!************************************************************************
      module optim
!************************************************************************
! OPTIMIZATION MODULE
! for basis set and geometry optimization procedures
! The calling routines are in basopt.f and geomopt.f 
!************************************************************************

      double precision, parameter :: ang2bohr = 1.d0/0.52917720859d0
      double precision, parameter :: PI = 3.141592653589793d0
      integer,parameter           :: lstr=512                     ! assumed max length of a line in the file MINP

      integer,parameter           :: iterfile=98                  ! file ITER it stores the interation number of geomopt
      integer,parameter           :: icoordfile = 78              ! file number for COORD.int
      integer,parameter           :: minptmpf = 79                ! MINP.tmp
      integer,parameter           :: mheaderf=80, mfooterf=81     ! MINP.header and MINP.footer - temporary files without geom spec
      integer,parameter           :: cenergies=99                 ! COORD.xyz.energies file for MOLDEN
      integer,parameter           :: crmsforce=97                 ! COORD.xyz.rforce file for MOLDEN (rms force)
      integer,parameter           :: cmaxforce=96                 ! COORD.xyz.mforce file for MOLDEN (max force)
      integer,parameter           :: coordfile=100                ! COORD.xyz.i file current geometry at step i
      integer,parameter           :: geomsfile=101                ! COORD.xyz.geometries file for MOLDEN it contains all the geoms during the optimization
      integer,parameter           :: pmatfile=102                 ! scr file to hold projection matrix element, PMAT.int
      integer,parameter           :: ifcfile=20                   ! file iface; MRCC interface file
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! ARRAYS, TYPES 
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      character(len=lstr), dimension(:), allocatable  :: alloc_zpname   ! array of types/parameter names in the Z-matrix
      double precision,    dimension(:), allocatable  :: alloc_zpvalue  ! array of parameter values in the Z-matrix

      character(len=lstr), dimension(:), allocatable  :: alloc_pname    ! array of types/parameter names 
      double precision,    dimension(:), allocatable  :: alloc_fgopt    ! array of parameter values
      integer,             dimension(:), allocatable  :: alloc_igopt    ! whether or not the given parameter should be optimized
      character(len=2),    dimension(:), allocatable  :: alloc_cgopt    ! to hold the special marks '++', '--', or '  '

      integer, dimension(:),             allocatable  :: alloc_atnums   ! array of atomic numbers
      character(len=2), dimension(:),    allocatable  :: alloc_atsymbols! array of atomic symbols
      integer, dimension(:,:),           allocatable  :: alloc_rim      ! redundant internal coordinate matrix

      double precision, dimension(:,:),  allocatable  :: alloc_coords   ! Cartesian coords
      double precision, dimension(:),    allocatable  :: alloc_intcoords! internal coords
      double precision, dimension(:,:),  allocatable  :: alloc_hessian  ! initial Hessian matrix

      type llist
        integer :: info
        type (llist), pointer :: next
      end type llist

      contains

!************************************************************************
      double precision function epsilon_()
!************************************************************************
! it gives back machine epsilon
!************************************************************************
      implicit none

      epsilon_ = 1.d0
      do while (1.d0 + 0.5d0*epsilon_ .gt. 1.d0)
        epsilon_ = 0.5d0*epsilon_
      enddo
      end function epsilon_
!************************************************************************

!************************************************************************
        subroutine cfillzero(str_array,n)
!************************************************************************
! This subroutine initialize a character array                          *
!************************************************************************
      implicit none
      integer,intent(in) :: n
      character(len=*),dimension(:),intent(inout) :: str_array
      !local
      integer :: i

      do i=1,n
        str_array(i) = ''
      enddo

      return
      end subroutine cfillzero
!************************************************************************


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! SIMPLEX ALGORITHM
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      subroutine simplex(f2min,args,ftol,xtol,maxiter,maxfun,fopt, &
      niter,nfeval,plevel,bool)
!************************************************************************
! The downhill simplex algorithm of Nelder and Mead;
! J.A. Nelder, R. Mead: Computer J., 7, 308 (1965);
! written by J Csontos
!
! The implementation is based on that of the numpy project
!

      implicit none 
      double precision, intent(in) :: ftol,xtol
      double precision, intent(out) :: fopt
      integer, intent(in) :: maxfun,maxiter,plevel
      integer, intent(out) :: niter,nfeval
      double precision, dimension(:), intent(inout) :: args
      logical, intent(out) :: bool 

!     local variables
      logical :: doshrink=.False.
      integer :: i,j,ndim
      double precision :: functol,argtol,fxr,fxe,fxc,fxcc
      ! vertj is an array for the function values at the vertices of the simplex
      double precision, dimension(size(args)) :: vertj,xavg,xr,xe,xc,xcc
      double precision, dimension(size(args)+1) :: fsim
      integer, dimension(size(args)+1) :: order
      double precision, dimension(size(args),size(args)+1) :: sim
      double precision, dimension(size(args),size(args)+1) :: tmp
      ! the parameters for the transformation of the simplex
      ! rho  -> factor of reflection
      ! psi  -> factor of contraction
      ! chi  -> factor of expansion
      ! del  -> factor for generating the initial simplex
      ! zdel -> factor for generating the initial simplex
      double precision, parameter :: rho=1.d0,chi=2.d0, & 
                                     psi=0.5d0, sigma=0.5d0
      double precision, parameter :: del=0.05d0, zdel=0.00025d0
      double precision, parameter :: eps=1d-6

      interface
        double precision function f2min(z)
          double precision, dimension(:), intent(inout) :: z
        end function
      end interface

! ***
! f2min   ->  external function to be minimized
! args()  ->  independent variables of f2min;
!         ->  as an input:  initial guess
!         ->  as an output: coordinates of the minimum
!
! ftol    ->  convergence criterion for the function values
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! fopt    ->  minimal value of f2min
! niter   ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! ***

      bool = .False.
      ndim = size(args) ! dimension of f2min
      nfeval = 0

      do j=1,ndim+1
        do i=1,ndim
          sim(i,j)=0.d0
        enddo
        fsim(j)=0.d0 ! array for function values at the vertices
      enddo

      fsim(1)=f2min(args)
      nfeval=nfeval+1

      ! generation of the initial simplex
      ! row-index stands for the coordinates of a given vertex (n)
      ! column index stands for the vertices (n+1)
      do i=1,ndim
          sim(i,1)=args(i)
          vertj(i)=0.d0 ! array for function arguments at the vertices
      enddo

      do j=2,ndim+1
         do i=1,ndim
            if(i.eq.(j-1)) then
              if(dabs(args(i)-0.d0).lt.eps) then
                sim(i,j)=zdel
              else
                sim(i,j)=(1.d0+del)*args(i)
              endif
            else
              sim(i,j)=args(i)
            endif
            vertj(i)=sim(i,j)
         enddo
         fsim(j)=f2min(vertj)
      enddo

      nfeval=nfeval+ndim
      ! rearrange the simplex points according to increasing function values
      ! best point first, worst point last
      ! the first column of sim contains the coordinates of the best point, etc.
      call quick_sort(fsim,order)
      call sort_mat(sim,order)
      niter=0

      if(plevel.gt.1) then 
        write(*,'(25x,3(a22))') &         
                'energy of worst point','energy change','step change'
        write(*,*)
      endif

      do while(nfeval<maxfun .and. niter<maxiter)
        argtol = fxtol(sim)
        functol = fftol(fsim)
        if(argtol.le.xtol.and.functol.le.ftol) exit ! convergence reached

        ! average arguments values without the worst point in the simplex
        call fxavg(sim,xavg)
        call transform(sim,xavg,xr)
        fxr = f2min(xr)
        nfeval = nfeval+1
        doshrink = .False.

        if(fxr.lt.fsim(1)) then
          call expansion(sim,xavg,xe)
          fxe = f2min(xe)
          nfeval = nfeval+1

          if(fxe.lt.fxr) then
            call columnR(sim,xe)
            fsim(ndim+1) = fxe
          else
            call columnR(sim,xr)
            fsim(ndim+1) = fxr
          endif

        else
          if(fxr.lt.fsim(ndim)) then
            call columnR(sim,xr)
            fsim(ndim+1) = fxr
          else
            if(fxr.lt.fsim(ndim+1)) then
              call contraction(sim,xavg,xc)
              fxc = f2min(xc)
              nfeval = nfeval+1

              if(fxc.le.fxr) then
                call columnR(sim,xc)
                fsim(ndim+1) = fxc
              else
                doshrink = .True.
              endif
            else
              call icontraction(sim,xavg,xcc)
              fxcc = f2min(xcc)
              nfeval = nfeval+1

              if(fxcc.lt.fsim(ndim+1)) then
                call columnR(sim,xcc)
                fsim(ndim+1) = fxcc
              else
                doshrink = .True.
              endif

              if(doshrink) then
                do i=2,ndim+1
                  call shrink(sim)
                  fsim(i) = f2min(sim(1:,i))
                enddo
                nfeval = nfeval+ndim
              endif

            endif
          endif
        endif
      
        ! again rearrange the simplex points according to increasing function values
        ! best point first, worst point last
        ! the first column of sim contains the coordinates of the best point, etc.
        call quick_sort(fsim,order)
        call sort_mat(sim,order)

        niter = niter+1
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,3(2x,f20.12))') &         
             'Iteration info ',niter,fsim(ndim+1),functol,argtol
        endif

      fopt = minval(fsim)
      args = sim(1:,1)
      enddo

      if(nfeval.ge.maxfun) then
        write(*,*) 'Maximum number of function evaluations &
                    has been exceeded.'
      else
        if(niter.ge.maxiter) then
          write(*,*) 'Maximum number of iterations &
                      has been exceeded.'
        else
          write(*,*) 'Optimization terminated successfully.'
          bool = .True.
        endif
      endif

!************************************************************************
      contains
!************************************************************************
!************************************************************************
      subroutine sort_mat(A,order)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! this subroutine rearrange its columns based on the given order

      implicit none
      double precision, dimension(:,:), intent(inout) :: A
      integer, dimension(size(A)+1), intent(in) :: order
!     local variables
      integer :: i,j,ndim
      double precision, dimension(size(A,1),size(A,1)+1) :: tmp

      ndim=size(A,1)

      do j=1,ndim+1
        do i=1,ndim
          tmp(i,j) = A(i,order(j))
        enddo
      enddo

      A = tmp
      end subroutine sort_mat
!************************************************************************
      subroutine shrink(A)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! this subroutine shrinks the simplex around the best point

      implicit none
      double precision, dimension(:,:), intent(inout) :: A
!     local variables
      integer :: i,j,ndim

      ndim=size(A,1)

      do j=2,ndim+1
        do i=1,ndim
          A(i,j) = A(i,1) + sigma*(A(i,j) - A(i,1))
        enddo
      enddo
      if(plevel.gt.2) write(*,*) 'shrink'

      end subroutine shrink 
!************************************************************************
      subroutine columnR(A,column)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! we only care about the last column - that is the worst point
! this subroutine replaces it with the vector column

      implicit none
      double precision, dimension(:,:), intent(inout) :: A
      double precision, dimension(size(A,1)), intent(in) :: column
!     local variables
      integer :: i,ndim

      ndim=size(A,1)

      do i=1,ndim
        A(i,ndim+1) = column(i)
      enddo

      end subroutine columnR
!************************************************************************
      subroutine icontraction(A,vec,vertex)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! we only care about the last column - that is the worst point
! this subroutine transform it using vec to a hopefully better place 
! vertex is the new point

      implicit none
      double precision, dimension(:,:), intent(in) :: A
      double precision, dimension(size(A,1)), intent(in) :: vec
      double precision, dimension(size(A,1)), intent(out) :: vertex
!     local variables
      integer :: i,ndim

      ndim=size(A,1)

      do i=1,ndim
        vertex(i) = (1.d0-psi)*vec(i) + psi*A(i,ndim+1)
      enddo
      if(plevel.gt.2) write(*,*) 'icontraction'

      end subroutine icontraction
!************************************************************************
      subroutine contraction(A,vec,vertex)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! we only care about the last column - that is the worst point
! this subroutine transform it using vec to a hopefully better place 
! vertex is the new point

      implicit none
      double precision, dimension(:,:), intent(in) :: A
      double precision, dimension(size(A,1)), intent(in) :: vec
      double precision, dimension(size(A,1)), intent(out) :: vertex
!     local variables
      integer :: i,ndim

      ndim=size(A,1)

      do i=1,ndim
        vertex(i) = (1.d0+psi*rho)*vec(i) - psi*rho*A(i,ndim+1)
      enddo
      if(plevel.gt.2) write(*,*) 'contraction'

      end subroutine contraction
!************************************************************************
      subroutine expansion(A,vec,vertex)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! we only care about the last column - that is the worst point
! this subroutine transform it using vec to a hopefully better place 
! vertex is the new point

      implicit none
      double precision, dimension(:,:), intent(in) :: A
      double precision, dimension(size(A,1)), intent(in) :: vec
      double precision, dimension(size(A,1)), intent(out) :: vertex
!     local variables
      integer :: i,ndim

      ndim=size(A,1)

      do i=1,ndim
        vertex(i) = (1.d0+rho*chi)*vec(i) - rho*chi*A(i,ndim+1)
      enddo
      if(plevel.gt.2) write(*,*) 'expansion'

      end subroutine expansion
!************************************************************************
      subroutine transform(A,vec,vertex)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! we only care about the last column - that is the worst point
! this subroutine transform it using vec to a hopefully better place 
! vertex is the new point

      implicit none
      double precision, dimension(:,:), intent(in) :: A
      double precision, dimension(size(A,1)), intent(in) :: vec
      double precision, dimension(size(A,1)), intent(out) :: vertex
!     local variables
      integer :: i,ndim

      ndim=size(A,1)

      do i=1,ndim
        vertex(i) = (1.d0+rho)*vec(i) - rho*A(i,ndim+1)
      enddo
      if(plevel.gt.2) write(*,*) 'transform'

      end subroutine transform
!************************************************************************
      double precision function fftol(y)
! y is a (ndim+1) dimension vector and it contains
! the fucntion values at the vertices
! this function returns the largest difference
! relative to the lowest function value

      implicit none
      double precision, dimension(:), intent(in) :: y
!     local variables
      integer :: i,ndim
      double precision, dimension(size(y)-1) :: z

      ndim=size(y)

      do i=2,ndim
        z(i-1)=dabs(y(i)-y(1))
      enddo

      fftol=maxval(z) ! it retuns the maximum of the vector z
      return 

      end function fftol
!************************************************************************
      double precision function fxtol(A)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! this function returns the maximum displacement of the 
! x-coordinates relative to the first point

      implicit none
      double precision, dimension(:,:), intent(in) :: A
!     local variables
      integer :: i,j,ndim
      double precision, dimension(size(A,1),size(A,1)) :: B

      ndim=size(A,1)

      do j=2,ndim+1
        do i=1,ndim
          B(i,j-1)=dabs(A(i,j)-A(i,1))
        enddo
      enddo

      fxtol=maxval(B) ! it retuns the maximum of the array B
      return 

      end function fxtol
!************************************************************************
      subroutine fxavg(A,avg)
! A is a (ndim)*(ndim+1) dimension matrix
! the columns of A contains the points of the simplex
! we don't care about the last column - that is the worst point
! this subroutine calculates the average of the arguments and 
! return the average x-coordinates in vector avg

      implicit none
      double precision, dimension(:,:), intent(in) :: A
      double precision, dimension(size(A,1)), intent(out) :: avg
!     local variables
      integer :: i,j,ndim
      double precision :: sum_j

      ndim=size(A,1)

      do i=1,ndim
        sum_j = 0.d0
        do j=1,ndim
          sum_j = sum_j + A(i,j)
        enddo
        avg(i) = sum_j/ndim
      enddo

      end subroutine fxavg
!************************************************************************

      end subroutine simplex
!************************************************************************

!************************************************************************
      recursive subroutine quick_sort(list, order)

! Quick sort routine from:
! Brainerd, W.S., Goldberg, C.H. & Adams, J.C. (1990) "Programmer's Guide to
! Fortran 90", McGraw-Hill  ISBN 0-07-000248-7, pages 149-150.
! Modified by Alan Miller to include an associated integer array which gives
! the positions of the elements in the original order.

      implicit none
      double precision, dimension (:), intent(inout)  :: list
      integer, dimension (:), intent(out)  :: order

!     local variable
      integer :: i

      do i = 1, size(list)
        order(i) = i
      enddo

      call quick_sort_1(1, size(list))

!************************************************************************
      contains
!************************************************************************
      recursive subroutine quick_sort_1(left_end, right_end)

      integer, intent(in) :: left_end, right_end

!     local variables
      integer             :: i, j, itemp
      double precision    :: reference, temp
      integer, parameter  :: max_simple_sort_size = 6

      if (right_end < left_end + max_simple_sort_size) then
        ! Use interchange sort for small lists
        call interchange_sort(left_end, right_end)

      else
        ! Use partition ("quick") sort
        reference = list((left_end + right_end)/2)
        i = left_end - 1; j = right_end + 1

        do
          ! Scan list from left end until element >= reference is found
          do
            i = i + 1
            if (list(i) >= reference) exit
          enddo
          ! Scan list from right end until element <= reference is found
          do
            j = j - 1
            if (list(j) <= reference) exit
          enddo


          if (i < j) then
          ! Swap two out-of-order elements
            temp = list(i); list(i) = list(j); list(j) = temp
            itemp = order(i); order(i) = order(j); order(j) = itemp
          elseif (i == j) then
              i = i + 1
              exit
          else
              exit
          endif
        enddo

        if (left_end < j)  call quick_sort_1(left_end, j)
        if (i < right_end) call quick_sort_1(i, right_end)
      endif

      end subroutine quick_sort_1
!************************************************************************

!************************************************************************
      subroutine interchange_sort(left_end, right_end)

      integer, intent(in) :: left_end, right_end

!     local variables
      integer             :: i, j, itemp
      double precision    :: temp

      do i = left_end, right_end - 1
        do j = i+1, right_end
          if (list(i) > list(j)) then
            temp = list(i); list(i) = list(j); list(j) = temp
            itemp = order(i); order(i) = order(j); order(j) = itemp
          endif
        enddo
      enddo

      end subroutine interchange_sort
!************************************************************************

      end subroutine quick_sort
!************************************************************************

!************************************************************************
      subroutine build_unit_Hessian(H)
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(inout) :: H
      !local
      integer :: i,j,ndim

      ndim = size(H,1)

      do j=1,ndim             ! initialization of arrays
        do i=1,ndim
          if(i .eq. j) then
            H(i,j) = 1.d0
          else
            H(i,j) = 0.d0
          endif
        enddo
      enddo

      end subroutine build_unit_Hessian
!************************************************************************

!************************************************************************
      subroutine get_Lindh_params(atnums,i,j,alpha,r)
!************************************************************************
      implicit none
      integer,dimension(:),intent(in) :: atnums
      integer, intent(in) :: i,j
      double precision,intent(out) :: alpha,r
      !local
      double precision, dimension(3,3) :: A_ij,R_ij

      A_ij  = reshape((/1.0000,0.3949,0.3949,&
                        0.3949,0.2800,0.2800,&
                        0.3949,0.2800,0.2800/), shape(A_ij))

      R_ij  = reshape((/1.35,2.10,2.53,&
                        2.10,2.87,3.40,&
                        2.53,3.40,3.40/), shape(R_ij))

      if(atnums(i).le.2) then
        if(atnums(j).le.2) then
          alpha = A_ij(1,1)
          r     = R_ij(1,1)
        elseif(atnums(j).le.10) then
          alpha = A_ij(1,2)
          r     = R_ij(1,2)
        elseif(atnums(j).le.18) then
          alpha = A_ij(1,3)
          r     = R_ij(1,3)
        else
          alpha = 0.d0 ! no Lindh parameter for this atom
          r     = 0.d0 ! no Lindh parameter for this atom
          write(*,*)'No Lindh parameter is available for this atom:', &
                     atnums(j)
        endif

      elseif(atnums(i).le.10) then
        if(atnums(j).le.2) then
          alpha = A_ij(2,1)
          r     = R_ij(2,1)
        elseif(atnums(j).le.10) then
          alpha = A_ij(2,2)
          r     = R_ij(2,2)
        elseif(atnums(j).le.18) then
          alpha = A_ij(2,3)
          r     = R_ij(2,3)
        else
          alpha = 0.d0 ! no Lindh parameter for this atom
          r     = 0.d0 ! no Lindh parameter for this atom
          write(*,*)'No Lindh parameter is available for this atom:', &
                     atnums(j)
        endif

      elseif(atnums(i).le.18) then
        if(atnums(j).le.2) then
          alpha = A_ij(3,1)
          r     = R_ij(3,1)
        elseif(atnums(j).le.10) then
          alpha = A_ij(3,2)
          r     = R_ij(3,2)
        elseif(atnums(j).le.18) then
          alpha = A_ij(3,3)
          r     = R_ij(3,3)
        else
          alpha = 0.d0 ! no Lindh parameter for this atom
          r     = 0.d0 ! no Lindh parameter for this atom
          write(*,*)'No Lindh parameter is available for this atom:', &
                     atnums(j)
        endif

      else
          alpha = 0.d0 ! no Lindh parameter for this atom
          r     = 0.d0 ! no Lindh parameter for this atom
          write(*,*)'No Lindh parameter is available for this atom:', &
                     atnums(i)
 
      endif


      end subroutine get_Lindh_params
!************************************************************************

!************************************************************************
      subroutine build_Lindh_Hessian(H,red_int_matrix,coords,atnums)
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(inout) :: H
      integer,dimension(:,:),intent(in) :: red_int_matrix
      double precision,dimension(:,:),intent(in) :: coords
      integer,dimension(:),intent(in) :: atnums
      !local
      double precision, parameter :: k_r  =  0.45  !Lindh et al, CPL, 241, 423 (1995)
      double precision, parameter :: k_f  =  0.15  !Lindh et al, CPL, 241, 423 (1995)
      double precision, parameter :: k_t  =  0.005 !Lindh et al, CPL, 241, 423 (1995)
      integer :: i,j,k,l,m,n,ndim,itype
      double precision :: r_ij,r_jk,r_kl
      double precision :: a_ij,a_jk,a_kl
      double precision :: rho_ij,rho_jk,rho_kl
      double precision :: r_refij,r_refjk,r_refkl

      ndim = size(H,1)

      do n=1,ndim             ! initialization of arrays
        do m=1,ndim
          if(m .eq. n) then
            itype=red_int_matrix(1,m) ! type of internal coord
            i    =red_int_matrix(2,m) ! 1st atom
            j    =red_int_matrix(3,m) ! 2nd atom
            k    =red_int_matrix(4,m) ! 3rd atom
            l    =red_int_matrix(5,m) ! 4th atom


            if(itype.eq.1 .or. itype.eq.4) then ! internal bond

              call get_Lindh_params(atnums,i,j,a_ij,r_refij)
              r_ij   = distance2(coords(1:3,i),coords(1:3,j))
              rho_ij = dexp(a_ij*(r_refij**2 - r_ij**2))

              H(m,m) = (k_r*rho_ij)
            endif
            if(itype.eq.2) then ! angle 

              call get_Lindh_params(atnums,i,j,a_ij,r_refij)
              call get_Lindh_params(atnums,j,k,a_jk,r_refjk)
              r_ij = distance2(coords(1:3,i),coords(1:3,j))
              r_jk = distance2(coords(1:3,j),coords(1:3,k))
              rho_ij = dexp(a_ij*(r_refij**2 - r_ij**2))
              rho_jk = dexp(a_jk*(r_refjk**2 - r_jk**2))

              H(m,m) = (k_f*rho_ij*rho_jk)
!             write(*,*)'Lindh angle: ', rho_ij,rho_jk
            endif
            if(itype.eq.3) then ! dihedral

              call get_Lindh_params(atnums,i,j,a_ij,r_refij)
              call get_Lindh_params(atnums,j,k,a_jk,r_refjk)
              call get_Lindh_params(atnums,k,l,a_kl,r_refkl)
              r_ij = distance2(coords(1:3,i),coords(1:3,j))
              r_jk = distance2(coords(1:3,j),coords(1:3,k))
              r_kl = distance2(coords(1:3,k),coords(1:3,l))
              rho_ij = dexp(a_ij*(r_refij**2 - r_ij**2))
              rho_jk = dexp(a_jk*(r_refjk**2 - r_jk**2))
              rho_kl = dexp(a_kl*(r_refkl**2 - r_kl**2))

              H(m,m) = (k_t*rho_ij*rho_jk*rho_kl)
!             write(*,*)'Lindh dihedral: ', rho_kl,r_kl,r_refkl
            endif

          else
            H(m,n) = 0.d0
          endif
        enddo
      enddo

!     call print_matrixR(H,9)
!     call mrccend(1)
      end subroutine build_Lindh_Hessian
!************************************************************************

!************************************************************************
      subroutine build_initial_Hessian(H,red_int_matrix,coords)
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(inout) :: H
      integer,dimension(:,:),intent(in) :: red_int_matrix
      double precision,dimension(:,:),intent(in) :: coords
      !local
      integer :: i,j,k,l,m,n,ndim,itype

      ndim = size(H,1)

      do n=1,ndim             ! initialization of arrays
        do m=1,ndim
          if(m .eq. n) then
            itype=red_int_matrix(1,m) ! type of internal coord
            i    =red_int_matrix(2,m) ! 1st atom
            j    =red_int_matrix(3,m) ! 2nd atom
            k    =red_int_matrix(4,m) ! 3rd atom
            l    =red_int_matrix(5,m) ! 4th atom

            if(itype.eq.1) then ! regular bond
              H(m,m) = 1/0.5d0
            endif
            if(itype.eq.2) then ! angle 
              H(m,m) = 1/0.2d0
            endif
            if(itype.eq.3) then ! dihedral
              H(m,m) = 1/0.1d0
            endif
            if(itype.eq.4) then ! auxiliary bond
              H(m,m) = 1/0.2d0
            endif
          else
            H(m,n) = 0.d0
          endif
        enddo
      enddo

!     call print_matrixR(H,9)
!     call mrccend(1)
      end subroutine build_initial_Hessian
!************************************************************************

!************************************************************************
      subroutine get_Schlegel_param(atnums,i,j,B)
!************************************************************************
      implicit none
      integer,dimension(:),intent(in) :: atnums
      integer, intent(in) :: i,j
      double precision,intent(out) :: B
      !local
      double precision, dimension(3,3) :: B_ij

      B_ij  = reshape((/-0.244, 0.352, 0.660,&
                         0.352, 1.085, 1.522,&
                         0.660, 1.552, 2.068/), shape(B_ij))

      if(atnums(i).le.2) then
        if(atnums(j).le.2) then
          B = B_ij(1,1)
        elseif(atnums(j).le.10) then
          B = B_ij(1,2)
        elseif(atnums(j).le.18) then
          B = B_ij(1,3)
        else
          B = 0.d0 ! no Schlegel parameter for this atom
          write(*,*) &
                 'No Schlegel parameter is available for this atom:', &
                     atnums(j)
        endif

      elseif(atnums(i).le.10) then
        if(atnums(j).le.2) then
          B = B_ij(2,1)
        elseif(atnums(j).le.10) then
          B = B_ij(2,2)
        elseif(atnums(j).le.18) then
          B = B_ij(2,3)
        else
          B = 0.d0 ! no Schlegel parameter for this atom
          write(*,*) &
                 'No Schlegel parameter is available for this atom:', &
                     atnums(j)
        endif

      elseif(atnums(i).le.18) then
        if(atnums(j).le.2) then
          B = B_ij(3,1)
        elseif(atnums(j).le.10) then
          B = B_ij(3,2)
        elseif(atnums(j).le.18) then
          B = B_ij(3,3)
        else
          B = 0.d0 ! no Schlegel parameter for this atom
          write(*,*) &
                 'No Schlegel parameter is available for this atom:', &
                     atnums(j)
        endif

      else
          B = 0.d0 ! no Schlegel parameter for this atom
          write(*,*) &
                 'No Schlegel parameter is available for this atom:', &
                     atnums(j)
 
      endif


      end subroutine get_Schlegel_param
!************************************************************************

!************************************************************************
      subroutine build_Schlegel_Hessian(H,red_int_matrix,coords,atnums)
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(inout) :: H
      integer,dimension(:,:),intent(in) :: red_int_matrix
      double precision,dimension(:,:),intent(in) :: coords
      integer,dimension(:),intent(in) :: atnums
      !local
      integer :: i,j,k,l,m,n,ndim,itype
      double precision :: r,r_cov,B,F
      double precision, parameter :: fac   =  1.3d0    !Schlegel, TCA, 66, 333 (1984) factror for minimal basis calcs
      double precision, parameter :: A     =  1.734d0  !Schlegel, TCA, 66, 333 (1984) constant A for bonds
      double precision, parameter :: AbH   =  0.160d0  !Schlegel, TCA, 66, 333 (1984) constant A for bends with hydrogens
      double precision, parameter :: Ab    =  0.250d0  !Schlegel, TCA, 66, 333 (1984) constant A for bends
      double precision, parameter :: At    =  0.0023d0 !Schlegel, TCA, 66, 333 (1984) constant A for torsions 
      double precision, parameter :: Bt    =  0.0700d0 !Schlegel, TCA, 66, 333 (1984) constant B for torsions 
      double precision, parameter :: Aoup  =  0.0450d0 !Schlegel, TCA, 66, 333 (1984) constant A for out of plane motions
!     d = 1.d0-r1*(r2 x r3)/(|r1||r2||r3|)
!     H = Aoup*d**4

      ndim = size(H,1)

      do n=1,ndim             ! initialization of arrays
        do m=1,ndim
          if(m .eq. n) then
            itype=red_int_matrix(1,m) ! type of internal coord
            i    =red_int_matrix(2,m) ! 1st atom
            j    =red_int_matrix(3,m) ! 2nd atom
            k    =red_int_matrix(4,m) ! 3rd atom
            l    =red_int_matrix(5,m) ! 4th atom


            if(itype.eq.1 .or. itype.eq.4) then ! internal bond

              call get_Schlegel_param(atnums,i,j,B)
              r = distance2(coords(1:3,i),coords(1:3,j))
              F = fac*A/((r-B)**3)

              H(m,m) = F
            endif
            if(itype.eq.2) then ! angle 
              if(atnums(i).eq.1 .or. atnums(k).eq.1) then
                F = fac*AbH
              else
                F = fac*Ab
              endif
              H(m,m) = F
            endif
            if(itype.eq.3) then ! dihedral
              r_cov = get_cov_radius(atnums(j)) 
              r_cov = r_cov + get_cov_radius(atnums(k))
              r = distance2(coords(1:3,j),coords(1:3,k))
              F = fac*(At - Bt*(r - r_cov))
              H(m,m) = dabs(F)
            endif

          else
            H(m,n) = 0.d0
          endif
        enddo
      enddo


!     call print_matrixR(H,9)
!     call mrccend(1)
      end subroutine build_Schlegel_Hessian
!************************************************************************


!************************************************************************
      subroutine wrt_molden(niter,energy,gnorm,gmax)
! it writes energies and geometries suitable to append to file MOLDEN

      implicit none
      integer, intent(in)          :: niter
      double precision, intent(in) :: energy,gnorm,gmax
      !local
      character(len=512) :: si,scommand

      if(niter .eq. 0) then
#ifdef WINTEL
        call ishell('echo [GEOCONV]>COORD.xyz.energies')
        call ishell('echo energy>>COORD.xyz.energies')
#else
        call ishell('echo "[GEOCONV]" > COORD.xyz.energies')
        call ishell('echo "energy" >> COORD.xyz.energies')
#endif
        open(cenergies,file='COORD.xyz.energies',position='append')
        write(cenergies,103) energy
        close(cenergies)
#ifdef WINTEL
        call ishell('echo rms-force>>COORD.xyz.rforce')
#else
        call ishell('echo "rms-force" >> COORD.xyz.rforce')
#endif
        open(crmsforce,file='COORD.xyz.rforce',position='append')
        write(crmsforce,103) gnorm
        close(crmsforce)
#ifdef WINTEL
        call ishell('echo max-force>>COORD.xyz.mforce')
#else
        call ishell('echo "max-force" >> COORD.xyz.mforce')
#endif
        open(cmaxforce,file='COORD.xyz.mforce',position='append')
        write(cmaxforce,103) gmax
        close(cmaxforce)

        call ishell('cp COORD.xyz COORD.xyz.0')
#ifdef WINTEL
        call ishell('echo [GEOMETRIES] (XYZ)>COORD.xyz.geometries')
#else
        call ishell('echo "[GEOMETRIES] (XYZ)" > COORD.xyz.geometries')
#endif
        call wrt_xyz_geoms(niter)
      else
        open(cenergies,file='COORD.xyz.energies',position='append')
        write(cenergies,103) energy
        close(cenergies)

        open(crmsforce,file='COORD.xyz.rforce',position='append')
        write(crmsforce,103) gnorm
        close(crmsforce)

        open(cmaxforce,file='COORD.xyz.mforce',position='append')
        write(cmaxforce,103) gmax
        close(cmaxforce)

        si = int2char(niter)
        scommand='cp COORD.xyz COORD.xyz.'//trim(adjustl(si))
        call ishell(scommand)
        call wrt_xyz_geoms(niter)

      endif

103   format(f22.12,a2)

      end subroutine wrt_molden
!************************************************************************
      subroutine wrt_xyz_geoms(niter)
! it appends the current geometry to the file COORD.xyx.geometries

      implicit none
      integer, intent(in) :: niter
      !local
      character(len=512) :: fname,line,s10
      integer :: i,natoms,l

      fname = 'COORD.xyz.'//trim(adjustl(int2char(niter)))

      open(unit=coordfile,file=fname,status='old')                      ! current geometry at step niter
      open(unit=geomsfile,file='COORD.xyz.geometries',position='append')! all geometries concatenated to this file

      read(coordfile,'(I10)') natoms ! first line, number of atoms
      s10 = int2char(natoms)
      l   = len_trim(adjustl(s10))
      write(geomsfile,'(a)') s10(1:l)
 
      read(coordfile,104) line       ! comment line
      l = len_trim(adjustl(line))
      write(geomsfile,'(a)') line(1:l)

      do i=1,natoms
        read(coordfile,104) line
        l = len_trim(adjustl(line))
        write(geomsfile,'(a)') line(1:l)
      enddo

      close(coordfile)
      close(geomsfile)

104   format(a512)

      end subroutine wrt_xyz_geoms
!************************************************************************
      subroutine wrt_iter(niter)
! it writes the current iteration number into the file ITER

      implicit none
      integer, intent(in) :: niter

      open(unit=iterfile,file='ITER') ! # interations is written in the file ITER
      write(iterfile,*) niter
      close(iterfile)

      end subroutine wrt_iter
!************************************************************************
      subroutine read_iter(niter)
      !it reads the current iteration number from the file ITER

      implicit none
      integer, intent(out) :: niter

      open(unit=iterfile,file='ITER') ! # interation number is read from the file ITER
      read(iterfile,*) niter
      close(iterfile)

      end subroutine read_iter
!************************************************************************
      function outer_product(vecA,vecB)
! it returns the outer product, vecA'vecB, of vecA and vecB
! row vectors are assumed, therefore transpose of vecA times vecB will 
! be returned (dimA,dimB)

      implicit none
      double precision, dimension(:), intent(in) :: vecA,vecB
      double precision, dimension(size(vecA),size(vecB)) ::outer_product
      !local
      integer i,j,dimA,dimB

      dimA = size(vecA)
      dimB = size(vecB)

      do j=1,dimB
        do i=1,dimA
          outer_product(i,j) = vecA(i) * vecB(j)
        enddo
      enddo

      end function outer_product
!************************************************************************
      function direction(H,grad)
! it gives the search direction back as a vector

      implicit none
      double precision, dimension(:,:), intent(in) :: H
      double precision, dimension(:), intent(in) :: grad
      double precision, dimension(size(grad)) :: direction
      !local
      integer i,j,ndim

      ndim = size(grad)
      do i=1,ndim
        direction(i) = 0.d0
      enddo

      do j=1,ndim
        do i=1,ndim
          direction(j) = direction(j) - H(i,j)*grad(i)
!         write(*,*) H(i,J),grad(i),direction(j)
        enddo
      enddo

      end function direction
!************************************************************************

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! GDIIS ALGORITHM
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      subroutine diis_extrap(niter,ndmax,xk,gk,Hk,sk)
!************************************************************************
! DIIS extrapolation                                                   
! nact - iteration step
! nmax - dimension of the vector to be extrapolated
! pvec - the updated vector
! evec - the original vector
! ndmax - maximum no. of iterations
!************************************************************************

      implicit none
      integer,intent(in) :: niter,ndmax
      double precision, dimension(:), intent(inout) :: xk,gk ! geometry and gradient
      double precision, dimension(:,:), intent(in)  :: Hk    ! inverse Hessian matrix
      double precision, dimension(:), intent(inout) :: sk    ! step(error)
      !local
      double precision, dimension(:), allocatable   :: pvec ! updated vector
      double precision, dimension(:), allocatable   :: evec ! original vector
      integer :: nmax
      double precision, dimension(:), allocatable           :: cvec ! vector of coefficients
      double precision, dimension(:,:), allocatable         :: bmat ! scalar product B matrix
      double precision, dimension(:,:), allocatable         :: invbmat ! inverse of the B matrix
      double precision, dimension(:), allocatable           :: bvec ! inverse of the B matrix
      integer, parameter :: diisfile=110,errfile=111,ifltln=8 !MRCC common file!
      integer, parameter :: gdiisfile=112

      nmax = size(xk)

      allocate(pvec(nmax))
      allocate(evec(nmax))
      allocate(cvec(ndmax))
      allocate(bmat(ndmax,ndmax))
      allocate(invbmat(ndmax,ndmax))
      allocate(bvec(ndmax*ndmax))

      cvec = 0.d0

      evec = -1.d0*matrix_dot_vec(Hk,gk)
      write(*,*) 'trust diis evec'
      call print_vectorR(evec,9)

!     evec = sk
      write(*,*) 'trust diis xk'
      call print_vectorR(xk,9)
      write(*,*) 'trust diis gk'
      call print_vectorR(gk,9)
      pvec = xk

      call diismod(niter,nmax,pvec,evec,gk,ndmax,diisfile,errfile,&
                   gdiisfile,ifltln,bmat,bvec,invbmat,cvec)

      xk = xk - 1.d0*matrix_dot_vec(Hk,gk)
      write(*,*) 'trust diis xk after', cvec
      call print_vectorR(xk,9)
      write(*,*) 'trust diis gk'
      call print_vectorR(gk,9)

      deallocate(pvec)
      deallocate(evec)
      deallocate(cvec)
      deallocate(bmat)
      deallocate(invbmat)
      deallocate(bvec)

      end subroutine diis_extrap
!************************************************************************

!************************************************************************
      subroutine diismod(nact,nmax,pvec,evec,gvec,ndmax,diisfile,&
                         errfile,gdiisfile,ifltln,bmat,bvec,invbmat,&
                         cvec)
!************************************************************************
!* DIIS extrapolation                                                   *
!* nact - iteration step
!* nmax - dimension of the vector to be extrapolated
!* pvec - the updated vector
!* evec - the original vector
!* ndmax - maximum no. of iterations
!************************************************************************
      implicit none
      integer nact,iact,ii,i,j,nmax,file,ndmax,iact1,iact2,diisfile
      integer gdiisfile,errfile,ifltln
      integer ipiv(ndmax),ndiis
      real*8 bmat(ndmax,ndmax),bvec(ndmax**2),invbmat(ndmax,ndmax),ddot
      real*8 cvec(ndmax),sum,sum1,evec(nmax),scr(nmax),pvec(nmax)
      real*8 gvec(nmax)
!
      ndiis=8
      open(diisfile,status='unknown',access='direct',recl=nmax*ifltln)
      open(errfile,status='unknown',access='direct',recl=nmax*ifltln)
      open(gdiisfile,status='unknown',access='direct',recl=nmax*ifltln)
! Save amplitudes and calculate new error vector
!        call dfillzero(pvec,nmax)
        write(diisfile,rec=nact) pvec
!        do j=1,nmax
!         pvec(j)=pvec(j)-evec(j)
!        enddo
!       write(errfile,rec=nact) pvec
        write(errfile,rec=nact) evec
        write(gdiisfile,rec=nact) gvec
! Update the scalar product matrix (B matrix)
       do iact1=max(1,nact-ndiis+1),nact
        read(errfile,rec=iact1) pvec
         do iact2=max(1,nact-ndiis+1),iact1
          read(errfile,rec=iact2) scr
          sum=ddot(nmax,pvec,1,scr,1)
          bmat(iact1-max(0,nact-ndiis),iact2-max(0,nact-ndiis))=sum
          bmat(iact2-max(0,nact-ndiis),iact1-max(0,nact-ndiis))=sum
         enddo
       enddo
! Invert B matrix
       do j=1,min(nact,ndiis)
         do i=1,j
          bvec(i+(j-1)*j/2)=bmat(i,j)
         enddo
       enddo
      call dsptrf('U',min(nact,ndiis),bvec,ipiv,invbmat,i)
      call dsptri('U',min(nact,ndiis),bvec,ipiv,invbmat,i)
       do j=1,min(nact,ndiis)
         do i=1,j
          invbmat(i,j)=bvec(i+(j-1)*j/2)
          invbmat(j,i)=invbmat(i,j)
         enddo
       enddo
! Denominator
      sum=0.d0
       do i=1,min(nact,ndiis)
         do j=1,min(nact,ndiis)
          sum=sum+invbmat(i,j)
         enddo
       enddo
      sum=1.d0/sum
! Coefficients
       do i=1,min(nact,ndiis)
        sum1=0.d0
         do j=1,min(nact,ndiis)
          sum1=sum1+invbmat(i,j)
         enddo
        cvec(i)=sum*sum1
       enddo
! New amplitudes
       do i=1,nmax
        pvec(i)=0.d0
       enddo
       do iact=1,min(nact,ndiis)
        sum=cvec(iact)
        read(diisfile,rec=iact+max(0,nact-ndiis)) scr
         do j=1,nmax
          pvec(j)=pvec(j)+sum*scr(j)
         enddo
       enddo
! New gradient
!      do i=1,nmax
!       gvec(i)=0.d0
!      enddo
!      do iact=1,min(nact,ndiis)
!       sum=cvec(iact)
!       read(gdiisfile,rec=iact+max(0,nact-ndiis)) scr
!        do j=1,nmax
!         gvec(j)=gvec(j)+sum*scr(j)
!        enddo
!      enddo

      close(diisfile)
      close(errfile)
      close(gdiisfile)
!
      return
      end subroutine diismod
!***********************************************************************


!************************************************************************
      subroutine gdiisdel(nact,nmax,ifltln)
!************************************************************************
! this subroutine deletes the last gdiis information
!* DIIS extrapolation                                                   *
!* nact - iteration step
!* nmax - dimension of the vector to be extrapolated
!* pvec - the updated vector
!* evec - the original vector
!* ndmax - maximum no. of iterations
!************************************************************************
      implicit none
      integer, intent(in) :: nact,nmax,ifltln
      !local
      integer, parameter  :: diisfile=110,errfile=111, gdiisfile=112
      integer, parameter  :: diisftmp=210,errftmp=211, gdiisftmp=212
      integer :: i,istat
      real*8 pvec(nmax)
#ifdef WINTEL
      logical :: fileexist
#endif
!
#if !defined (gfortran) && !defined (G95)
      integer*4 system
      external system
#endif

      open(diisfile, file='DIIS.geo', status='unknown',access='direct',recl=nmax*ifltln)
      open(diisftmp, file='DIIS.tmp', status='unknown',access='direct',recl=nmax*ifltln)
      open(errfile,  file='DIISE.geo',status='unknown',access='direct',recl=nmax*ifltln)
      open(errftmp,  file='DIISE.tmp',status='unknown',access='direct',recl=nmax*ifltln)
      open(gdiisfile,file='DIISG.geo',status='unknown',access='direct',recl=nmax*ifltln)
      open(gdiisftmp,file='DIISG.tmp',status='unknown',access='direct',recl=nmax*ifltln)

      do i=1,nact-1
        read (diisfile,  rec=i) pvec
        write(diisftmp,  rec=i) pvec

        read (errfile,   rec=i) pvec
        write(errftmp,   rec=i) pvec

        read (gdiisfile, rec=i) pvec
        write(gdiisftmp, rec=i) pvec
      enddo
  
      close(diisfile)
      close(diisftmp)
      close(errfile)
      close(errftmp)
      close(gdiisfile)
      close(gdiisftmp)
!
#ifdef WINTEL
      inquire(file="DIIS.tmp", exist=fileexist)
      if(fileexist) istat=system('mv DIIS.tmp DIIS.geo')
      inquire(file="DIISE.tmp", exist=fileexist)
      if(fileexist) istat=system('mv DIISE.tmp DIISE.geo')
      inquire(file="DIISG.tmp", exist=fileexist)
      if(fileexist) istat=system('mv DIISG.tmp DIISG.geo')
#else
      istat=system('test -e DIIS.tmp')
      if(istat.eq.0) istat=system('mv DIIS.tmp DIIS.geo')
      istat=system('test -e DIISE.tmp')
      if(istat.eq.0) istat=system('mv DIISE.tmp DIISE.geo')
      istat=system('test -e DIISG.tmp')
      if(istat.eq.0) istat=system('mv DIISG.tmp DIISG.geo')
#endif
      

      return
      end subroutine gdiisdel
!***********************************************************************

!************************************************************************
      subroutine gdiismod(nact,nmax,pvec,evec,gvec,ndmax,ifltln,bmat,&
                          bvec,invbmat,cvec)
!************************************************************************
!* DIIS extrapolation                                                   *
!* nact - iteration step
!* nmax - dimension of the vector to be extrapolated
!* pvec - the updated vector
!* evec - the original vector
!* ndmax - maximum no. of iterations
!************************************************************************
      implicit none
      integer, intent(in) ::nact,nmax,ndmax,ifltln
      !local
      integer, parameter  :: diisfile=110,errfile=111, gdiisfile=112
      integer :: iact,ii,i,j,iact1,iact2
      integer :: ipiv(ndmax),ndiis
      real*8 bmat(ndmax,ndmax),bvec(ndmax**2),invbmat(ndmax,ndmax),ddot
      real*8 cvec(ndmax),sum,sum1,evec(nmax),scr(nmax),pvec(nmax)
      real*8 gvec(nmax)
!
      ndiis=4
      open(diisfile, file='DIIS.geo', status='unknown',access='direct',recl=nmax*ifltln)
      open(errfile,  file='DIISE.geo',status='unknown',access='direct',recl=nmax*ifltln)
      open(gdiisfile,file='DIISG.geo',status='unknown',access='direct',recl=nmax*ifltln)
! Save amplitudes and calculate new error vector
!        call dfillzero(pvec,nmax)
        write(diisfile,rec=nact) pvec
!        do j=1,nmax
!         pvec(j)=pvec(j)-evec(j)
!        enddo
!       write(errfile,rec=nact) pvec
        write(errfile,rec=nact) evec
        write(gdiisfile,rec=nact) gvec
! Update the scalar product matrix (B matrix)
       do iact1=max(1,nact-ndiis+1),nact
        read(errfile,rec=iact1) pvec
         do iact2=max(1,nact-ndiis+1),iact1
          read(errfile,rec=iact2) scr
          sum=ddot(nmax,pvec,1,scr,1)
          bmat(iact1-max(0,nact-ndiis),iact2-max(0,nact-ndiis))=sum
          bmat(iact2-max(0,nact-ndiis),iact1-max(0,nact-ndiis))=sum
         enddo
       enddo
! Invert B matrix
       do j=1,min(nact,ndiis)
         do i=1,j
          bvec(i+(j-1)*j/2)=bmat(i,j)
         enddo
       enddo
      call dsptrf('U',min(nact,ndiis),bvec,ipiv,invbmat,i)
      call dsptri('U',min(nact,ndiis),bvec,ipiv,invbmat,i)
       do j=1,min(nact,ndiis)
         do i=1,j
          invbmat(i,j)=bvec(i+(j-1)*j/2)
          invbmat(j,i)=invbmat(i,j)
         enddo
       enddo
! Denominator
      sum=0.d0
       do i=1,min(nact,ndiis)
         do j=1,min(nact,ndiis)
          sum=sum+invbmat(i,j)
         enddo
       enddo
      sum=1.d0/sum
! Coefficients
       do i=1,min(nact,ndiis)
        sum1=0.d0
         do j=1,min(nact,ndiis)
          sum1=sum1+invbmat(i,j)
         enddo
        cvec(i)=sum*sum1
       enddo
! New amplitudes
       do i=1,nmax
        pvec(i)=0.d0
       enddo
       do iact=1,min(nact,ndiis)
        sum=cvec(iact)
        read(diisfile,rec=iact+max(0,nact-ndiis)) scr
         do j=1,nmax
          pvec(j)=pvec(j)+sum*scr(j)
         enddo
       enddo
! New gradient
       do i=1,nmax
        gvec(i)=0.d0
       enddo
       do iact=1,min(nact,ndiis)
        sum=cvec(iact)
        read(gdiisfile,rec=iact+max(0,nact-ndiis)) scr
         do j=1,nmax
          gvec(j)=gvec(j)+sum*scr(j)
         enddo
       enddo
       write(*,*)'inside gdiismod'
       call print_vectorR(gvec,9)

      close(diisfile)
      close(errfile)
      close(gdiisfile)
!
      return
      end subroutine gdiismod
!***********************************************************************

!************************************************************************
      subroutine gdiismod2(nact,nmax,pvec,evec,gvec,ndmax,ifltln,bmat,&
                          bvec,invbmat,cvec)
!************************************************************************
!* DIIS extrapolation                                                   *
!* nact - iteration step
!* nmax - dimension of the vector to be extrapolated
!* pvec - the updated vector
!* evec - the original vector
!* ndmax - maximum no. of iterations
!************************************************************************
      implicit none
      integer, intent(in) ::nact,nmax,ndmax,ifltln
      !local
      integer, parameter  :: diisfile=110,errfile=111, gdiisfile=112
      integer :: iact,ii,i,j,iact1,iact2
      integer :: ipiv(ndmax),ndiis
      real*8 bmat(ndmax,ndmax),bvec(ndmax**2),invbmat(ndmax,ndmax),ddot
      real*8 cvec(ndmax),sum,sum1,evec(nmax),scr(nmax),pvec(nmax)
      real*8 gvec(nmax)
!
      ndiis=4
      open(diisfile, file='DIIS.geo', status='unknown',access='direct',recl=nmax*ifltln)
      open(errfile,  file='DIISE.geo',status='unknown',access='direct',recl=nmax*ifltln)
      open(gdiisfile,file='DIISG.geo',status='unknown',access='direct',recl=nmax*ifltln)
! Save amplitudes and calculate new error vector
!        call dfillzero(pvec,nmax)
        write(diisfile,rec=nact) pvec
!        do j=1,nmax
!         pvec(j)=pvec(j)-evec(j)
!        enddo
!       write(errfile,rec=nact) pvec
        write(errfile,rec=nact) evec
        write(gdiisfile,rec=nact) gvec
! Update the scalar product matrix (B matrix)
       do iact1=max(1,nact-ndiis+1),nact
        read(errfile,rec=iact1) pvec
         do iact2=max(1,nact-ndiis+1),iact1
          read(errfile,rec=iact2) scr
          sum=ddot(nmax,pvec,1,scr,1)
          bmat(iact1-max(0,nact-ndiis),iact2-max(0,nact-ndiis))=sum
          bmat(iact2-max(0,nact-ndiis),iact1-max(0,nact-ndiis))=sum
         enddo
       enddo
! Invert B matrix
       do j=1,min(nact,ndiis)
         do i=1,j
          bvec(i+(j-1)*j/2)=bmat(i,j)
         enddo
       enddo
      call dsptrf('U',min(nact,ndiis),bvec,ipiv,invbmat,i)
      call dsptri('U',min(nact,ndiis),bvec,ipiv,invbmat,i)
       do j=1,min(nact,ndiis)
         do i=1,j
          invbmat(i,j)=bvec(i+(j-1)*j/2)
          invbmat(j,i)=invbmat(i,j)
         enddo
       enddo
! Denominator
      sum=0.d0
       do i=1,min(nact,ndiis)
         do j=1,min(nact,ndiis)
          sum=sum+invbmat(i,j)
         enddo
       enddo
      sum=1.d0/sum
! Coefficients
       do i=1,min(nact,ndiis)
        sum1=0.d0
         do j=1,min(nact,ndiis)
          sum1=sum1+invbmat(i,j)
         enddo
        cvec(i)=sum*sum1
       enddo
! New amplitudes
       do i=1,nmax
        pvec(i)=0.d0
       enddo
       do iact=1,min(nact,ndiis)
        sum=cvec(iact)
        read(diisfile,rec=iact+max(0,nact-ndiis)) scr
         do j=1,nmax
          pvec(j)=pvec(j)+sum*scr(j)
         enddo
       enddo

       !csonti print vector differences
       do iact=1,min(nact,ndiis)
        read(diisfile,rec=iact+max(0,nact-ndiis)) scr
        write(*,*)'vec diffs trust', dot_product(pvec-scr,pvec-scr)
       enddo
       !csonti

! New gradient
       do i=1,nmax
        gvec(i)=0.d0
       enddo
       do iact=1,min(nact,ndiis)
        sum=cvec(iact)
        read(gdiisfile,rec=iact+max(0,nact-ndiis)) scr
         do j=1,nmax
          gvec(j)=gvec(j)+sum*scr(j)
         enddo
       enddo
       write(*,*)'inside gdiismod'
       call print_vectorR(gvec,9)

      close(diisfile)
      close(errfile)
      close(gdiisfile)
!
      return
      end subroutine gdiismod2
!***********************************************************************

!************************************************************************
      subroutine gdiis_del(niter,ndim)
!************************************************************************
      implicit none
      integer,intent(in) :: niter,ndim
      !local
      integer, parameter :: ifltln=8 !MRCC common file!

      call gdiisdel(niter,ndim,ifltln)

      end subroutine gdiis_del
!************************************************************************

!************************************************************************
      subroutine gdiis_extrap2(niter,ndmax,xk,gk,ek)
!************************************************************************
      implicit none
      integer,intent(in) :: niter,ndmax
      double precision, dimension(:), intent(inout) :: xk,gk ! geometry and gradient
      double precision, dimension(:), intent(in)    :: ek    ! error vector
      !local
      integer :: nmax
      double precision, dimension(:), allocatable           :: cvec ! vector of coefficients
      double precision, dimension(:,:), allocatable         :: bmat ! scalar product B matrix
      double precision, dimension(:,:), allocatable         :: invbmat ! inverse of the B matrix
      double precision, dimension(:), allocatable           :: bvec ! inverse of the B matrix
      integer, parameter :: ifltln=8 !MRCC common file!

      nmax = size(xk)

      allocate(cvec(ndmax))
      allocate(bmat(ndmax,ndmax))
      allocate(invbmat(ndmax,ndmax))
      allocate(bvec(ndmax*ndmax))

      cvec = 0.d0

!     write(*,*) 'trust before diis ek'
!     call print_vectorR(ek,9)
!     write(*,*) 'trust before diis xk'
!     call print_vectorR(xk,9)
!     write(*,*) 'trust before diis gk'
!     call print_vectorR(gk,9)

      call gdiismod2(niter,nmax,xk,ek,gk,ndmax,ifltln,&
                    bmat,bvec,invbmat,cvec)

!     write(*,*) 'trust after diis ek'
!     call print_vectorR(ek,9)
!     write(*,*) 'trust after diis xk'
!     call print_vectorR(xk,9)
!     write(*,*) 'trust after diis gk'
!     call print_vectorR(gk,9)

      write(*,*) 'trust diis coefficients'
      call print_vectorR(cvec,9)

      deallocate(cvec)
      deallocate(bmat)
      deallocate(invbmat)
      deallocate(bvec)

      end subroutine gdiis_extrap2
!************************************************************************

!************************************************************************
      subroutine gdiis_extrap(niter,ndmax,xk,gk,ek)
!************************************************************************
      implicit none
      integer,intent(in) :: niter,ndmax
      double precision, dimension(:), intent(inout) :: xk,gk ! geometry and gradient
      double precision, dimension(:), intent(in)    :: ek    ! error vector
      !local
      integer :: nmax
      double precision, dimension(:), allocatable           :: cvec ! vector of coefficients
      double precision, dimension(:,:), allocatable         :: bmat ! scalar product B matrix
      double precision, dimension(:,:), allocatable         :: invbmat ! inverse of the B matrix
      double precision, dimension(:), allocatable           :: bvec ! inverse of the B matrix
      integer, parameter :: ifltln=8 !MRCC common file!

      nmax = size(xk)

      allocate(cvec(ndmax))
      allocate(bmat(ndmax,ndmax))
      allocate(invbmat(ndmax,ndmax))
      allocate(bvec(ndmax*ndmax))

      cvec = 0.d0

!     write(*,*) 'trust before diis ek'
!     call print_vectorR(ek,9)
!     write(*,*) 'trust before diis xk'
!     call print_vectorR(xk,9)
!     write(*,*) 'trust before diis gk'
!     call print_vectorR(gk,9)

      call gdiismod(niter,nmax,xk,ek,gk,ndmax,ifltln,&
                    bmat,bvec,invbmat,cvec)

!     write(*,*) 'trust after diis ek'
!     call print_vectorR(ek,9)
!     write(*,*) 'trust after diis xk'
!     call print_vectorR(xk,9)
!     write(*,*) 'trust after diis gk'
!     call print_vectorR(gk,9)

      write(*,*) 'trust diis coefficients'
      call print_vectorR(cvec,9)

      deallocate(cvec)
      deallocate(bmat)
      deallocate(invbmat)
      deallocate(bvec)

      end subroutine gdiis_extrap
!************************************************************************

!************************************************************************
      subroutine gdiis(fegrad,args,Hk,&
                      ftol,gtol,xtol,maxiter,maxfun,finit,&
                      fopt,niter,nfeval,ngeval,plevel,lmolden,bool)
!************************************************************************
! The Broyden-Fletcher-Goldfarb-Shanno algorithm;
! written by J Csontos
!
! Hk is the inverse of the Hessian, see Nocedal and Wright, equation 8.16

      implicit none 
      double precision, intent(in)  :: ftol,gtol,xtol
      double precision, intent(out) :: finit
      double precision, intent(out) :: fopt
      integer, intent(in) :: maxfun,maxiter,plevel
      integer, intent(out) :: niter,nfeval,ngeval
      double precision, dimension(:), intent(inout) :: args
      logical, intent(in)  :: lmolden ! Is molden file required?
      logical, intent(out) :: bool 

      double precision, parameter :: Delta_0=0.5d0
      double precision, parameter :: eta=1.0d-3
!     double precision, parameter :: r=1.0d-2
!     double precision, parameter :: etol=1.0d-6

!     local variables
      integer :: i,j,ndim,nfseval,ngseval
      double precision :: gnorm,snorm,gmax,smax,rtmp,alpha,thres
      double precision :: pred,ared,q,Delta_k,a,b
      double precision :: fval,fval_prev,functol,argtol
      double precision, dimension(size(args)) :: gfk,gfk_prev,xk,xk_prev
      double precision, dimension(size(args)) :: sk,ska,ek,yk,zk,Bs
      double precision, dimension(size(args)+1) :: results
      double precision, dimension(size(args),size(args)) :: Hk,Pmat,Pinv,One

      double precision, dimension(size(args)) :: gval
      logical :: notConverged,lBaker
      double precision :: alpha_k,sBs,ss
      character(len=512) :: si,scommand
      double precision,dimension(3,3) :: Atest,Ctest

      interface
        function fegrad(z)
          double precision, dimension(:), intent(inout) :: z
          double precision, dimension(size(z)+1) :: fegrad
        end function
      end interface

!     Atest = reshape((/1,2,4,0,-1,1,2,3,8/), shape(Atest))
!     call matrix_inverse(Atest,Ctest)
!     Atest = matmul(Atest,Ctest)

!     write(*,*) ' matrix test i/o:'
!     call print_matrixR(Ctest,3)
!     write(*,*) ' matrix test i/o:'
!     call print_matrixR(Atest,3)
!     call mrccend(1)

! ***
! f2min   ->  external function to be minimized
! args()  ->  independent variables of f2min;
!         ->  as an input:  initial guess
!         ->  as an output: coordinates of the minimum
! grad    ->  gradient of f2min
!
! gtol    ->  convergence criterion (largest element of the gradient vector)
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! finit   ->  initial value of f2min
! fopt    ->  minimal value of f2min
! niter   ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! ngeval  ->  number of gradient evaluations performed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! ***

      ndim = size(args) ! dimension of the grad vector (fegrad(1:ndim)); the energy is stored in fegrad(ndim+1)
      nfeval = 0
      ngeval = 0
      niter  = 0
      bool = .False.
      lBaker = .True. ! Baker's convergence criteria are set

!     call print_matrixR(Hk,9)
      call print_inverse(Hk)

      results = fegrad(args)
      finit   = results(ndim+1) ! initial fucntion value
      ngeval  = ngeval+1
      nfeval  = nfeval+1

      do i=1,ndim
        gfk(i) = results(i)
      enddo

      fval      = finit              ! current function value
      fval_prev = fval + 5000.0d0       ! previous function value

      xk    = args
      gnorm = vecnorm(gfk,2)    ! Euclidean
      gmax  = vecnorm(gfk,1000) ! infinit

      call wrt_iter(niter)  ! # interations is written in the file ITER
      if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info
      
      Delta_k = Delta_0
      notConverged = .True.
      if( gnorm <= gtol) notConverged = .False.
      if(.not. notConverged) then
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
           'Iteration info ',niter,fval,gnorm,gmax,0.d0,0.d0,0.d0
        endif
      endif

      do while(notConverged .and. niter < maxiter)! csonti introduce ftol and xtol here

        niter = niter+1


        ek = -1.d0*matrix_dot_vec(Hk,gfk)
        write(*,*)'Iteration info gfk*ek',&
                    dot_product(gfk,ek),vecnorm(ek,2)
        write(*,*)'vector ek:'
        call print_vectorR(ek,9)
!       alpha = direction_alpha(ek,gfk)

        gfk_prev=gfk
        xk_prev=xk 

        call gdiis_extrap(niter,20,xk,gfk,ek) ! xk, gfk are updated

        call read_Pmat(ndim,Pmat)
        write(*,*) 'Pmat*Pinv i/o:'
        gfk = matrix_dot_vec(Pmat,gfk)

        sk = - 1.0d0*matrix_dot_vec(Hk,gfk)
        xk = xk + sk

        call quadratic(sk,gfk_prev,Hk,pred,Bs) !csonti test

        results    = fegrad(xk)
        fval_prev  = fval
        fval       = results(ndim+1)

        nfeval  = nfeval + 1
        ngeval  = ngeval + 1

        do i=1,ndim
          gfk(i)= results(i)
          yk(i) = gfk(i) - gfk_prev(i)
        enddo

        sk = xk - xk_prev

        ared  = fval_prev - fval  ! actual reduction in function value

        gnorm = vecnorm(gfk,2)    ! Euclidean
        gmax  = vecnorm(gfk,1000) ! infinit
        snorm = vecnorm(sk,2)
        smax  = vecnorm(sk,1000)

        if(lBaker) then !Baker conv criteria; JCC, 14, 1085 (1993)
          if( gmax <= 3.d-4 .and. (ared<=1.d-6 .or. smax<=3.d-4) ) then 
            notConverged = .False.
          endif
        else
          if( gnorm <= gtol) notConverged = .False.
        endif

        if(.not. notConverged) then
          if(plevel.gt.1) then 
            write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
          endif
          exit
        endif


        call wrt_iter(niter)
        if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info

        if(niter.eq.1) then 
          write(*,'(5x,a)') &         
             'All quantities are in atomic units.'
          write(*,'(5x,a,a5,6(a22))') &         
             'Iteration info ','step','energy','rms force','max force',&
             'rms step','max step','energy change'
        endif
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
        endif

!       call bfgs_update_damped(sk,yk,Hk,q,Bs)
        sk = matrix_dot_vec(Pmat,sk)
        call bfgs_update(sk,yk,Hk)

!       write(*,*) 'Projection matrix after i/o:'
!       call print_matrixR(Pmat,10)
!       write(*,*) 'Projection matrix inverse after i/o:'
!       call print_matrixR(Pinv,10)
!       One = matmul(Hk,Pinv)
!       write(*,*) 'Pmat*Pinv i/o:'
!       call print_matrixR(One,10)

!       Hk=matmul(Hk,Pmat)
!       Hk=matmul(Pmat,Hk)

        call matrix_pseudoinverse(Pmat,Pinv)
        Hk=matmul(Hk,Pinv)
        Hk=matmul(Pinv,Hk)
!       call matrix_inverse(Pmat,Hk)

!       call print_inverse(Hk)

      enddo

      fopt = fval
      args = xk

      if(.not. notConverged .and. &
         niter .lt. maxiter) then
         bool=.True.
         if(lmolden) then
           call ishell('cat COORD.xyz.energies >> MOLDEN')
           call ishell('cat COORD.xyz.rforce >> MOLDEN')
           call ishell('cat COORD.xyz.mforce >> MOLDEN')
           call ishell('cat COORD.xyz.geometries >> MOLDEN')
           call ishell('rm -f COORD.xyz.* COORD.xyz.energies')
           call ishell('rm -f COORD.xyz.rforce COORD.xyz.mforce')
           call ishell('rm -f COORD.xyz.geometries')
         endif
      endif


      end subroutine gdiis
!************************************************************************


!     old_fval = results(ndim+1)
!     old_old_fval = old_fval + 5000.d0


!       if(vecnorm(pk,2)>Delta_k ) then
!         pk = (Delta_k/vecnorm(pk,2))*pk
!       endif

!       call line_search(fegrad,xk,pk,gfk,old_fval,old_old_fval,&
!                        alpha_k,nfseval,ngseval,fval,gval,lmessage)
!       write(*,*) 'csonti alpha:', alpha_k
!       nfeval = nfeval + nfseval
!       ngeval = ngeval + ngseval
!       old_old_fval= old_fval
!       old_fval = fval

!       sk = alpha_k*pk
!       xk = xk + sk
!       
!       call quadratic(sk,gfk,Hk,pred,Bs) !csonti test
!       write(*,*) 'trust gdiis predicted, s-length', pred,vecnorm(sk,2)

!       yk = gval - gfk
!       gfk = gval


!       ared  = old_fval - old_old_fval  ! actual reduction in function value

!       call bfgs_update_damped(sk,yk,Hk,q,Bs)


!************************************************************************
      subroutine gdiis2(fegrad,args,Hk,&
                      ftol,gtol,xtol,maxiter,maxfun,finit,&
                      fopt,niter,nfeval,ngeval,plevel,lmolden,bool)
!************************************************************************
! The Broyden-Fletcher-Goldfarb-Shanno algorithm;
! written by J Csontos
!
! Hk is the inverse of the Hessian, see Nocedal and Wright, equation 8.16

      implicit none 
      double precision, intent(in)  :: ftol,gtol,xtol
      double precision, intent(out) :: finit
      double precision, intent(out) :: fopt
      integer, intent(in) :: maxfun,maxiter,plevel
      integer, intent(out) :: niter,nfeval,ngeval
      double precision, dimension(:), intent(inout) :: args
      logical, intent(in)  :: lmolden ! Is molden file required?
      logical, intent(out) :: bool 

      double precision, parameter :: Delta_0=0.5d0
      double precision, parameter :: eta=1.0d-3
!     double precision, parameter :: r=1.0d-2
!     double precision, parameter :: etol=1.0d-6

!     local variables
      integer :: i,j,ndim,nfseval,ngseval,nditer
      double precision :: gnorm,snorm,gmax,smax,rtmp,alpha,thres
      double precision :: pred,ared,q,Delta_k,a,b
      double precision :: fval,fval_prev,functol,argtol
      double precision, dimension(size(args)) :: gfk,gfk_prev,xk,xk_prev
      double precision, dimension(size(args)) :: sk,ek,yk,zk,Bs
      double precision, dimension(size(args)+1) :: results
      double precision, dimension(size(args),size(args)) :: Hk

      double precision, dimension(size(args)) :: gval
      logical :: notConverged,lBaker
      double precision :: alpha_k,sBs,ss
      character(len=512) :: si,scommand

      interface
        function fegrad(z)
          double precision, dimension(:), intent(inout) :: z
          double precision, dimension(size(z)+1) :: fegrad
        end function
      end interface

! ***
! f2min   ->  external function to be minimized
! args()  ->  independent variables of f2min;
!         ->  as an input:  initial guess
!         ->  as an output: coordinates of the minimum
! grad    ->  gradient of f2min
!
! gtol    ->  convergence criterion (largest element of the gradient vector)
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! finit   ->  initial value of f2min
! fopt    ->  minimal value of f2min
! niter   ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! ngeval  ->  number of gradient evaluations performed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! ***

      ndim = size(args) ! dimension of the grad vector (fegrad(1:ndim)); the energy is stored in fegrad(ndim+1)
      nfeval = 0
      ngeval = 0
      niter  = 0
      nditer = 0
      bool = .False.
      lBaker = .True. ! Baker's convergence criteria are set

!     call print_matrixR(Hk,9)

      results = fegrad(args)
      finit   = results(ndim+1) ! initial fucntion value
      ngeval  = ngeval+1
      nfeval  = nfeval+1

      do i=1,ndim
        gfk(i) = results(i)
      enddo

      fval      = finit              ! current function value
      fval_prev = fval + 5000.0d0       ! previous function value

      xk    = args
      gnorm = vecnorm(gfk,2)    ! Euclidean
      gmax  = vecnorm(gfk,1000) ! infinit

      call wrt_iter(niter)  ! # interations is written in the file ITER
      if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info
      
      Delta_k = Delta_0
      notConverged = .True.
      if( gnorm <= gtol) notConverged = .False.
      if(.not. notConverged) then
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
           'Iteration info ',niter,fval,gnorm,gmax,0.d0,0.d0,0.d0
        endif
      endif

      do while(notConverged .and. niter < maxiter)! csonti introduce ftol and xtol here

        niter = niter+1
        nditer = nditer+1

        ek = -1.d0*matrix_dot_vec(Hk,gfk)
!       call dogleg(Delta_k,ek,gfk,Hk,sk,pred,Bs)  ! paramater pred is equal to the predicted reduction in the function value
!       write(*,*)'Iteration info gfk*ek',  dot_product(gfk,ek)
!       alpha = direction_alpha(ek,gfk)

        gfk_prev=gfk
        xk_prev=xk 

        if(niter>0) call gdiis_extrap2(nditer,20,xk,gfk,ek) ! xk, gfk are updated

        sk = - 1.0d0*matrix_dot_vec(Hk,gfk)
!       if(vecnorm(sk,2)>Delta_k ) then
!         sk = (Delta_k/vecnorm(ek,2))*sk
!       endif
!       write(*,*) 'trust gdiis predicted, s-length', pred,vecnorm(sk,2)
        xk = xk + sk
        sk = xk - xk_prev
!       call quadratic(sk,gfk_prev,Hk,pred,Bs) !csonti test
        write(*,*) 'trust gdiis predicted, s-length', pred,vecnorm(sk,2)

!       if(vecnorm(xk-xk_prev,2)>Delta_k ) then
!         sk = (Delta_k/vecnorm(xk-xk_prev,2))*(xk-xk_prev)
!         xk = xk_prev + sk
!       endif
        results    = fegrad(xk)
        fval_prev  = fval
        fval       = results(ndim+1)

        nfeval  = nfeval + 1
        ngeval  = ngeval + 1

        do i=1,ndim
          gfk(i)= results(i)
          yk(i) = gfk(i) - gfk_prev(i)
        enddo


        ared  = fval_prev - fval  ! actual reduction in function value

!       if(ared>0) then
!         nditer = nditer + 1
!         call gdiis_extrap(nditer,20,sk,yk,ek)
!       endif

        gnorm = vecnorm(gfk,2)    ! Euclidean
        gmax  = vecnorm(gfk,1000) ! infinit
        snorm = vecnorm(sk,2)
        smax  = vecnorm(sk,1000)

        if(lBaker) then !Baker conv criteria; JCC, 14, 1085 (1993)
          if( gmax <= 3.d-4 .and. (ared<=1.d-6 .or. smax<=3.d-4) ) then 
            notConverged = .False.
          endif
        else
          if( gnorm <= gtol) notConverged = .False.
        endif

        if(.not. notConverged) then
          if(plevel.gt.1) then 
            write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
          endif
          exit
        endif


        call wrt_iter(niter)
        if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info

        if(niter.eq.1) then 
          write(*,'(5x,a)') &         
             'All quantities are in atomic units.'
          write(*,'(5x,a,a5,6(a22))') &         
             'Iteration info ','step','energy','rms force','max force',&
             'rms step','max step','energy change'
        endif
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
        endif

!       call quadratic(sk,gfk_prev,Hk,pred,Bs)
!       q = ared/pred
!       if(q .le. eta) then
!         xk  = xk_prev
!         gfk = gfk_prev
!         fval = fval_prev
!         call gdiis_del(nditer,size(xk)) ! last step deleted
!         nditer = nditer-1
!       call trustR_update(ared,pred,Delta_k,snorm)
!         if(q<0) then
!           call dogleg(Delta_k,ek,gfk,Hk,sk,pred,Bs)  ! paramater pred is equal to the predicted reduction in the function value
!           xk = xk + ek
!           results    = fegrad(xk)
!           fval_prev  = fval
!           fval       = results(ndim+1)
!           write(*,*) 'trust dogleg step energy', fval

!           nfeval  = nfeval + 1
!           ngeval  = ngeval + 1

!           do i=1,ndim
!             gfk(i)= results(i)
!           enddo
!           ared  = fval_prev - fval  ! actual reduction in function value
!           snorm = vecnorm(sk,2)
!           call trustR_update(ared,pred,Delta_k,vecnorm(ek,2))
!           call bfgs_update(sk,yk,Hk)
!         endif
!       else
!         call trustR_update(ared,pred,Delta_k,snorm)
          if(vecnorm(sk,2)<Delta_k) call bfgs_update(sk,yk,Hk)
!       endif


!       write(*,*) 'trust skyk',dot_product(sk,yk),&
!        1e-2*dot_product(sk,Bs),ared,pred,ared/pred,nditer
!       if(ared > 0) then
!         call bfgs_update(sk,yk,Hk)
!       endif

      enddo

      fopt = fval
      args = xk

      if(.not. notConverged .and. &
         niter .lt. maxiter) then
         bool=.True.
         if(lmolden) then
           call ishell('cat COORD.xyz.energies >> MOLDEN')
           call ishell('cat COORD.xyz.rforce >> MOLDEN')
           call ishell('cat COORD.xyz.mforce >> MOLDEN')
           call ishell('cat COORD.xyz.geometries >> MOLDEN')
           call ishell('rm -f COORD.xyz.* COORD.xyz.energies')
           call ishell('rm -f COORD.xyz.rforce COORD.xyz.mforce')
           call ishell('rm -f COORD.xyz.geometries')
         endif
      endif


      end subroutine gdiis2
!************************************************************************

!************************************************************************
      subroutine gdiis_o(fegrad,args,Hk,&
                      ftol,gtol,xtol,maxiter,maxfun,finit,&
                      fopt,niter,nfeval,ngeval,plevel,lmolden,bool)
!************************************************************************
! The Broyden-Fletcher-Goldfarb-Shanno algorithm;
! written by J Csontos
!
! Hk is the inverse of the Hessian, see Nocedal and Wright, equation 8.16

      implicit none 
      double precision, intent(in)  :: ftol,gtol,xtol
      double precision, intent(out) :: finit
      double precision, intent(out) :: fopt
      integer, intent(in) :: maxfun,maxiter,plevel
      integer, intent(out) :: niter,nfeval,ngeval
      double precision, dimension(:), intent(inout) :: args
      logical, intent(in)  :: lmolden ! Is molden file required?
      logical, intent(out) :: bool 

      double precision, parameter :: Delta_0=0.5d0
      double precision, parameter :: eta=1.0d-3
!     double precision, parameter :: r=1.0d-2
!     double precision, parameter :: etol=1.0d-6

!     local variables
      integer :: i,j,ndim,nfseval,ngseval
      double precision :: gnorm,snorm,gmax,smax,rtmp,alpha,thres
      double precision :: pred,ared,q,Delta_k,a,b
      double precision :: fval,fval_prev,functol,argtol
      double precision, dimension(size(args)) :: gfk,gfk_prev,xk,xk_prev
      double precision, dimension(size(args)) :: sk,ska,ek,yk,zk,Bs
      double precision, dimension(size(args)+1) :: results
      double precision, dimension(size(args),size(args)) :: Hk

      double precision, dimension(size(args)) :: gval
      logical :: notConverged,lBaker
      double precision :: alpha_k,sBs,ss
      character(len=512) :: si,scommand

      interface
        function fegrad(z)
          double precision, dimension(:), intent(inout) :: z
          double precision, dimension(size(z)+1) :: fegrad
        end function
      end interface

! ***
! f2min   ->  external function to be minimized
! args()  ->  independent variables of f2min;
!         ->  as an input:  initial guess
!         ->  as an output: coordinates of the minimum
! grad    ->  gradient of f2min
!
! gtol    ->  convergence criterion (largest element of the gradient vector)
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! finit   ->  initial value of f2min
! fopt    ->  minimal value of f2min
! niter   ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! ngeval  ->  number of gradient evaluations performed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! ***

      ndim = size(args) ! dimension of the grad vector (fegrad(1:ndim)); the energy is stored in fegrad(ndim+1)
      nfeval = 0
      ngeval = 0
      niter  = 0
      bool = .False.
      lBaker = .True. ! Baker's convergence criteria are set

!     call print_matrixR(Hk,9)

      results = fegrad(args)
      finit   = results(ndim+1) ! initial fucntion value
      ngeval  = ngeval+1
      nfeval  = nfeval+1

      do i=1,ndim
        gfk(i) = results(i)
      enddo

      fval      = finit              ! current function value
      fval_prev = fval + 5000.0d0       ! previous function value

      xk    = args
      gnorm = vecnorm(gfk,2)    ! Euclidean
      gmax  = vecnorm(gfk,1000) ! infinit

      call wrt_iter(niter)  ! # interations is written in the file ITER
      if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info
      
      Delta_k = Delta_0
      notConverged = .True.
      if( gnorm <= gtol) notConverged = .False.
      if(.not. notConverged) then
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
           'Iteration info ',niter,fval,gnorm,gmax,0.d0,0.d0,0.d0
        endif
      endif

      do while(notConverged .and. niter < maxiter)! csonti introduce ftol and xtol here

        niter = niter+1

!       call print_inverse(Hk)

        ek = -1.d0*matrix_dot_vec(Hk,gfk)
        write(*,*)'Iteration info gfk*ek',&
                    dot_product(gfk,ek),vecnorm(ek,2)
!       alpha = direction_alpha(ek,gfk)

        gfk_prev=gfk
        xk_prev=xk 

        call gdiis_extrap(niter,20,xk,gfk,ek) ! xk, gfk are updated

        sk = - 1.0d0*matrix_dot_vec(Hk,gfk)
        xk = xk + sk

        results    = fegrad(xk)
        fval_prev  = fval
        fval       = results(ndim+1)

        nfeval  = nfeval + 1
        ngeval  = ngeval + 1

        do i=1,ndim
          gfk(i)= results(i)
          yk(i) = gfk(i) - gfk_prev(i)
        enddo

        sk = xk - xk_prev

        ared  = fval_prev - fval  ! actual reduction in function value

        gnorm = vecnorm(gfk,2)    ! Euclidean
        gmax  = vecnorm(gfk,1000) ! infinit
        snorm = vecnorm(sk,2)
        smax  = vecnorm(sk,1000)

        if(lBaker) then !Baker conv criteria; JCC, 14, 1085 (1993)
          if( gmax <= 3.d-4 .and. (ared<=1.d-6 .or. smax<=3.d-4) ) then 
            notConverged = .False.
          endif
        else
          if( gnorm <= gtol) notConverged = .False.
        endif

        if(.not. notConverged) then
          if(plevel.gt.1) then 
            write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
          endif
          exit
        endif


        call wrt_iter(niter)
        if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info

        if(niter.eq.1) then 
          write(*,'(5x,a)') &         
             'All quantities are in atomic units.'
          write(*,'(5x,a,a5,6(a22))') &         
             'Iteration info ','step','energy','rms force','max force',&
             'rms step','max step','energy change'
        endif
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
        endif

        call bfgs_update(sk,yk,Hk)

      enddo

      fopt = fval
      args = xk

      if(.not. notConverged .and. &
         niter .lt. maxiter) then
         bool=.True.
         if(lmolden) then
           call ishell('cat COORD.xyz.energies >> MOLDEN')
           call ishell('cat COORD.xyz.rforce >> MOLDEN')
           call ishell('cat COORD.xyz.mforce >> MOLDEN')
           call ishell('cat COORD.xyz.geometries >> MOLDEN')
           call ishell('rm -f COORD.xyz.* COORD.xyz.energies')
           call ishell('rm -f COORD.xyz.rforce COORD.xyz.mforce')
           call ishell('rm -f COORD.xyz.geometries')
         endif
      endif


      end subroutine gdiis_o
!************************************************************************

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! BFGS ALGORITHM
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      subroutine bfgs_B(fegrad,args,Bk,&
                      ftol,gtol,xtol,maxiter,maxfun,finit,&
                      fopt,niter,nfeval,ngeval,plevel,lmolden,bool)
!************************************************************************
! The Broyden-Fletcher-Goldfarb-Shanno algorithm;
! written by J Csontos
!
! Bk is the Hessian, see Nocedal and Wright, equation 8.19
! Hessian update is utilized 
! (it is not efficient at the moment, it should use Cholesky factors)

      implicit none 
      double precision, intent(in)  :: ftol,gtol,xtol
      double precision, intent(out) :: finit
      double precision, intent(out) :: fopt
      integer, intent(in) :: maxfun,maxiter,plevel
      integer, intent(out) :: niter,nfeval,ngeval
      double precision, dimension(:), intent(inout) :: args
      logical, intent(in)  :: lmolden ! Is molden file required?
      logical, intent(out) :: bool 

      double precision, parameter :: Delta_0=0.5d0
      double precision, parameter :: eta=1.0d-3
!     double precision, parameter :: r=1.0d-2
!     double precision, parameter :: etol=1.0d-6

!     local variables
      integer :: i,j,ndim,nfseval,ngseval
      double precision :: gnorm,snorm,gmax,smax,rtmp,alpha,thres
      double precision :: pred,ared,q,Delta_k,a,b
      double precision :: fval,fval_prev,functol,argtol
      double precision, dimension(size(args)) :: gfk,gfk_prev,xk,xkp1,stp_diag
      double precision, dimension(size(args)) :: sk,pk,yk,zk,Bs,eval
      double precision, dimension(size(args)+1) :: results
      double precision, dimension(size(args),size(args)) :: Hk,evec,evec2
      double precision, dimension(size(args),size(args)) :: Bk,Pmat,Hinv,One

      double precision, dimension(size(args)) :: gval
      logical :: lmessage,notConverged,lBaker
      double precision :: alpha_k,sBs,ss
      character(len=512) :: si,scommand
      double precision,dimension(3,3) :: Atest,Ctest

      interface
        function fegrad(z)
          double precision, dimension(:), intent(inout) :: z
          double precision, dimension(size(z)+1) :: fegrad
        end function
      end interface

! ***
! f2min   ->  external function to be minimized
! args()  ->  independent variables of f2min;
!         ->  as an input:  initial guess
!         ->  as an output: coordinates of the minimum
! grad    ->  gradient of f2min
!
! gtol    ->  convergence criterion (largest element of the gradient vector)
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! finit   ->  initial value of f2min
! fopt    ->  minimal value of f2min
! niter   ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! ngeval  ->  number of gradient evaluations performed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! ***

      ndim = size(args) ! dimension of the grad vector (fegrad(1:ndim)); the energy is stored in fegrad(ndim+1)
      nfeval = 0
      ngeval = 0
      niter  = 0
      bool = .False.
      lBaker = .False. !.True. ! Baker's convergence criteria are set

!     call print_matrixR(Hk,9)
!     Hk = Hinv

!!!!!!!!!!!!!!!!!!!!!!!!!! TEST
!      write(*,*) 'Bk:'
!      call print_matrixR(Bk,10)
!
!!     call matrix_inverse_sym(Bk,Hk)
!      call matrix_diag(Bk,evec,eval)
!
!      write(*,*) 'evec:'
!      call print_matrixR(evec,10)
!
!      write(*,*) 'evals:'
!      write(*,*) eval

!!!!!!!!!!!!!!!!!!!!!!!!!! TEST

      results = fegrad(args)
      finit   = results(ndim+1) ! initial function value
      ngeval  = ngeval+1
      nfeval  = nfeval+1

      do i=1,ndim
        gfk(i) = results(i)
      enddo
!      evec2=transpose(evec)
!      gfk_prev=matrix_dot_vec(evec2,gfk)
!      write(*,*) 'gfk_prev:'
!      write(*,*) gfk_prev
!      write(*,*) 'evec2:'
!      call print_matrixR(evec2,10)
!
!      One=matmul(evec2,matmul(Bk,evec))
!
!
!      do i=1,ndim
!        if(dabs(One(i,i)) .gt. 1d-6) then
!          stp_diag(i) = -gfk_prev(i)/One(i,i)
!        else
!          stp_diag(i) = 0.d0
!        endif
!      enddo
!      write(*,*) 'stp_diag'
!      write(*,*) stp_diag
!
!      write(*,*) 'stp_origin'
!      write(*,*) matmul(evec,stp_diag)
!
!      write(*,*) 'internal coords'
!      write(*,*) args
!      call mrccend(1)
!
!
!      call matrix_diag(Bk,evec,eval)
!      evec2=transpose(evec)
!      gfk_prev=matrix_dot_vec(evec2,gfk)
!      One=matmul(evec2,matmul(Bk,evec))
!      do i=1,ndim
!        if(dabs(One(i,i)) .gt. 1d-6) then
!          stp_diag(i) = -gfk_prev(i)/One(i,i)
!        else
!          stp_diag(i) = 0.d0
!        endif
!      enddo
!      write(*,*) matmul(evec,stp_diag)



      fval      = finit              ! current function value
      fval_prev = fval + 5000.0d0       ! previous function value

      xk    = args
      gnorm = vecnorm(gfk,2)    ! Euclidean
      gmax  = vecnorm(gfk,1000) ! infinit

      call wrt_iter(niter)  ! # interations is written in the file ITER
      if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info
      
      Delta_k = Delta_0
      notConverged = .True.
      if( gnorm <= gtol) notConverged = .False.
      if(.not. notConverged) then
        if(plevel.gt.1) then 
          call print_iter_info(niter,fval,gnorm,gmax,snorm,smax,ared)
        endif
      endif

      do while(notConverged .and. niter < maxiter)! csonti introduce ftol and xtol here
        call matrix_pseudoinverse(Bk,Hk)
!       call matrix_inverse_sym(Bk,Hk)
!       call matrix_inverse(Bk,Hk)
        pk = -1.d0*matrix_dot_vec(Hk,gfk)
!      call matrix_diag(Bk,evec,eval)
!      evec2=transpose(evec)
!      gfk_prev=matrix_dot_vec(evec2,gfk)
!      One=matmul(evec2,matmul(Bk,evec))
!      do i=1,ndim
!        if(dabs(One(i,i)) .gt. 1d-6) then
!          stp_diag(i) = -gfk_prev(i)/One(i,i)
!        else
!          stp_diag(i) = 0.d0
!        endif
!      enddo
!      write(*,*) matmul(evec,stp_diag)
!      write(*,*) pk 
!      pk = matmul(evec,stp_diag)
!      gfk=gfk_prev
!      call mrccend(1)

!       write(*,*)'Iteration info gfk*pk',  dot_product(gfk,pk)
        call dogleg(Delta_k,pk,gfk,Hk,sk,pred,Bs)  ! paramater pred is equal to the predicted reduction in the function value
!       call ddogleg(Delta_k,pk,gfk,Hk,sk,pred,Bs) !csonti test
        alpha = direction_alpha(sk,gfk)


        niter = niter+1

        xk = xk + sk

        results    = fegrad(xk)
        fval_prev  = fval
        fval       = results(ndim+1)

        nfeval  = nfeval + 1
        ngeval  = ngeval + 1

        do i=1,ndim
          yk(i) = results(i) - gfk(i)
          gfk_prev(i) = gfk(i)
          gfk(i)= results(i)
        enddo

        ared  = fval_prev - fval  ! actual reduction in function value

        gnorm = vecnorm(gfk,2)    ! Euclidean
        gmax  = vecnorm(gfk,1000) ! infinit
        snorm = vecnorm(sk,2)
        smax  = vecnorm(sk,1000)

        if(lBaker) then !Baker conv criteria; JCC, 14, 1085 (1993)
          if( gmax <= 3.d-4 .and. (dabs(ared)<=1.d-6 .or. smax<=3.d-4) ) then 
            notConverged = .False.
          endif
        else
          if( gmax <= gtol .and. (dabs(ared)<=ftol .or. smax <= xtol) ) then
            notConverged = .False.
          endif
        endif

        if(.not. notConverged) then
          if(plevel.gt.1) then 
            call print_iter_info(niter,fval,gnorm,gmax,snorm,smax,ared)
          endif
          exit
        endif


        call wrt_iter(niter)
        if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info

        if(plevel.gt.1) then 
          call print_iter_info(niter,fval,gnorm,gmax,snorm,smax,ared)
        endif

        q = ared/pred
!        if(q .le. eta) then
!          xk  = xk - sk
!          gfk = gfk_prev
!          fval = fval_prev
!        endif

        call trustR_update(ared,pred,Delta_k,snorm)

        call bfgs_update_damped_B(sk,yk,Bk,q,Bs)
!       call bfgs_update_B(sk,yk,Bk)
!     write(*,*) 'Pmat*Pinv i/o:'
      call read_Pmat(ndim,Pmat)
!     call print_matrixR(Pmat,10)
      Bk = matmul(Bk,Pmat)
      Bk = matmul(Pmat,Bk)

      enddo

      fopt = fval
      args = xk

      if(niter .ge. maxiter) then
        write(*,*)'ERROR:  Maximum number of iterations exceeded.'
      endif

      if(.not. notConverged .and. &
         niter .lt. maxiter) then
         bool=.True.
         if(lmolden) then
           call ishell('cat COORD.xyz.energies >> MOLDEN')
           call ishell('cat COORD.xyz.rforce >> MOLDEN')
           call ishell('cat COORD.xyz.mforce >> MOLDEN')
           call ishell('cat COORD.xyz.geometries >> MOLDEN')
           call ishell('rm -f COORD.xyz.* COORD.xyz.energies')
           call ishell('rm -f COORD.xyz.rforce COORD.xyz.mforce')
           call ishell('rm -f COORD.xyz.geometries')
         endif
      endif


      end subroutine bfgs_B
!************************************************************************


!************************************************************************
      subroutine print_iter_info(niter,fval,gnorm,gmax,snorm,smax,ared)
!************************************************************************
!     it prints information about the current geometry optimization step
!************************************************************************
      integer,intent(in) :: niter
      double precision,intent(in) :: fval,gnorm,gmax,snorm,smax,ared
          write(*,*)
          write(*,1003)
          write(*,'(15x,a,I10)') 'GEOMETRY OPTIMIZATION STEP', niter

          write(*,*)
          write(*,'(5x,(a22),f20.12)') 'energy',fval
          write(*,'(5x,(a22),f20.12)') 'rms force',gnorm
          write(*,'(5x,(a22),f20.12)') 'max force',gmax
          write(*,'(5x,(a22),f20.12)') 'rms step',snorm
          write(*,'(5x,(a22),f20.12)') 'max step',smax
          write(*,'(5x,(a22),f20.12)') 'energy change',ared
          
          write(*,*)
          write(*,'(14x,a)') &         
             'All quantities are in atomic units'
          write(*,1003)

1003  format(1x,70('@'))

      end subroutine print_iter_info 
!************************************************************************

!************************************************************************
      subroutine bfgs(fegrad,args,Hk,&
                      ftol,gtol,xtol,maxiter,maxfun,finit,&
                      fopt,niter,nfeval,ngeval,plevel,lmolden,bool)
!************************************************************************
! The Broyden-Fletcher-Goldfarb-Shanno algorithm;
! written by J Csontos
!
! Hk is the inverse of the Hessian, see Nocedal and Wright, equation 8.16
! inverse Hessian update is utilized

      implicit none 
      double precision, intent(in)  :: ftol,gtol,xtol
      double precision, intent(out) :: finit
      double precision, intent(out) :: fopt
      integer, intent(in) :: maxfun,maxiter,plevel
      integer, intent(out) :: niter,nfeval,ngeval
      double precision, dimension(:), intent(inout) :: args
      logical, intent(in)  :: lmolden ! Is molden file required?
      logical, intent(out) :: bool 

      double precision, parameter :: Delta_0=0.5d0
      double precision, parameter :: eta=1.0d-3
!     double precision, parameter :: r=1.0d-2
!     double precision, parameter :: etol=1.0d-6

!     local variables
      integer :: i,j,ndim,nfseval,ngseval
      double precision :: gnorm,snorm,gmax,smax,rtmp,alpha,thres
      double precision :: pred,ared,q,Delta_k,a,b
      double precision :: fval,fval_prev,functol,argtol
      double precision, dimension(size(args)) :: gfk,gfk_prev,xk,xkp1
      double precision, dimension(size(args)) :: sk,pk,yk,zk,Bs
      double precision, dimension(size(args)+1) :: results
!     double precision, dimension(size(args),size(args)) :: Hk
      double precision, dimension(size(args),size(args)) :: Hk,Pmat,Hinv,One

      double precision, dimension(size(args)) :: gval
      logical :: lmessage,notConverged,lBaker
      double precision :: alpha_k,sBs,ss
      character(len=512) :: si,scommand
      double precision,dimension(3,3) :: Atest,Ctest

      interface
        function fegrad(z)
          double precision, dimension(:), intent(inout) :: z
          double precision, dimension(size(z)+1) :: fegrad
        end function
      end interface

! ***
! f2min   ->  external function to be minimized
! args()  ->  independent variables of f2min;
!         ->  as an input:  initial guess
!         ->  as an output: coordinates of the minimum
! grad    ->  gradient of f2min
!
! gtol    ->  convergence criterion (largest element of the gradient vector)
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! finit   ->  initial value of f2min
! fopt    ->  minimal value of f2min
! niter   ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! ngeval  ->  number of gradient evaluations performed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! ***

      ndim = size(args) ! dimension of the grad vector (fegrad(1:ndim)); the energy is stored in fegrad(ndim+1)
      nfeval = 0
      ngeval = 0
      niter  = 0
      bool = .False.
      lBaker = .True. ! Baker's convergence criteria are set

!     call print_matrixR(Hk,9)
      call matrix_pseudoinverse(Hk,Hinv)
      Hk = Hinv

      results = fegrad(args)
      finit   = results(ndim+1) ! initial fucntion value
      ngeval  = ngeval+1
      nfeval  = nfeval+1

      do i=1,ndim
        gfk(i) = results(i)
      enddo

      fval      = finit              ! current function value
      fval_prev = fval + 5000.0d0       ! previous function value

      xk    = args
      gnorm = vecnorm(gfk,2)    ! Euclidean
      gmax  = vecnorm(gfk,1000) ! infinit

      call wrt_iter(niter)  ! # interations is written in the file ITER
      if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info
      
      Delta_k = Delta_0
      notConverged = .True.
      if( gnorm <= gtol) notConverged = .False.
      if(.not. notConverged) then
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
           'Iteration info ',niter,fval,gnorm,gmax,0.d0,0.d0,0.d0
        endif
      endif

      do while(notConverged .and. niter < maxiter)! csonti introduce ftol and xtol here
        pk = -1.d0*matrix_dot_vec(Hk,gfk)
        write(*,*)'Iteration info gfk*pk',  dot_product(gfk,pk)
        call dogleg(Delta_k,pk,gfk,Hk,sk,pred,Bs)  ! paramater pred is equal to the predicted reduction in the function value
!       call ddogleg(Delta_k,pk,gfk,Hk,sk,pred,Bs) !csonti test
        alpha = direction_alpha(sk,gfk)


        niter = niter+1

        xk = xk + sk

        results    = fegrad(xk)
        fval_prev  = fval
        fval       = results(ndim+1)

        nfeval  = nfeval + 1
        ngeval  = ngeval + 1

        do i=1,ndim
          yk(i) = results(i) - gfk(i)
          gfk_prev(i) = gfk(i)
          gfk(i)= results(i)
        enddo

        ared  = fval_prev - fval  ! actual reduction in function value

        gnorm = vecnorm(gfk,2)    ! Euclidean
        gmax  = vecnorm(gfk,1000) ! infinit
        snorm = vecnorm(sk,2)
        smax  = vecnorm(sk,1000)

        if(lBaker) then !Baker conv criteria; JCC, 14, 1085 (1993)
          if( gmax <= 3.d-4 .and. (ared<=1.d-6 .or. smax<=3.d-4) ) then 
            notConverged = .False.
          endif
        else
          if( gnorm <= gtol) notConverged = .False.
        endif

        if(.not. notConverged) then
          if(plevel.gt.1) then 
            write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
          endif
          exit
        endif


        call wrt_iter(niter)
        if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info

        if(niter.eq.1) then 
          write(*,'(5x,a)') &         
             'All quantities are in atomic units.'
          write(*,'(5x,a,a5,6(a22))') &         
             'Iteration info ','step','energy','rms force','max force',&
             'rms step','max step','energy change'
        endif
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
        endif

        q = ared/pred
!        if(q .le. eta) then
!          xk  = xk - sk
!          gfk = gfk_prev
!          fval = fval_prev
!        endif

        call trustR_update(ared,pred,Delta_k,snorm)

        call bfgs_update_damped(sk,yk,Hk,q,Bs)
!     write(*,*) 'Pmat*Pinv i/o:'
!     call read_Pmat(ndim,Pmat)
!     call print_matrixR(Pmat,10)
!     Hk = matmul(Hk,Pmat)
!     Hk = matmul(Pmat,Hk)

!     write(*,*) 'Hk before inverse:'
!     call print_matrixR(Hk,10)

!     call matrix_inverse(Hk,Hinv)

!     write(*,*) 'Hinv:'
!     call print_matrixR(Hinv,10)

!     One = matmul(Hk,Hinv)
!     write(*,*) 'One:'
!     call print_matrixR(One,10)


!     Atest = reshape((/1,2,4,0,-1,1,2,3,8/), shape(Atest))
!     write(*,*) ' matrix A:'
!     call print_matrixR(Atest,3)

!     call matrix_inverse(Atest,Ctest)
!     write(*,*) ' matrix Ainv:'
!     call print_matrixR(Ctest,3)

!     Atest = matmul(Atest,Ctest)

!     write(*,*) ' matrix test i/o:'
!     call print_matrixR(Atest,3)

!     call mrccend(1)
 


!       sk = matrix_dot_vec(Pmat,sk)
!       call bfgs_update(sk,yk,Hk)

!       write(*,*) 'Projection matrix after i/o:'
!       call print_matrixR(Pmat,10)
!       write(*,*) 'Projection matrix inverse after i/o:'
!       call print_matrixR(Pinv,10)
!       One = matmul(Hk,Pinv)
!       write(*,*) 'Pmat*Pinv i/o:'
!       call print_matrixR(One,10)

!       Hk=matmul(Hk,Pmat)
!       Hk=matmul(Pmat,Hk)

!       call matrix_pseudoinverse(Pmat,Pinv)
!       Hk=matmul(Hk,Pinv)
!       Hk=matmul(Pinv,Hk)
!       call matrix_inverse(Pmat,Hk)

!       call print_inverse(Hk)


      enddo

      fopt = fval
      args = xk

      if(.not. notConverged .and. &
         niter .lt. maxiter .and. &
         .not.lmessage) then
         bool=.True.
         if(lmolden) then
           call ishell('cat COORD.xyz.energies >> MOLDEN')
           call ishell('cat COORD.xyz.rforce >> MOLDEN')
           call ishell('cat COORD.xyz.mforce >> MOLDEN')
           call ishell('cat COORD.xyz.geometries >> MOLDEN')
           call ishell('rm -f COORD.xyz.* COORD.xyz.energies')
           call ishell('rm -f COORD.xyz.rforce COORD.xyz.mforce')
           call ishell('rm -f COORD.xyz.geometries')
         endif
      endif


      end subroutine bfgs
!************************************************************************

!************************************************************************
      subroutine bfgs_rfo(fegrad,args,Hk,&
                      ftol,gtol,xtol,maxiter,maxfun,finit,&
                      fopt,niter,nfeval,ngeval,plevel,lmolden,bool)
!************************************************************************
! The Broyden-Fletcher-Goldfarb-Shanno algorithm;
! written by J Csontos
!
! Hk is the inverse of the Hessian, see Nocedal and Wright, equation 8.16

      implicit none 
      double precision, intent(in)  :: ftol,gtol,xtol
      double precision, intent(out) :: finit
      double precision, intent(out) :: fopt
      integer, intent(in) :: maxfun,maxiter,plevel
      integer, intent(out) :: niter,nfeval,ngeval
      double precision, dimension(:), intent(inout) :: args
      logical, intent(in)  :: lmolden ! Is molden file required?
      logical, intent(out) :: bool 

      double precision, parameter :: Delta_0=0.5d0
      double precision, parameter :: eta=1.0d-3
!     double precision, parameter :: r=1.0d-2
!     double precision, parameter :: etol=1.0d-6

!     local variables
      integer :: i,j,ndim,nfseval,ngseval
      double precision :: gnorm,snorm,gmax,smax,rtmp,alpha,thres
      double precision :: pred,ared,q,Delta_k,a,b
      double precision :: fval,fval_prev,functol,argtol
      double precision, dimension(size(args)) :: gfk,gfk_prev,xk,xkp1
      double precision, dimension(size(args)) :: sk,pk,yk,zk,Bs
      double precision, dimension(size(args)+1) :: results
      double precision, dimension(size(args),size(args)) :: Hk,Pmat

      double precision, dimension(size(args)) :: gval
      logical :: lmessage,notConverged,lBaker
      double precision :: alpha_k,sBs,ss
      character(len=512) :: si,scommand

      interface
        function fegrad(z)
          double precision, dimension(:), intent(inout) :: z
          double precision, dimension(size(z)+1) :: fegrad
        end function
      end interface

! ***
! f2min   ->  external function to be minimized
! args()  ->  independent variables of f2min;
!         ->  as an input:  initial guess
!         ->  as an output: coordinates of the minimum
! grad    ->  gradient of f2min
!
! gtol    ->  convergence criterion (largest element of the gradient vector)
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! finit   ->  initial value of f2min
! fopt    ->  minimal value of f2min
! niter   ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! ngeval  ->  number of gradient evaluations performed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! ***

      ndim = size(args) ! dimension of the grad vector (fegrad(1:ndim)); the energy is stored in fegrad(ndim+1)
      nfeval = 0
      ngeval = 0
      niter  = 0
      bool = .False.
      lBaker = .True. ! Baker's convergence criteria are set

!     call print_matrixR(Hk,9)

      results = fegrad(args)
      finit   = results(ndim+1) ! initial fucntion value
      ngeval  = ngeval+1
      nfeval  = nfeval+1

      do i=1,ndim
        gfk(i) = results(i)
      enddo

      fval      = finit              ! current function value
      fval_prev = fval + 5000.0d0       ! previous function value

      xk    = args
      gnorm = vecnorm(gfk,2)    ! Euclidean
      gmax  = vecnorm(gfk,1000) ! infinit

      call wrt_iter(niter)  ! # interations is written in the file ITER
      if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info
      
      Delta_k = Delta_0
      notConverged = .True.
      if( gnorm <= gtol) notConverged = .False.
      if(.not. notConverged) then
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
           'Iteration info ',niter,fval,gnorm,gmax,0.d0,0.d0,0.d0
        endif
      endif

      do while(notConverged .and. niter < maxiter)! csonti introduce ftol and xtol here
        pk = -1.d0*matrix_dot_vec(Hk,gfk)
        write(*,*)'Iteration info gfk*pk',  dot_product(gfk,pk)
        call dogleg(Delta_k,pk,gfk,Hk,sk,pred,Bs)  ! paramater pred is equal to the predicted reduction in the function value
!       call ddogleg(Delta_k,pk,gfk,Hk,sk,pred,Bs) !csonti test
        alpha = direction_alpha(sk,gfk)


        niter = niter+1

        xk = xk + sk

        results    = fegrad(xk)
        fval_prev  = fval
        fval       = results(ndim+1)

        nfeval  = nfeval + 1
        ngeval  = ngeval + 1

        do i=1,ndim
          yk(i) = results(i) - gfk(i)
          gfk_prev(i) = gfk(i)
          gfk(i)= results(i)
        enddo

        ared  = fval_prev - fval  ! actual reduction in function value

        gnorm = vecnorm(gfk,2)    ! Euclidean
        gmax  = vecnorm(gfk,1000) ! infinit
        snorm = vecnorm(sk,2)
        smax  = vecnorm(sk,1000)

        if(lBaker) then !Baker conv criteria; JCC, 14, 1085 (1993)
          if( gmax <= 3.d-4 .and. (ared<=1.d-6 .or. smax<=3.d-4) ) then 
            notConverged = .False.
          endif
        else
          if( gnorm <= gtol) notConverged = .False.
        endif

        if(.not. notConverged) then
          if(plevel.gt.1) then 
            write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
          endif
          exit
        endif


        call wrt_iter(niter)
        if(lmolden) call wrt_molden(niter,fval,gnorm,gmax)! MOLDEN info

        if(niter.eq.1) then 
          write(*,'(5x,a)') &         
             'All quantities are in atomic units.'
          write(*,'(5x,a,a5,6(a22))') &         
             'Iteration info ','step','energy','rms force','max force',&
             'rms step','max step','energy change'
        endif
        if(plevel.gt.1) then 
          write(*,'(5x,a,i5,6(2x,f20.12))') &         
             'Iteration info ',niter,fval,gnorm,gmax,snorm,smax,ared
        endif

        q = ared/pred
        if(q .le. eta) then
          xk  = xk - sk
          gfk = gfk_prev
          fval = fval_prev
        endif

        call trustR_update(ared,pred,Delta_k,snorm)

        call bfgs_update_damped(sk,yk,Hk,q,Bs)
        call read_Pmat(ndim,Pmat)
        write(*,*) 'Projection matrix after i/o:'
        call print_matrixR(Pmat,10)
        Hk=matmul(Hk,Pmat)
        Hk=matmul(Pmat,Hk)

      enddo

      fopt = fval
      args = xk

      if(.not. notConverged .and. &
         niter .lt. maxiter .and. &
         .not.lmessage) then
         bool=.True.
         if(lmolden) then
           call ishell('cat COORD.xyz.energies >> MOLDEN')
           call ishell('cat COORD.xyz.rforce >> MOLDEN')
           call ishell('cat COORD.xyz.mforce >> MOLDEN')
           call ishell('cat COORD.xyz.geometries >> MOLDEN')
           call ishell('rm -f COORD.xyz.* COORD.xyz.energies')
           call ishell('rm -f COORD.xyz.rforce COORD.xyz.mforce')
           call ishell('rm -f COORD.xyz.geometries')
         endif
      endif


      end subroutine bfgs_rfo
!************************************************************************

!************************************************************************
        subroutine bfgs_update_B(sk,yk,Bk)
!************************************************************************
!************************************************************************
! BFGS Update
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Nocedal and Wright, eq. 18.22, pg. 538

      implicit none
      double precision, dimension(:), intent(in) :: sk
      double precision, dimension(:), intent(in) :: yk
      double precision, dimension(:,:),intent(inout) ::Bk
      !local
      double precision :: sBs,thres,theta_k,denom,rho_k,r_k
      double precision, dimension(:,:), allocatable :: Id,A1,A2,A3
      double precision, dimension(:), allocatable ::Bs 
      integer :: i,j,ndim

      ndim = size(sk)

      allocate(Id(ndim,ndim))
      allocate(A1(ndim,ndim))
      allocate(Bs(ndim))

      allocate(A2(ndim,ndim))
      allocate(A3(ndim,ndim))

      do j=1,ndim             ! initialization of arrays
        do i=1,ndim
          if(i .eq. j) then
            Id(i,j)  = 1.d0
          else
            Id(i,j)  = 0.d0
          endif
          A1(i,j) = 0.d0
          A2(i,j) = 0.d0
          A3(i,j) = 0.d0
        enddo
      enddo



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      denom = dot_product(yk,sk)
      if(denom < 1.d-6) then ! csonti
        rho_k = 1.d6
      else
        rho_k = 1.d0/denom
      endif

      Bs  = matrix_dot_vec(Bk,sk)
      sBs = dot_product(sk,Bs) ! r_k = 1/sBs

      if(sBs < 1.d-6) then ! csonti
        r_k = 1.d6
      else
        r_k = 1.d0/sBs
      endif


      A1 = outer_product(Bs,sk)
      A2 = matmul(A1,Bk)

      A3 = outer_product(yk,yk)
      do j=1,ndim
        do i=1,ndim
          A2(i,j) = -1.d0*r_k * A2(i,j)
          A3(i,j) = rho_k * A3(i,j)
        enddo
      enddo

      do j=1,ndim
        do i=1,ndim
          Bk(i,j) = Bk(i,j) + A2(i,j) + A3(i,j) 
        enddo
      enddo

!     call print_matrixR(Bk,9)

      deallocate(Id)
      deallocate(A1)
      deallocate(Bs)

      deallocate(A2)
      deallocate(A3)

      end subroutine bfgs_update_B
!************************************************************************

!************************************************************************
        subroutine bfgs_update_damped_B(sk,yk,Bk,q,Bs)
!************************************************************************
!************************************************************************
! BFGS Update
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Nocedal and Wright, eq. 18.22, pg. 538

      implicit none
      double precision, dimension(:), intent(in) :: sk,  Bs
      double precision, dimension(:), intent(inout) :: yk
      double precision, intent(in) :: q
      double precision, dimension(:,:),intent(inout) ::Bk
      !local
      double precision :: sBs,thres,theta_k,denom,rho_k,r_k,sy
      double precision, dimension(:,:), allocatable :: Id,A1,A2,A3
      integer :: i,j,ndim

      ndim = size(sk)

      allocate(Id(ndim,ndim))
      allocate(A1(ndim,ndim))

      allocate(A2(ndim,ndim))
      allocate(A3(ndim,ndim))

      do j=1,ndim             ! initialization of arrays
        do i=1,ndim
          if(i .eq. j) then
            Id(i,j)  = 1.d0
          else
            Id(i,j)  = 0.d0
          endif
          A1(i,j) = 0.d0
          A2(i,j) = 0.d0
          A3(i,j) = 0.d0
        enddo
      enddo

      sy  = dot_product(sk,yk)
      sBs = dot_product(sk,Bs) ! r_k = 1/sBs

      thres = 0.2d0*sBs

      if(sy .ge. thres) then
        theta_k = 1.0d0
      else
!       theta_k = (0.8d0*sBs)/(sBs-sy)
!       theta_k = (0.8d0*thres)/(thres-sy)
        theta_k = (0.8d0*sBs)/(sBs-sy)
      endif

      yk(:) = theta_k*yk(:) + (1.d0-theta_k)*Bs(:)
!     write(*,'(A,3f12.6)')'trust theta_k: ',theta_k,sy,0.2d0*sBs



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      denom = dot_product(yk,sk)
      if(denom < 1.d-6) then ! csonti
        rho_k = 1.d6
      else
        rho_k = 1.d0/denom
      endif

      if(sBs < 1.d-6) then ! csonti
        r_k = 1.d6
      else
        r_k = 1.d0/sBs
      endif


      A1 = outer_product(Bs,sk)
      A2 = matmul(A1,Bk)

      A3 = outer_product(yk,yk)
      do j=1,ndim
        do i=1,ndim
          A2(i,j) = -1.d0*r_k * A2(i,j)
          A3(i,j) = rho_k * A3(i,j)
        enddo
      enddo

      do j=1,ndim
        do i=1,ndim
          Bk(i,j) = Bk(i,j) + A2(i,j) + A3(i,j) 
        enddo
      enddo

!     call print_matrixR(Bk,9)

      deallocate(Id)
      deallocate(A1)

      deallocate(A2)
      deallocate(A3)

      end subroutine bfgs_update_damped_B
!************************************************************************

!************************************************************************
        subroutine bfgs_update(sk,yk,Hk)
!************************************************************************
!************************************************************************
! BFGS Update
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Nocedal and Wright, eq. 18.22, pg. 538

      implicit none
      double precision, dimension(:), intent(in) :: sk
      double precision, dimension(:), intent(in) :: yk
      double precision, dimension(:,:),intent(inout) ::Hk
      !local
      double precision :: sBs,thres,theta_k,denom,rho_k
      double precision, dimension(:,:), allocatable :: Id,A1,A2,A3
      integer :: i,j,ndim

      ndim = size(sk)

      allocate(Id(ndim,ndim))
      allocate(A1(ndim,ndim))
      allocate(A2(ndim,ndim))
      allocate(A3(ndim,ndim))

      do j=1,ndim             ! initialization of arrays
        do i=1,ndim
          if(i .eq. j) then
            Id(i,j)  = 1.d0
          else
            Id(i,j)  = 0.d0
          endif
          A1(i,j) = 0.d0
          A2(i,j) = 0.d0
          A3(i,j) = 0.d0
        enddo
      enddo



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      denom = dot_product(yk,sk)
      if(denom < 1.d-12) then
        rho_k = 1.d3
      else
        rho_k = 1.d0/denom
      endif
      A1 = outer_product(sk,yk)
      A2 = outer_product(yk,sk)
      A3 = outer_product(sk,sk)
      do j=1,ndim
        do i=1,ndim
          A1(i,j) = Id(i,j) - rho_k * A1(i,j)
          A2(i,j) = Id(i,j) - rho_k * A2(i,j)
          A3(i,j) = rho_k * A3(i,j)
        enddo
      enddo

      Hk = matmul( A1,matmul( Hk,A2) )
      do j=1,ndim
        do i=1,ndim
          Hk(i,j) = Hk(i,j) + A3(i,j) 
        enddo
      enddo

!     call print_matrixR(Hk,9)

      deallocate(Id)
      deallocate(A1)
      deallocate(A2)
      deallocate(A3)

      end subroutine bfgs_update
!************************************************************************

!************************************************************************
        subroutine bfgs_update_damped(sk,yk,Hk,q,Bs)
!************************************************************************
!************************************************************************
! BFGS Update
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Nocedal and Wright, eq. 18.22, pg. 538

      implicit none
      double precision, dimension(:), intent(in) :: sk,  Bs
      double precision, dimension(:), intent(inout) :: yk
      double precision, intent(in) :: q
      double precision, dimension(:,:),intent(inout) ::Hk
      !local
      double precision :: sy,sBs,thres,theta_k,denom,rho_k
      double precision, dimension(:,:), allocatable :: Id,A1,A2,A3
      integer :: i,j,ndim

      ndim = size(sk)

      allocate(Id(ndim,ndim))
      allocate(A1(ndim,ndim))
      allocate(A2(ndim,ndim))
      allocate(A3(ndim,ndim))

      do j=1,ndim             ! initialization of arrays
        do i=1,ndim
          if(i .eq. j) then
            Id(i,j)  = 1.d0
          else
            Id(i,j)  = 0.d0
          endif
          A1(i,j) = 0.d0
          A2(i,j) = 0.d0
          A3(i,j) = 0.d0
        enddo
      enddo


!     Bs    = matrix_dot_vec(Hk,sk)
      sy  = dot_product(sk,yk)
      sBs = dot_product(sk,Bs)
      thres = 0.2d0*sBs

      if(sy .ge. thres) then
        theta_k = 1.0d0
      else
!       theta_k = (0.8d0*sBs)/(sBs-sy)
!       theta_k = (0.8d0*thres)/(thres-sy)
        theta_k = (0.8d0*sBs)/(sBs-sy)
      endif

      yk(:) = theta_k*yk(:) + (1.d0-theta_k)*Bs(:)
      write(*,'(A,3f12.6)')'trust theta_k: ',theta_k,sy,0.2d0*sBs

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      denom = dot_product(yk,sk)
      if(denom < 1.d-12) then
        rho_k = 1.d3
      else
        rho_k = 1.d0/denom
      endif
      A1 = outer_product(sk,yk)
      A2 = outer_product(yk,sk)
      A3 = outer_product(sk,sk)
      do j=1,ndim
        do i=1,ndim
          A1(i,j) = Id(i,j) - rho_k * A1(i,j)
          A2(i,j) = Id(i,j) - rho_k * A2(i,j)
          A3(i,j) = rho_k * A3(i,j)
        enddo
      enddo

      Hk = matmul( A1,matmul( Hk,A2) )
      do j=1,ndim
        do i=1,ndim
          Hk(i,j) = Hk(i,j) + A3(i,j) 
        enddo
      enddo
! BFGS Update END
!     endif
!     else
!       Hk = Id
!     endif

!     write(*,*)'inverse Hessian matrix'
!     call print_matrixR(Hk,9)

      deallocate(Id)
      deallocate(A1)
      deallocate(A2)
      deallocate(A3)

      end subroutine bfgs_update_damped
!************************************************************************

!************************************************************************
      subroutine trustR_update(ared,pred,Delta_k,snorm)
!************************************************************************
! it updates the trust radius (Delta_k)
!************************************************************************

      implicit none
      double precision, intent(in) :: ared,pred,snorm
      double precision, intent(inout) :: Delta_k
      !local
      double precision :: q
      double precision, parameter :: Delta_max=0.8d0
      double precision, parameter :: Delta_min=0.0001d0
      double precision, parameter :: scale_up=1.2d0
      double precision, parameter :: scale_down=0.67d0
      

      q=ared/pred

      if(q < 0.25d0) then
        Delta_k = scale_down*snorm!Delta_k!
        if(Delta_k < Delta_min) Delta_k = Delta_min
!       write(*,'(A,5f12.6)') &
!         'The trust radius was decreased in this step to',&
!          Delta_k,q,snorm,ared,pred
      else
        if(q > 0.75d0 .and. snorm > 0.8d0*Delta_k) then
          Delta_k = min(scale_up*Delta_k,Delta_max)
!         write(*,'(A,5f12.6)') &
!           'The trust radius was increased in this step to',&
!           Delta_k,q,snorm,ared,pred
        else
          Delta_k = 1.0d0*Delta_k
!         write(*,'(A,5f12.6)') & 
!           'The trust radius was not changed in this step',&
!            Delta_k,q,snorm,ared,pred
        endif
      endif

      return
      end subroutine trustR_update
!************************************************************************

!************************************************************************
      double precision function direction_alpha(step,gradient)
!************************************************************************
! it calculates the angle between the step and gradient vector
! that should be between -90 and -180 somewhere
!************************************************************************
      implicit none
      double precision, dimension(:), intent(in) :: step,gradient
      !local
      double precision :: rtmp

      rtmp = dot_product(step,gradient)/&
             (sqrt(dot_product(gradient,gradient))&
             *sqrt(dot_product(step,step)))
      if(rtmp> 1.d0) rtmp =  1.d0
      if(rtmp<-1.d0) rtmp = -1.d0
      direction_alpha = (-180.d0/PI)*dacos(rtmp)
      if(direction_alpha > -95.d0) then
        write(*,*) 'Warning: the optimizer might move uphill!!!'
        write(*,'(A,f8.1)')&
              'Iteration info direction alpha: ',direction_alpha
      endif

      end function direction_alpha
!************************************************************************

!************************************************************************
      subroutine ddogleg(Delta,pB,grad,H,sk,pred,Bs)
!************************************************************************
! double dogleg Dennis, Robert, Schnabel section 6.4.2 pg 139-147.

      implicit none
      
      double precision, intent(in)                  :: Delta ! trust radius
      double precision, dimension(:), intent(in)    :: pB    !
      double precision, dimension(:), intent(in)    :: grad  ! gradient vector
      double precision, dimension(:,:), intent(in)  :: H     ! the inverse Hessian matrix, B^-1 = H
      double precision, dimension(:), intent(out)   :: sk    ! the dogleg step
      double precision, intent(out)                 :: pred  ! predicted function value
      double precision, dimension(:), intent(out)   :: Bs
      !local
      integer :: ndim
      double precision :: pB_norm,pU_norm,nom,denom,a,hb,c,kappa,kappa2
      double precision :: gBg,gHg,gamma_,eta_
      double precision, dimension(size(grad)) :: pU,pBU,gtmp
      double precision, dimension(:,:), allocatable :: B,E


      ndim = size(grad)
      allocate(B(ndim,ndim))
      allocate(E(ndim,ndim)) ! just for test remove csonti

      call matrix_pseudoinverse(H,B)
!     write(*,*)'inverse Hessian'
!     call print_matrixR(H,9)
!     write(*,*)'Hessian'
!     call print_matrixR(B,9)
!     E = matmul(H,B)
!     write(*,*)'H*B'
!     call print_matrixR(E,9)


      pB_norm = vecnorm(pB,2)

      ! CASE A -- the full Newton step is inside the trust radius
      if(pB_norm .le. Delta) then 
        write(*,*) &
          'The full Newton step is inside the trust radius'
        sk    = pB
        Bs    = matrix_dot_vec(B,sk)
        pred  = -1.d0*(dot_product(grad,sk)+0.5d0*dot_product(sk,Bs))
        if(allocated(B)) deallocate(B)
        if(allocated(E)) deallocate(E)
        return 
      endif


      gtmp  = matrix_dot_vec(B,grad)
!     call print_vectorR(gtmp,8)
!     write(*,*)'dogleg after return'

!     gtmp  = dot_product(B,transpose(grad))
!     call print_vectorR(gtmp,8)
      nom   = dot_product(grad,grad)
      denom = dot_product(grad,gtmp)
      pU = -1.d0*(nom/denom)*grad
!     write(*,*)'grad'
!     call print_vectorR(grad,8)
!     write(*,*)'pU',-1.d0*(nom/denom)
!     call print_vectorR(pU,8)

      ! CASE B -- the Cauchy step reaches outside the border of the trust radius; let's scale it back to the border
      pU_norm = vecnorm(pU,2)
      if(pU_norm .gt. Delta) then 
        sk    = (Delta/pU_norm)*pU
!       sk    = sk +  (Delta/pB_norm)*pB
        Bs    = matrix_dot_vec(B,sk)
        pred  = -1.d0*(dot_product(grad,sk)+0.5d0*dot_product(sk,Bs))
        write(*,'(A)',advance='no') &
          'Let us scale the Cauchy step (steepest descent) back '
        write(*,*) 'to the trust border'
        if(allocated(B)) deallocate(B)
        if(allocated(E)) deallocate(E)
        return 
      endif
      

      ! CASE C -- quadratic part; we need the point of intersection with the trust radius
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      gBg    = dot_product(grad,matrix_dot_vec(B,grad))
      gHg    = dot_product(grad,matrix_dot_vec(H,grad))
      gamma_ = (dot_product(grad,grad))**2/(gBg*gHg)
      eta_   = 0.8d0*gamma_ + 0.2d0
      sk     = eta_*pB
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


      pBU = sk - pU
      a  = dot_product(pBU,pBU)
      hb = dot_product(pU,pBU) !half b, hb=0.5*b
      c  = dot_product(pU,pU) -  Delta**2
      kappa = (-1.d0*hb + sqrt(hb**2 - a*c))/a! kappa = (tau-1)
      kappa2 = (-1.d0*hb - sqrt(hb**2 - a*c))/a! kappa = (tau-1)

      write(*,'(A)',advance='no') &
        'We need the intersection of the trust radius with that of'
      write(*,*) 'the dogleg step (Cauchy step + Newton step)'
      write(*,*) 'Trust radius tau is equal to ',kappa+1.d0
      write(*,*) 'Trust radius tau2 is equal to ',kappa2+1.d0

      sk    = pU + kappa*pBU
      Bs    = matrix_dot_vec(B,sk)
      pred  = -1.d0*(dot_product(grad,sk)+0.5d0*dot_product(sk,Bs))

      if(allocated(B)) deallocate(B)   ! deallocation in case A, B, and C !csonti
      if(allocated(E)) deallocate(E)
      return

      end subroutine ddogleg
!************************************************************************

!************************************************************************
      subroutine dogleg(Delta,pB,grad,H,sk,pred,Bs)
!************************************************************************
! see Nocedal and Wright, equations 4.10-4.13, pg. 71-73.

      implicit none
      
      double precision, intent(in)                  :: Delta ! trust radius
      double precision, dimension(:), intent(in)    :: pB    !
      double precision, dimension(:), intent(in)    :: grad  ! gradient vector
      double precision, dimension(:,:), intent(in)  :: H     ! the inverse Hessian matrix, B^-1 = H
      double precision, dimension(:), intent(out)   :: sk    ! the dogleg step
      double precision, intent(out)                 :: pred  ! predicted function value
      double precision, dimension(:), intent(out)   :: Bs
      !local
      integer :: ndim
      double precision :: pB_norm,pU_norm,nom,denom,a,hb,c,kappa
      double precision, dimension(size(grad)) :: pU,pBU,gtmp
      double precision, dimension(:,:), allocatable :: B,E


      ndim = size(grad)
      allocate(B(ndim,ndim))
      allocate(E(ndim,ndim)) ! just for test remove csonti

      call matrix_pseudoinverse(H,B)
!     write(*,*)'inverse Hessian'
!     call print_matrixR(H,9)
!     write(*,*)'Hessian'
!     call print_matrixR(B,9)
!     E = matmul(H,B)
!     write(*,*)'H*B'
!     call print_matrixR(E,9)


      pB_norm = vecnorm(pB,2)

      ! CASE A -- the full Newton step is inside the trust radius
      if(pB_norm .le. Delta) then 
!       write(*,*) &
!         'The full Newton step is inside the trust radius'
        sk    = pB
        Bs    = matrix_dot_vec(B,sk)
        pred  = -1.d0*(dot_product(grad,sk)+0.5d0*dot_product(sk,Bs))
        if(allocated(B)) deallocate(B)
        if(allocated(E)) deallocate(E)
        return 
      endif


      gtmp  = matrix_dot_vec(B,grad)
!     call print_vectorR(gtmp,8)
!     write(*,*)'dogleg after return'

!     gtmp  = dot_product(B,transpose(grad))
!     call print_vectorR(gtmp,8)
      nom   = dot_product(grad,grad)
      denom = dot_product(grad,gtmp)
      pU = -1.d0*(nom/denom)*grad
!     write(*,*)'grad'
!     call print_vectorR(grad,8)
!     write(*,*)'pU',-1.d0*(nom/denom)
!     call print_vectorR(pU,8)

      ! CASE B -- the Cauchy step reaches outside the border of the trust radius; let's scale it back to the border
      pU_norm = vecnorm(pU,2)
      if(pU_norm .gt. Delta) then 
        sk    = (Delta/pU_norm)*pU
        Bs    = matrix_dot_vec(B,sk)
        pred  = -1.d0*(dot_product(grad,sk)+0.5d0*dot_product(sk,Bs))
!       write(*,'(A)',advance='no') &
!         'Let us scale the Cauchy step (steepest descent) back '
!       write(*,*) 'to the trust border'
        if(allocated(B)) deallocate(B)
        if(allocated(E)) deallocate(E)
        return 
      endif

      ! CASE C -- quadratic part; we need the point of intersection with the trust radius
      pBU = pB - pU
      a  = dot_product(pBU,pBU)
      hb = dot_product(pU,pBU) !half b, hb=0.5*b
      c  = dot_product(pU,pU) -  Delta**2
      kappa = (-1.d0*hb + sqrt(hb**2 - a*c))/a! kappa = (tau-1)

!     write(*,'(A)',advance='no') &
!       'We need the intersection of the trust radius with that of'
!     write(*,*) 'the dogleg step (Cauchy step + Newton step)'
!     write(*,*) 'Trust radius tau is equal to ',kappa+1.d0

      sk    = pU + kappa*pBU
      Bs    = matrix_dot_vec(B,sk)
      pred  = -1.d0*(dot_product(grad,sk)+0.5d0*dot_product(sk,Bs))

      if(allocated(B)) deallocate(B)   ! deallocation in case A, B, and C !csonti
      if(allocated(E)) deallocate(E)
      return

      end subroutine dogleg
!************************************************************************

!************************************************************************
      subroutine print_inverse(H)
!************************************************************************
! see Nocedal and Wright, equations 4.10-4.13, pg. 71-73.

      implicit none
      
      double precision, dimension(:,:), intent(in)  :: H     ! the inverse Hessian matrix, B^-1 = H
      !local
      integer :: ndim
      double precision, dimension(:,:), allocatable :: B


      ndim = size(H,1)
      allocate(B(ndim,ndim))

      call matrix_pseudoinverse(H,B)
      write(*,*)'HESSIAN MATRIX'
      call print_MatrixR(B,9)

      if(allocated(B)) deallocate(B)   ! deallocation in case A, B, and C !csonti
      return

      end subroutine print_inverse 
!************************************************************************

!************************************************************************
      subroutine quadratic(sk,grad,H,pred,Bs)
!************************************************************************
! see Nocedal and Wright, equations 4.10-4.13, pg. 71-73.

      implicit none
      
      double precision, dimension(:), intent(in)    :: sk    !
      double precision, dimension(:), intent(in)    :: grad  ! gradient vector
      double precision, dimension(:,:), intent(in)  :: H     ! the inverse Hessian matrix, B^-1 = H
      double precision, intent(out)                 :: pred  ! predicted function value
      double precision, dimension(:), intent(out)   :: Bs
      !local
      integer :: ndim
      double precision :: pB_norm,pU_norm,nom,denom,a,hb,c,kappa
      double precision, dimension(size(grad)) :: pU,pBU,gtmp
      double precision, dimension(:,:), allocatable :: B,E


      ndim = size(grad)
      allocate(B(ndim,ndim))
      allocate(E(ndim,ndim)) ! just for test remove csonti

      call matrix_pseudoinverse(H,B)

      Bs    = matrix_dot_vec(B,sk)
      pred  = -1.d0*(dot_product(grad,sk)+0.5d0*dot_product(sk,Bs))

      if(allocated(B)) deallocate(B)   ! deallocation in case A, B, and C !csonti
      if(allocated(E)) deallocate(E)
      return

      end subroutine quadratic 
!************************************************************************


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! LINEAR ALGEBRA
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      double precision function vecnorm(vec,p)
! it returns the usual p-norm of the vector vec
! the infinit norm can be called with p>999

      implicit none
      double precision, dimension(:),intent(in) :: vec
      integer, intent(in) :: p
      !local
      integer i,ndim

      ndim = size(vec)

      if(p .gt. 999) then ! this is the infinity norm
        if ( dabs(maxval(vec)) .gt. dabs(minval(vec)) ) then
          vecnorm = dabs(maxval(vec))
        else
          vecnorm = dabs(minval(vec))
        endif
      else
        vecnorm = 0.d0
        do i=1,ndim
          vecnorm = vecnorm + dabs( vec(i) )**(p)
        enddo
        vecnorm = vecnorm**(1.d0/p)
      endif 

      return
      end function vecnorm
!************************************************************************

!************************************************************************
      function vec_norm_comp(A)
!************************************************************************
! it returns the components of the normal vector corresponding to A
! A is a 3-dimensional vector
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A
      double precision, dimension(3) :: vec_norm_comp
      !local
      integer :: i

      do i=1,3
        vec_norm_comp(i) = A(i)/vec_length(A)
      enddo

      return
      end function vec_norm_comp
!************************************************************************

!************************************************************************
      function vec_comp(A,B)
!************************************************************************
! it returns the components of the vector pointing from A to B
! A and B are 3-dimensional points
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A,B
      double precision, dimension(3) :: vec_comp
      !local
      integer :: i

      do i=1,3
        vec_comp(i) = B(i)-A(i)
      enddo

      return
      end function vec_comp
!************************************************************************

!************************************************************************
      double precision function vec_length(A)
!************************************************************************
! it returns the length of A (Euclidean norm, ||A||)
! for a 3D vector
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A

      vec_length = 0.d0
      vec_length = sqrt( vec_dot(A,A) )

      return
      end function vec_length
!************************************************************************

!************************************************************************
      double precision function vec_dot(A,B)
!************************************************************************
! it returns the dot product of vectors A and B
! A and B are 3-dimensional vectors
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A,B
      !local
      integer :: i

      vec_dot = 0.d0
      do i=1,3
        vec_dot = vec_dot + A(i)*B(i)
      enddo

      return
      end function vec_dot
!************************************************************************

!************************************************************************
      function vec_cross(A,B)
!************************************************************************
! it returns the cross product of vectors A and B
! A and B are 3-dimensional vectors
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A,B
      double precision, dimension(3) :: vec_cross
      !local
      double precision,dimension(3)  :: AxB

      call crproduct(A,B,AxB) 
      vec_cross = AxB

      return
      end function vec_cross
!************************************************************************

!************************************************************************
      double precision function vec_mixed(A,B,C)
!************************************************************************
! it returns the mixed product of vectors A, B, and C
!
! AxB.C
!
! A, B, and C are 3-dimensional vectors
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A,B,C

      vec_mixed = vec_dot(vec_cross(A,B),C)

      return
      end function vec_mixed
!************************************************************************

!************************************************************************
      subroutine crproduct(a,b,acb)
!************************************************************************
! Computes cross product of a,b in 3 dimensions
!************************************************************************
      real*8 a(3),b(3),acb(3)
!
      acb(1)=a(2)*b(3)-a(3)*b(2)
      acb(2)=a(3)*b(1)-a(1)*b(3)
      acb(3)=a(1)*b(2)-a(2)*b(1)
      return
      end subroutine crproduct
!
!************************************************************************

!************************************************************************
      function w_prime_comp(u,v)
!************************************************************************
! it returns the component of the vector w' 
! JCP, 117, 9160 (2002), Eq. 24
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: u,v
      double precision, dimension(3) :: w_prime_comp
      !local
      double precision, dimension(3) :: vtmp
      double precision, parameter :: eps=1.d-6 ! csonti more precise definition is required
      double precision,dimension(3),parameter::v2=(/ 1.d0,-1.d0,1.d0/),&
                                               v3=(/-1.d0, 1.d0,1.d0/) 

      vtmp = vec_cross(u,v) 
      if(vec_length(vtmp).ge.eps) then ! u and v are not parallel
        w_prime_comp = vtmp
      else
        vtmp = vec_cross(u,v2)
        if(vec_length(vtmp).ge.eps) then 
          w_prime_comp = vtmp
        else
          w_prime_comp = vec_cross(u,v3)
        endif
      endif

      return
      end function w_prime_comp
!************************************************************************

!************************************************************************
      subroutine build_projection(B,P)
!************************************************************************
! it builds the projection matrix P from B
!
! P = B*B+
!
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: B                  !m x n
      double precision,dimension(size(B,1),size(B,1)),intent(inout)::P !m x m
      !local
      double precision,dimension(size(B,2),size(B,1))       :: Bp      !n x m
      integer :: i,j,n,m

      m = size(B,1)
      n = size(B,2)

      call matrix_pseudoinverse(B,Bp)
!     call print_matrixR(B,10)
!     call print_matrixR(Bp,10)
!     P=matmul(B,Bp)
!     call print_matrixR(P,10)

      call dgemm('n','n',m,m,n,1.d0,B,m,Bp,n,0.d0,P,m)

      end subroutine build_projection
!************************************************************************

!************************************************************************
      subroutine matrix_inverse_sym(B,Binv)
!************************************************************************
! it computes Binv the inverse of symmetric positive definite matrix B 
! using lapack routines:
!
! dpotrf - compute the Cholesky factorization of a real symmetric 
!          positive definite matrix 
! dpotri - compute the inverse of a real symmetric positive
!          definite matrix A using the Cholesky factorization 
!          A = U**T*U or A = L*L**T computed by DPOTRF
!
! it returns Binv
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: B                  !n x n
      double precision,dimension(size(B,1),size(B,1)),&
      intent(inout)::Binv!n x n
      !local
      integer :: n,info

      external dpotrf
      external dpotri

      n = size(B,1)

      Binv = B   ! to keep B intact

!     write(*,*) 'B in inverse routine'
!     call print_matrixR(Binv,9)

      call dpotrf('L',n,Binv,n,info)
      if(info .ne. 0) then
        write(*,*) &
        'Cholesky factorization failed in matrix_inverse_sym',info
        call mrccend(1)
      endif
!     write(*,*) 'LU decomposition of B'
!     call print_matrixR(Binv,9)

      call dpotri('L',n,Binv,n,info)
      if(info .ne. 0) then
        write(*,*) &
        'Matrix inversion failed in matrix_inverse_sym'
        call mrccend(1)
      endif
!     call print_matrixR(Binv,9)


      end subroutine matrix_inverse_sym 
!************************************************************************

!************************************************************************
      subroutine matrix_inverse(B,Binv)
!************************************************************************
! it computes Binv the inverse of matrix B using lapack routines:
!
! dgetrf - LU decomposition (partial pivoting with raw interchanges)
! dgetri - This method inverts U and then computes inv(A) by solving
!          the system inv(A)*L = inv(U) for inv(A)
!
! it returns Binv
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: B                  !n x n
      double precision,dimension(size(B,1),size(B,1)),&
      intent(inout)::Binv!n x n
      !local
      integer,dimension(size(B,1))                          :: ipiv
      integer :: n,res,i
      integer :: LWORK,info
      double precision,dimension(:),allocatable             :: WORK    !

      external dgetrf
      external dgetri
      n = size(B,1)

      LWORK = n*n
      allocate(WORK(LWORK),STAT=res)
      if(res .eq. 0) then

        Binv = B   ! to keep B intact

!       write(*,*) 'B in inverse routine'
!       call print_matrixR(Binv,9)

        call dgetrf(n,n,Binv,n,IPIV,info)
        if(info .ne. 0) then
          write(*,*) &
          'LU decomposition failed in subroutine matrix_inverse'
          call mrccend(1)
        endif
!       write(*,*) 'LU decomposition of B'
!       call print_matrixR(Binv,9)

        call dgetri(n,Binv,n,IPIV,WORK,LWORK,info)
        if(info .ne. 0) then
          write(*,*) &
          'Matrix inversion failed in subroutine matrix_inverse'
          call mrccend(1)
        endif
!       call print_matrixR(Binv,9)

        deallocate(WORK)
      else
        write(*,*) &
        'Memory allocation was not successfull in matrix_nverse'
        call mrccend(1)
      endif



      end subroutine matrix_inverse 
!************************************************************************

!************************************************************************
      subroutine matrix_diag(B,evec,eval)
!************************************************************************
! it computes the diagonal form of B using lapack routines:
!
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: B                  !n x n
      double precision,dimension(size(B,1),size(B,1)),&
      intent(inout) :: evec!n x n matrix of eigenvectors
      double precision,dimension(size(B,1)),intent(inout) :: eval
      !local
      double precision,dimension(:)  ,allocatable   :: WORK    !
      integer :: res,info,LWORK,n


      n = size(B,1)

      evec = B                                       ! keep B
      LWORK=-1                                       ! check for optimal LWORK value
      allocate(WORK(1))
      call dsyev('V','U',n,evec,n,eval,WORK,LWORK,info) ! lapack routine for diagonalization
      LWORK=WORK(1)
      deallocate(WORK)


      allocate(WORK(LWORK),STAT=res)
      if(res .eq. 0) then
        call dsyev('V','U',n,evec,n,eval,WORK,LWORK,info)
        if(info.ne.0) then
          write(*,*) 'Fatal error at the diagonalization!'
          call mrccend(1)
        endif
        deallocate(WORK)
      else
        write(*,*) &
          'Memory allocation was not successfull in matrix_diag'
          call mrccend(1)
      endif

      end subroutine matrix_diag
!************************************************************************

!************************************************************************
      subroutine matrix_pseudoinverse(B,Bp)
!************************************************************************
! it computes B+ the pseudoinverse of matrix B using singular value
! decomposition:
!
! B  = U * SIGMA * V^T ; the svd of B
! B+ = V * (U * SIGMA^-1)^T
!
! it returns Bp
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: B                  !m x n
      double precision,dimension(size(B,2),size(B,1)),intent(inout)::Bp!n x m
      !local
      double precision,dimension(size(B,1),size(B,1))       :: U       !m x m
      double precision,dimension(size(B,2),size(B,2))       :: VT      !n x n
      double precision,dimension(min(size(B,1),size(B,2)))  :: s
      double precision,dimension(size(B,1),size(B,2))       :: USi     !
      double precision,dimension(size(B,1),size(B,2))       :: O       ! temp matrix
      integer :: n,m,lmax,res
      integer :: LWORK,IWORK(8*min(size(B,1),size(B,2))),info
      double precision,dimension(:),allocatable             :: WORK    !

      m = size(B,1)
      n = size(B,2)

      LWORK=-1                                                  ! check for optimal LWORK value
      allocate(WORK(1))
      call dgesdd('A',m,n,B,m,s,U,m,VT,n,WORK,LWORK,IWORK,info) ! lapack routine for SVD 
      LWORK=WORK(1)
      deallocate(WORK)

      lmax = 2*(min(m,n)*(6+4*min(m,n))+max(m,n))               ! double of that can be found in the lapack manual
      if(LWORK .gt. lmax ) then  
        LWORK = lmax
      endif

      allocate(WORK(LWORK),STAT=res)
      if(res .eq. 0) then
        O = B ! to keep B intact

        call dgesdd('A',m,n,O,m,s,U,m,VT,n,WORK,LWORK,IWORK,info) ! lapack routine for SVD 

        call matrix_pseudo1(U,s,USi)
        call matrix_pseudo2(VT,USi,Bp)

        deallocate(WORK)
      else
        write(*,*) &
        'Memory allocation was not successfull in matrix_pseudoinverse'
        call mrccend(1)
      endif

      end subroutine matrix_pseudoinverse
!************************************************************************

!************************************************************************
      subroutine matrix_pseudo2(VT,USi,Bp)
!************************************************************************
! it computes the matrix multiplication, V * USi^T, 
! where USi = U * SIGMA^-1
!
! B  = U * SIGMA * V^T the svd of B
! B+ = V * SIGMA^-1^T * U^T the pseudoinverse of B or otherwise
! B+ = V * (U * SIGMA^-1)^T
!
! it returns B+ /Bp/ the pseudoinverse of B
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: VT
      double precision,dimension(:,:),intent(in) :: USi
      double precision,&
             dimension(size(VT,2),size(USi,1)),intent(inout) :: Bp
      !local
      integer :: n,m

      m = size(VT,2)
      n = size(USi,1)

      call dgemm('t','t',m,n,m,1.d0,VT,m,USi,n,0.d0,Bp,m)

      end subroutine matrix_pseudo2
!************************************************************************

!************************************************************************
      subroutine matrix_pseudo1(U,s,USi)
!************************************************************************
! it computes the matrix multiplication, U*SIGMA^-1, by scaling the
! column vectors of U with the blas routine sdcal
! s is an array which elements are the column by column scaling
! factors (the diagonal elements of SIGMA)
!
! B  = U * SIGMA * V^T the svd of B
! B+ = V * SIGMA^-1^T * U^T the pseudoinverse of B or otherwise
! B+ = V * (U * SIGMA^-1)^T
!
! it returns USi = U * SIGMA^-1
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in)    :: U
      double precision,dimension(:)                 :: s
      double precision,dimension(:,:),intent(inout) :: USi
      !local
      double precision,dimension(size(U,1)) :: vec,vec_zero
      integer :: i,j,k,n,m

      n = size(U,1)
      m = size(USi,2)
      k = size(s)

      do i=1,n
        vec_zero(i) = 0.d0
      enddo

      do i=1,m
        if(i .le. k) then
          vec = U(1:,i)
          if( s(i) .gt. 1.d-7 ) then ! the criterion is from Pulay - JCP, 96, 2856 (1992)
            call dscal(n,1.d0/s(i),vec,1)
          else
            call dscal(n,0.d0,vec,1)
          endif
          USi(1:,i) = vec
        else
          USi(1:,i) = vec_zero
        endif
      enddo

      end subroutine matrix_pseudo1
!************************************************************************

!************************************************************************
      function matrix_dot_vec(A,vec)
!************************************************************************
! it returns as A*vec
! just a simple test routine
! A is an mxn and the length of vec is n
!************************************************************************
      implicit none
      double precision, dimension(:,:), intent(in) :: A
      double precision, dimension(size(A,2)) :: vec
      double precision, dimension(size(A,1)) :: matrix_dot_vec
      !local
      integer :: ii,i,j,m,n,k

      m = size(A,1)
      n = size(A,2)

      do i=1,m
        matrix_dot_vec(i)=0.d0
        do j=1,n
            matrix_dot_vec(i) = matrix_dot_vec(i) + A(i,j)*vec(j)
!           write(*,*) A(i,j), vec(j)
        enddo
      enddo

      end function matrix_dot_vec
!************************************************************************

!************************************************************************
      subroutine matrix_multi(A,B,P)
!************************************************************************
! it returns P = A*B
! just a simple test routine
! A is an mxk and B is an kxn
!************************************************************************
      implicit none
      double precision, dimension(:,:), intent(in) :: A,B
      double precision, dimension(size(A,1),size(B,2)),intent(inout):: P
      !local
      integer :: ii,i,j,m,n,k

      m = size(A,1)
      n = size(B,2)
      k = size(A,2)

      do i=1,m
        do j=1,n
          P(i,j)=0.d0
          do ii=1,k
            P(i,j) = P(i,j) + A(i,ii)*B(ii,j)
!           write(*,*) A(i,ii),B(ii,j)
          enddo
        enddo
      enddo

      end subroutine matrix_multi
!************************************************************************

!************************************************************************
      subroutine matrix_transpose(A,AT)
!************************************************************************
! it returns the transpose of matrix A /AT/
! A is also retained
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: A
      double precision,dimension(:,:),intent(inout) :: AT
      !local
      double precision :: tmp
      integer :: i,j,ncol,nrow

      tmp = 0.d0

      ncol = size(A,1)
      nrow = size(A,2)

      do j=1,nrow
        do i=1,ncol
          AT(j,i)=A(i,j)
        enddo
      enddo

      end subroutine matrix_transpose
!************************************************************************

!************************************************************************
      double precision function distance2(A,B)
!************************************************************************
! it returns the Euclidean distance between two 3D-points
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A,B
      !local
      integer :: i

      distance2 = 0.d0

      do i=1,3
        distance2 = distance2 + (B(i)-A(i))**2
      enddo
      distance2 = dsqrt(distance2)

      return
      end function distance2
!************************************************************************

!************************************************************************
      double precision function distance6(x1,y1,z1, x2,y2,z2)
!************************************************************************
! it returns the Euclidean distance between two 3D-points
!************************************************************************
      implicit none
      double precision, intent(in) :: x1,y1,z1, x2,y2,z2


      distance6 = (x2-x1)**2 + (y2-y1)**2 + (z2-z1)**2
      distance6 = dsqrt(distance6)

      return
      end function distance6
!************************************************************************

!************************************************************************
      double precision function angle3(A,B,C)
!************************************************************************
! it returns the bond angle between atoms A-B-C; 
! the angle is measured at vertex atom B
!
! A   C
!  \ /
!   B
!
! i.e. the angle between bonds B-A and B-C is returned in radial
! JCP, 117, 9160 (2002), Eq. 23
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A, B, C
      !local
      double precision, dimension(3) :: up,vp,u,v
      double precision :: dtmp
      double precision, parameter :: eps = 1.d-8 ! csonti constraint

      up  = vec_comp(B,A)  ! u'
      vp  = vec_comp(B,C)  ! v'
      u   = vec_norm_comp(up)
      v   = vec_norm_comp(vp)

      dtmp = vec_dot(u,v)

      if( dtmp .gt. 1.d0 ) then
        if( dtmp-1.d0 .gt. eps ) then
          write(*,*) 'Something is wrong in function angle3!!!'
          call mrccend(1)
        else
          dtmp = 1.d0
        endif
      elseif( dtmp .lt. -1.d0 ) then
        if( -1.d0-dtmp .gt. eps ) then
          write(*,*) 'Something is wrong in function angle3!!!'
          call mrccend(1)
        else
          dtmp = -1.d0
        endif
      endif


      angle3 = dacos(dtmp)

      return
      end function angle3
!************************************************************************

!************************************************************************
      double precision function dihedral4(A,B,C,D)
!************************************************************************
! it returns the dihedral angle between bonds A-B and C-D; 
! i.e. the angle between the planes defined by points A,B,C and B,C,D
!
! side view:
!
! A        D
!  \      /
!   B -- C
!
! bond view (along B--C):
! 
! A   (D)
!  \ /    this conformation has a + sign 
!  B(C)
!
! (D)   A
!   \ /   this conformation has a - sign
!   B(C)
!
! A,B are in the front C,D are in the back
!
! the angle is restricted to be in the range -Pi,Pi and measured in radian
! JCP, 117, 9160 (2002), Eq. 31 as well as JCC, 16, 527 (1995)
! 
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A, B, C, D
      !local
      double precision, dimension(3) :: up,vp,wp,u,v,w,uxw,vxw
      double precision :: sinu,sinv,dtmp
      double precision, parameter :: eps = 1.d-8 ! csonti constraint

      up  = vec_comp(B,A)  ! u'
      vp  = vec_comp(C,D)  ! v'
      wp  = vec_comp(B,C)  ! w'

      u   = vec_norm_comp(up)
      v   = vec_norm_comp(vp)
      w   = vec_norm_comp(wp)
  
      uxw = vec_cross(u,w)
      vxw = vec_cross(v,w)

      sinu= dsqrt(1.d0 - (vec_dot(u,w))**2)
      sinv= dsqrt(1.d0 - (vec_dot(v,w))**2)

      dtmp =  vec_dot(uxw,vxw) / (sinu*sinv)

      if( dtmp .gt. 1.d0 ) then
        if( dtmp-1.d0 .gt. eps ) then
          write(*,*) 'Something is wrong in function dihedral4!!!'
          call mrccend(1)
        else
          dtmp = 1.d0
        endif
      elseif( dtmp .lt. -1.d0 ) then
        if( -1.d0-dtmp .gt. eps ) then
          write(*,*) 'Something is wrong in function dihedral4!!!'
          call mrccend(1)
        else
          dtmp = -1.d0
        endif
      endif

      dihedral4 = dacos( dtmp )

      if( vec_mixed(w,u,v) .lt. 0.d0 ) dihedral4 = -1.d0*dihedral4


      return
      end function dihedral4
!************************************************************************
      
!************************************************************************
      logical function is_parallel(A,B,C)
!************************************************************************
! it checks whether or not the lines (A,B) and (B,C) are parallel 
! or A, B, and C are on the same line
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A, B, C
      !local
      double precision, dimension(3) :: up,vp,u,v,uxv

      is_parallel=.False.

      up  = vec_comp(B,A)  ! u'
      vp  = vec_comp(B,C)  ! v'
      u   = vec_norm_comp(up)
      v   = vec_norm_comp(vp)
      uxv = vec_cross(u,v)

      if(vec_dot(uxv,uxv) .lt. 1e-6) is_parallel=.True.
!     write(*,*) 'parallel',is_parallel, A,B,C,up,vp


      return
      end function is_parallel
!************************************************************************


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! WILSON'S B-matrix related
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      subroutine build_Wilson_Bmat(natoms,coords,red_int_matrix,Bmatrix)
!************************************************************************
! it builds Wilson's B-matrix
!************************************************************************
      implicit none
      integer, intent(in) :: natoms
      double precision, dimension(3,natoms), intent(in) :: coords
      integer, dimension(:,:),intent(in) :: red_int_matrix 
      double precision, dimension(size(red_int_matrix,2),3*natoms),& 
      intent(inout)::Bmatrix
      !local
      integer :: i,j,ncol

      ncol = size(red_int_matrix,2)

      do i=1,ncol
        call zero_line_i(Bmatrix,i)
        if(red_int_matrix(1,i) .ne. 0) then
          call Bmat_line(natoms,coords,red_int_matrix(1:,i),i,Bmatrix) 
          ! the columns of red_int_matrix contain the info for the given internal coordinate
          ! based on this info the appropriate line/row is filled in the B-matrix
        endif
      enddo

      end subroutine build_Wilson_Bmat
!************************************************************************

!************************************************************************
      subroutine Bmat_line(natoms,coords,column,idx,Bmatrix)
!************************************************************************
! it builds one line in Wilson's B-matrix
!************************************************************************
      implicit none
      integer, intent(in) :: natoms
      double precision, dimension(3,natoms), intent(in) :: coords
      integer, dimension(:), intent(in) :: column
      integer, intent(in) :: idx
      double precision, dimension(:,:), intent(inout)::Bmatrix
      !local
      integer :: btype,i,j,k,l,ii
      double precision, dimension(6)  :: bond_vec
      double precision, dimension(9)  :: angle_vec
      double precision, dimension(12) :: dihedral_vec

      btype = column(1) ! type of int. coord
      i     = column(2) ! atom1
      j     = column(3) ! atom2
      k     = column(4) ! atom3
      l     = column(5) ! atom4
!     write(*,*)'i,j,k,l: ',i,j,k,l

      if(btype .eq. 1 .or. btype .eq. 4) then ! bond
       bond_vec = Bmat_bond(coords(1:3,i),coords(1:3,j))
!      write(*,*)'vec: ', bond_vec,idx
!      do ii=1,3
!       Bmatrix(idx,3*(i-1)+ii) = bond_vec(ii)
!       write(*,*)'idx,i,ii: ',idx,i,ii
!       Bmatrix(idx,3*(j-1)+ii) = bond_vec(3+ii)
!      enddo
       Bmatrix(idx,3*(i-1)+1:3*(i-1)+3) = bond_vec(1:3)
       Bmatrix(idx,3*(j-1)+1:3*(j-1)+3) = bond_vec(4:6)
      elseif(column(1) .eq. 2) then  ! angle
       angle_vec = Bmat_angle(coords(1:3,i),&
                              coords(1:3,j),&
                              coords(1:3,k))
       Bmatrix(idx,3*(i-1)+1:3*(i-1)+3) = angle_vec(1:3)
       Bmatrix(idx,3*(j-1)+1:3*(j-1)+3) = angle_vec(4:6)
       Bmatrix(idx,3*(k-1)+1:3*(k-1)+3) = angle_vec(7:9)
!      write(*,*)'vec: ', angle_vec
      elseif(column(1) .eq. 3) then  ! dihedral
       dihedral_vec = Bmat_dihedral(coords(1:3,i),&
                                    coords(1:3,j),&
                                    coords(1:3,k),&
                                    coords(1:3,l))
       Bmatrix(idx,3*(i-1)+1:3*(i-1)+3) = dihedral_vec(1:3)
       Bmatrix(idx,3*(j-1)+1:3*(j-1)+3) = dihedral_vec(4:6)
       Bmatrix(idx,3*(k-1)+1:3*(k-1)+3) = dihedral_vec(7:9)
       Bmatrix(idx,3*(l-1)+1:3*(l-1)+3) = dihedral_vec(10:12)
!      write(*,*)'vec: ', dihedral_vec
       endif


      end subroutine Bmat_line
!************************************************************************
     
!************************************************************************
      subroutine zero_line_i(A,i)
      implicit none
      double precision,dimension(:,:),intent(inout) :: A
      integer, intent(in) :: i
      !local
      integer :: k,l,ndim

      ndim = size(A,2)

      do k=1,ndim
        A(i,k) = 0.d0
      enddo


      end subroutine zero_line_i
!************************************************************************

!************************************************************************
      function Bmat_bond(A, B)
!************************************************************************
! it returns the component of Wilson's B-matrix for the bonds
! between A and B
! JCP, 117, 9160 (2002), Eq. 17
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A, B
      double precision, dimension(6) :: Bmat_bond
      !local
      double precision, dimension(3) :: up,u

      up   = vec_comp(A,B) ! u'
      u    = vec_norm_comp(up)
      Bmat_bond(1:3) = -1.d0*u
      Bmat_bond(4:6) = +1.d0*u


      return

      end function Bmat_bond
!************************************************************************

!************************************************************************
      function Bmat_angle(A, B, C)
!************************************************************************
! it returns the component of Wilson's B-matrix for the bond angles
! of A-B-C; the angle is measured at vertex atom B
! i.e. the angle between bonds B-A and B-C
! JCP, 117, 9160 (2002), Eq. 25
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A, B, C
      double precision, dimension(9) :: Bmat_angle
      !local
      double precision, dimension(3) :: up,vp,u,v,wp,w,uxw,wxv
      double precision :: lu,lv


      up  = vec_comp(B,A)
      lu  = vec_length(up) ! lambda_u

      vp  = vec_comp(B,C)
      lv  = vec_length(vp) ! lambda_v

      u   = vec_norm_comp(up)
      v   = vec_norm_comp(vp)

      wp = w_prime_comp(u,v)
      w  = vec_norm_comp(wp)

      uxw = vec_cross(u,w)
      wxv = vec_cross(w,v)

      Bmat_angle(1:3) = uxw/lu 
      Bmat_angle(4:6) = (-1.d0*uxw/lu - 1.d0*wxv/lv)
      Bmat_angle(7:9) = wxv/lv


      return

      end function Bmat_angle
!************************************************************************

!************************************************************************
      function Bmat_dihedral(A, B, C, D)
!************************************************************************
! it returns the component of Wilson's B-matrix for the dihedral angles
! the angle is restricted to be in the range -Pi,Pi and measured in radian
! JCP, 117, 9160 (2002), Eq. 34
!************************************************************************
      implicit none
      double precision, dimension(3), intent(in) :: A, B, C, D
      double precision, dimension(12) :: Bmat_dihedral
      !local
      double precision, dimension(3) :: up,vp,wp, u,v,w, uxw,vxw
      double precision :: lu,lv,lw, sinu2,sinv2, cosu,cosv, su,sv,cu,cv


      up  = vec_comp(B,A)
      lu  = vec_length(up) ! lambda_u

      vp  = vec_comp(C,D)
      lv  = vec_length(vp) ! lambda_v

      wp  = vec_comp(B,C)
      lw  = vec_length(wp) ! lambda_w

      u   = vec_norm_comp(up)
      v   = vec_norm_comp(vp)
      w   = vec_norm_comp(wp)


      uxw = vec_cross(u,w)
      vxw = vec_cross(v,w)

      sinu2= (1.d0 - (vec_dot(u,w))**2)
      sinv2= (1.d0 - (vec_dot(v,w))**2)

      cosu=  1.d0*vec_dot(u,w)
      cosv= -1.d0*vec_dot(v,w)

      su  = 1.d0/(lu*sinu2)
      sv  = 1.d0/(lv*sinv2)

      cu  = cosu/(lw*sinu2)
      cv  = cosv/(lw*sinv2)

      Bmat_dihedral(1:3)   =  1.d0*uxw*su + 0.d0 +  0.d0
      Bmat_dihedral(4:6)   = -1.d0*uxw*su + 0.d0 +  1.d0*(uxw*cu+vxw*cv)
      Bmat_dihedral(7:9)   =  0.d0        + vxw*sv -1.d0*(uxw*cu+vxw*cv)
      Bmat_dihedral(10:12) =  0.d0    -1.d0*vxw*sv +0.d0


      return

      end function Bmat_dihedral
!************************************************************************

!************************************************************************
      function get_cartcoords(natoms,params)
!************************************************************************
! it extracts the Cartesian coordinates from the parameter space
! no constrains are handled at the moment (csonti)
!************************************************************************
      implicit none
      integer, intent(in) :: natoms
      double precision, dimension(:), intent(in):: params 
      double precision, dimension(3,natoms)  :: get_cartcoords
      !local
      integer :: i,j


      do j=1,natoms
        do i=1,3
          get_cartcoords(i,j)=params(3*(j-1)+i)
        enddo
      enddo

      return
      end function get_cartcoords
!************************************************************************

!*************************************************************************
      subroutine build_red_int_matrix(atnums,coords,red_int_matrix)
!************************************************************************
! it builds red_int_matrix (matrix of redundant internal coordinates)
! its format as follows:
!
!   red_int_matrix(1,i)  ! type
!   red_int_matrix(2,i)  ! first  atom
!   red_int_matrix(3,i)  ! second atom
!   red_int_matrix(4,i)  ! third  atom
!   red_int_matrix(5,i)  ! fourth atom
!
! i refers to the i^th internal coordinate
!
! atnums is an array of atomic numbers
!************************************************************************
      implicit none
      integer, dimension(:), intent(in)               :: atnums
      double precision, dimension(3,size(atnums))     :: coords
      integer, dimension(:,:),intent(out),allocatable :: red_int_matrix
      !local
      integer, dimension(size(atnums),size(atnums)) :: bond_matrix
      integer :: natoms,nb,na,nd,nintcoord

      natoms=size(atnums)

      call build_bond_matrix(atnums,coords,bond_matrix)
      nb = nint_bonds(bond_matrix)
      na = nint_angles(bond_matrix)
      nd = nint_dihedrals(bond_matrix,coords)

      nintcoord = nb + na + nd                           ! # internal coordinates
      allocate(red_int_matrix(5,nintcoord))              ! csonti allocate

      call get_bonds(bond_matrix,red_int_matrix)
      call get_angles(bond_matrix,red_int_matrix)
      call get_dihedrals(bond_matrix,coords,red_int_matrix)

!     write(*,*) 'bond_matrix:'
!     call print_matrixI(bond_matrix,9)
!     call network_comp(bond_matrix)
!     call mrccend(1)
!     write(*,*) 'red_int_matrix:'
!     call print_matrixI(red_int_matrix,12)

!     write(*,*) 'Number of internal coordinate bonds:     ', nb
!     write(*,*) 'Number of internal coordinate angles:    ', na
!     write(*,*) 'Number of internal coordinate dihedrals: ', nd

!     call print_intcoords(red_int_matrix,coords,'bond')
!     call print_intcoords(red_int_matrix,coords,'aux bond')
!     call print_intcoords(red_int_matrix,coords,'angle')
!     call print_intcoords(red_int_matrix,coords,'dihedral')

      end subroutine build_red_int_matrix
!************************************************************************

!************************************************************************
      subroutine print_intcoords(rim,coords,inttype)
!************************************************************************
! it prints the internal coordinates using the redundant internal 
! coordinate matrix (rim) and the cartesian coordinates (coords)
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in) :: rim     ! redundant internal coordinate matrix
      double precision, dimension(:,:),intent(in) :: coords  ! cartesian coordinates
      character(len=*), intent(in)       :: inttype ! type of the internal coordinate
      !local
      integer :: natoms,nintcoord,code,i,k,l,m,n,ivalue
      double precision :: fvalue

      natoms     = size(coords,2)
      nintcoord  = size(rim,2)

      if(inttype(1:4) .eq. 'bond') then
        ivalue = 1
      elseif(inttype(1:5) .eq. 'angle') then
        ivalue = 2
      elseif(inttype(1:8) .eq. 'dihedral') then
        ivalue = 3
      elseif(inttype(1:8) .eq. 'aux bond') then
        ivalue = 4
      else
        write(*,*) 'Invalid internal coordinate type was asked in &
              subroutine print_intcoords!!!'
        call mrccend(1)
      endif

      do i=1,nintcoord
        code = rim(1,i)
        k = rim(2,i)
        l = rim(3,i)
        m = rim(4,i)
        n = rim(5,i)

        if(code .gt. 0) then
          if(code .eq. ivalue) then
            if(ivalue .eq. 1) then
              fvalue = &
              distance2(coords(1:3,k),&
                        coords(1:3,l))
              write(*,101) 'bond:',k,'-',l, fvalue

            elseif(ivalue .eq. 2) then
              fvalue = &
              angle3(coords(1:3,k),&
                     coords(1:3,l),&
                     coords(1:3,m))
              fvalue = 180.d0*fvalue/PI
              write(*,102) 'angle:',k,'-',l,'-',m, fvalue

            elseif(ivalue .eq. 3) then
              fvalue = &
              dihedral4(coords(1:3,k),&
                        coords(1:3,l),&
                        coords(1:3,m),&
                        coords(1:3,n))
              fvalue = 180.d0*fvalue/PI
              write(*,103) 'dihedral:',k,'-',l,'-',m,'-',n, fvalue

            elseif(ivalue .eq. 4) then
              fvalue = &
              distance2(coords(1:3,k),&
                        coords(1:3,l))
              write(*,101) 'aux bond:',k,'-',l, fvalue

            endif
          endif
        else
          write(*,*) 'Invalid internal coordinate index in the &
            redundant internal coordinate matrix!!!'
          call mrccend(1)
        endif
      enddo
101   format(A20,I4,A4,I4,F12.6)
102   format(A20,I4,A4,I4,A4,I4,F12.6)
103   format(A20,I4,A4,I4,A4,I4,A4,I4,F12.6)
      end subroutine print_intcoords
!************************************************************************

!************************************************************************
      subroutine build_bond_matrix(atnum,coords,bond_matrix)
!************************************************************************
! it builds the bond matrix which is like the distance matrix but
! (i,j)=1 if there is a "regular" bond zero otherwise
! auxiliary bonds are indicated by the integer 4, i.e., (i,j)=4
! H-bonds will be indicated by 5 and inter-fragment bonds by 6
!************************************************************************
      implicit none
      integer, dimension(:), intent(in)                   :: atnum ! atomic/proton number
      double precision,dimension(3,size(atnum)),intent(in):: coords
      integer, dimension(:,:),intent(out)                 :: bond_matrix
      !local
      double precision d,cr,rb,ab
      integer :: i,j,natoms

      natoms = size(atnum)

      do j=1,natoms
         bond_matrix(j,j) = 0
         do i=j+1,natoms
          d  = distance2(coords(1:3,j),coords(1:3,i))
          cr = get_cov_radius(atnum(j))+get_cov_radius(atnum(i)) ! sum of covalent radii
          rb = 1.3d0*cr ! regular bond
          ab = 2.6d0*cr ! auxiliary bond !csonti
          if( d .le. rb ) then
!           write(*,*)   'Regular bond between   ', i,j,d
            bond_matrix(i,j) = 1
            bond_matrix(j,i) = 1
          else
            if (d .le. ab) then
!             write(*,*) 'Auxiliary bond between ', i,j,d
              bond_matrix(i,j) = 4
              bond_matrix(j,i) = 4
            else
!             write(*,*) 'No bond between        ', i,j,d,ab
              bond_matrix(i,j) = 0
              bond_matrix(j,i) = 0
            endif 
          endif
         enddo ! i
      enddo ! j

      return
      end subroutine build_bond_matrix
!************************************************************************

!************************************************************************
      subroutine get_bonds(bond_matrix,red_int_matrix)
!************************************************************************
! the first few columns (bond-related part) of the redundant internal 
! coordinate matrix, red_int_matrix, are set up.
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)   :: bond_matrix
      integer, dimension(:,:),intent(inout) :: red_int_matrix 
      !local
      integer :: i,j,k,natoms,nrow,ncol,code

      natoms = size(bond_matrix,1)
      nrow   = size(red_int_matrix,1)
      ncol   = size(red_int_matrix,2)

      do j=1,ncol
        do i=1,nrow
          red_int_matrix(i,j) = 0
        enddo
      enddo

      k = 0
      do j=1,natoms
         do i=j+1,natoms
          code = bond_matrix(i,j)
          if(code .eq. 1) then
            k = k+1
            red_int_matrix(1,k) = 1 ! type:  regular bond
            red_int_matrix(2,k) = j ! first  atom
            red_int_matrix(3,k) = i ! second atom
            red_int_matrix(4,k) = 0 ! third  atom
            red_int_matrix(5,k) = 0 ! fourth atom

          else
            if(code .eq. 4) then
              k = k+1
              red_int_matrix(1,k) = 4 ! type:  auxiliary bond
              red_int_matrix(2,k) = j ! first  atom
              red_int_matrix(3,k) = i ! second atom
              red_int_matrix(4,k) = 0 ! third  atom
              red_int_matrix(5,k) = 0 ! fourth atom
            !else H-bond, inter fragment bond will follow here
            ! 5 and 6
            endif 
          endif
         enddo ! i
      enddo ! j

      return
      end subroutine get_bonds
!************************************************************************

!************************************************************************
      integer function nint_bonds(bond_matrix)
!************************************************************************
! number of internal coordinate bonds is returned
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)   :: bond_matrix
      !local
      integer :: i,j,k,natoms,code

      natoms = size(bond_matrix,1)

      k = 0
      do j=1,natoms
         do i=j+1,natoms
          code = bond_matrix(i,j)
          if(code .eq. 1) then
            k = k+1
          else
            if(code .eq. 4) then
              k = k+1
            !else H-bond, inter fragment bond will follow here
            ! 5 and 6
            endif 
          endif
         enddo ! i
      enddo ! j

      nint_bonds = k

      return
      end function nint_bonds
!************************************************************************

!************************************************************************
      subroutine get_angles(bond_matrix,red_int_matrix)
!************************************************************************
! it determines the angles between bonds
! the second (angle-related) part of the redundant internal 
! coordinate matrix, red_int_matrix, is also set up.
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)    :: bond_matrix
      integer, dimension(:,:),intent(inout) :: red_int_matrix 
      !local
      integer :: i,j,natoms,ncol
      integer :: first = 1
      integer :: first_zero !first zero position in red_int_matrix

      natoms = size(bond_matrix,1)
      ncol   = size(red_int_matrix,2)

      do i=1,ncol
        if(red_int_matrix(1,i) .eq. 0) then
          first_zero = i
          exit
        endif
      enddo

      do j=1,natoms
        call rec_g_angle(first,bond_matrix(1:natoms,j),red_int_matrix,&
                         first_zero,j)
      enddo

      end subroutine get_angles
!************************************************************************

!************************************************************************
      recursive subroutine rec_g_angle(first,column,red_int_matrix,&
                                       first_zero,cidx)
!************************************************************************
      implicit none
      integer, intent(inout)                 :: first,first_zero
      integer, dimension(:), intent(in)      :: column
      integer, dimension(:,:), intent(inout) :: red_int_matrix
      integer, intent(in)                    :: cidx !column index of bond matrix
      !local
      integer :: i,k,last,nbonds
      logical :: lfirst 
      integer :: first_bond !first occasion of a bond in the given column of the bond matrix


      last = size(column)

      first_bond = 0
      nbonds     = 0
      lfirst = .True.

      do i=first,last 
        if(column(i) .eq. 1) then
          nbonds = nbonds+1
          if (lfirst) then 
            first_bond = i
            lfirst = .False.
          endif
        endif
        if(nbonds .eq. 2) exit
      enddo

      if(nbonds .eq. 2) then
        do k=first_bond+1,last
          if(column(k) .eq. 1) then
            red_int_matrix(1,first_zero) = 2          ! type:  angle
            red_int_matrix(2,first_zero) = first_bond ! first  atom
            red_int_matrix(3,first_zero) = cidx       ! second, vertex atom
            red_int_matrix(4,first_zero) = k          ! third  atom
            red_int_matrix(5,first_zero) = 0          ! fourth atom
            first_zero = first_zero+1
          endif
        enddo
        if(first_bond .lt. last) then
          first_bond = first_bond+1
          call rec_g_angle(first_bond,column,red_int_matrix,&
                           first_zero,cidx)
        endif
      endif
      
      return
      end subroutine rec_g_angle
!************************************************************************

!************************************************************************
      integer function nint_angles(bond_matrix)
!************************************************************************
! number of internal coordinate angles is returned
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)    :: bond_matrix
      !local
      integer :: j,natoms
      integer :: first = 1

      nint_angles = 0
      natoms = size(bond_matrix,1)

      nint_angles = 0
      do j=1,natoms
        call rec_n_angle(first,nint_angles,bond_matrix(1:natoms,j))
      enddo

      return
      end function nint_angles
!************************************************************************

!************************************************************************
      recursive subroutine rec_n_angle(first,nint_angles,column)
!************************************************************************
      implicit none
      integer, intent(inout)            :: first,nint_angles
      integer, dimension(:), intent(in) :: column
      !local
      integer :: i,j,k,last,nbonds
      logical :: lfirst 
      integer :: first_bond !first occasion of a bond in the given column of the bond matrix


      last = size(column)

      first_bond = 0
      nbonds     = 0
      lfirst = .True.

      do i=first,last 
        if(column(i) .eq. 1) then
          nbonds = nbonds+1
          if (lfirst) then 
            first_bond = i
            lfirst = .False.
          endif
        endif
        if(nbonds .eq. 2) exit
      enddo

      if(nbonds .eq. 2) then
        j = 0
        do k=first_bond+1,last
          if(column(k) .eq. 1) then
            j = j+1
          endif
        enddo
        nint_angles = nint_angles + j
        if(first_bond .lt. last) then
          first_bond = first_bond+1
          call rec_n_angle(first_bond,nint_angles,column)
        endif
      endif
      

      return

      end subroutine rec_n_angle
!************************************************************************

!************************************************************************
      subroutine get_angles_o(bond_matrix,red_int_matrix)
!************************************************************************
! it determines the angles between bonds
! the second (angle-related) part of the redundant internal 
! coordinate matrix, red_int_matrix, is also set up.
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)    :: bond_matrix
      integer, dimension(:,:),intent(inout) :: red_int_matrix 
      !local
      integer :: i,j,k,natoms,nbonds,ncol
      logical :: lfirst 
      integer :: first_bond !first occasion of a bond in the given column of the bond matrix
      integer :: first_zero !first zero position in red_int_matrix

      natoms = size(bond_matrix,1)
      ncol   = size(red_int_matrix,2)

      do i=1,ncol
        if(red_int_matrix(1,i) .eq. 0) then
          first_zero = i
          exit
        endif
      enddo

      first_bond=0
      do j=1,natoms
        nbonds = 0
        lfirst = .True.
        do i=1,natoms 
          if(bond_matrix(i,j) .eq. 1) then
            nbonds = nbonds+1
            if (lfirst) then 
              first_bond=i
              lfirst = .False.
            endif
          endif
          if(nbonds .eq. 2) exit
        enddo
        if(nbonds .eq. 2) then
          do k=first_bond+1,natoms                        ! csonti implement a recursive algorithm here
            if(bond_matrix(k,j) .eq. 1) then
              red_int_matrix(1,first_zero) = 2          ! type:  angle
              red_int_matrix(2,first_zero) = first_bond ! first  atom
              red_int_matrix(3,first_zero) = j          ! second, vertex atom
              red_int_matrix(4,first_zero) = k          ! third  atom
              red_int_matrix(5,first_zero) = 0          ! fourth atom
              first_zero = first_zero+1
            endif
          enddo
        endif
      enddo

      end subroutine get_angles_o
!************************************************************************

!************************************************************************
      integer function nint_angles_o(bond_matrix)
!************************************************************************
! number of internal coordinate angles is returned
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)    :: bond_matrix
      !local
      integer :: i,j,k,l,natoms,nbonds
      logical :: lfirst 
      integer :: first_bond !first occasion of a bond in the given column of the bond matrix
      integer :: first_zero !first zero position in red_int_matrix


      nint_angles_o = 0
      natoms = size(bond_matrix,1)

      first_bond=0
      do j=1,natoms
        nbonds = 0
        lfirst = .True.
        do i=1,natoms 
          if(bond_matrix(i,j) .eq. 1) then
            nbonds = nbonds+1
            if (lfirst) then 
              first_bond=i
              lfirst = .False.
            endif
          endif
          if(nbonds .eq. 2) exit
        enddo
        if(nbonds .eq. 2) then
          l = 0
          do k=first_bond+1,natoms                        ! csonti implement a recursive algorithm here
            if(bond_matrix(k,j) .eq. 1) then
              l = l+1
            endif
          enddo
          nint_angles_o = nint_angles_o + l
        endif
      enddo
      
      
      return
      end function nint_angles_o
!************************************************************************

!************************************************************************
      subroutine get_dihedrals(bond_matrix,coords,red_int_matrix)
!************************************************************************
! it determines the dihedral angles between the first and last bonds of 
! three consecutive bonds
! the third (dihedral-related) part of the redundant internal 
! coordinate matrix, red_int_matrix, is also set up.
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)    :: bond_matrix
      double precision,dimension(3,size(bond_matrix,1)),intent(in):: coords
      integer, dimension(:,:),intent(inout) :: red_int_matrix 
      !local
      integer :: i,j,k,l,natoms,nrow,ncol,mloc,mvalue,nb
      logical :: lfirst 
      integer :: first_zero !first zero position in red_int_matrix
      logical :: no_regular !are there any regular (A-B-C-D type) dihedral
      integer, dimension(:), allocatable :: a_idx
      integer :: idxi,idxj,idxk,idxl
      logical :: bool1, bool2

      natoms = size(bond_matrix,1)
      nrow   = size(red_int_matrix,1)
      ncol   = size(red_int_matrix,2)

      do i=1,ncol
        if(red_int_matrix(1,i) .eq. 0) then
          first_zero = i
          exit
        endif
      enddo

      natoms = size(bond_matrix,1)
      no_regular = .True.

      do i=1,natoms
        do j=1,natoms 
          if(bond_matrix(i,j) .eq. 1) then
            do k=1,natoms
              if((bond_matrix(j,k) .eq. 1) .and. (k .ne. i) ) then
                do l=i,natoms
                  if((bond_matrix(k,l) .eq. 1) .and. &
                    ( (l .ne. j) .and. (l .ne. i) ) ) then
                      bool1 = is_parallel(coords(1:3,i),&
                                    coords(1:3,j),coords(1:3,k))
                      bool2 = is_parallel(coords(1:3,j),&
                                    coords(1:3,k),coords(1:3,l))
                      if((.not. bool1) .and. (.not. bool2)) then ! none of the ends of the dihedral coincides with a line (0 or 180 degree bond angle)
                        red_int_matrix(1,first_zero) = 3          ! type:  dihedral
                        red_int_matrix(2,first_zero) = i          ! first  atom
                        red_int_matrix(3,first_zero) = j          ! second, vertex atom
                        red_int_matrix(4,first_zero) = k          ! third  atom
                        red_int_matrix(5,first_zero) = l          ! fourth atom
                        first_zero = first_zero + 1
                        no_regular = .False.                      ! regular dihedral angles are present
                      endif
                  endif
                enddo !l
              endif
            enddo !k
          endif
        enddo !j
      enddo !i

      ! auxiliary dihedral angles where they need; for molecules like BH3, NH3, and CH4

      mloc   = 1 ! location of the atom with maximum coordination number
      mvalue = 0 ! number of atoms that are coordinated

      if(natoms .ge. 4 .and. no_regular) then 
        do j=1,natoms
          nb = 0
          do i=1,natoms
            if(bond_matrix(i,j) .eq. 1) then
              nb = nb+1
            endif
          enddo
          if(nb .gt. mvalue) then
            mvalue = nb
            mloc   = j
          endif
        enddo
      
        if(mvalue .gt. 0) then
          allocate(a_idx(mvalue))
          j = 0
          do i=1,natoms
            if(bond_matrix(i,mloc) .eq. 1) then
              j = j+1
              a_idx(j) = i
            endif
          enddo


          idxj = mloc    ! i-j-k-l the central atom is always the second in the definition of the dihedral
          do i=1,mvalue  ! the cyclic permutation goes here among the atoms bonded to the central atom
            idxi = i
            idxk = mod(i+1,mvalue)
            idxl = mod(i+2,mvalue)
            if(idxk .eq. 0) idxk = mvalue
            if(idxl .eq. 0) idxl = mvalue
            bool1 = is_parallel(coords(1:3,a_idx(idxi)),&
                                coords(1:3,idxj),&
                                coords(1:3,a_idx(idxk)))
            bool2 = is_parallel(coords(1:3,idxj),&
                                coords(1:3,a_idx(idxk)),&
                                coords(1:3,a_idx(idxl)))
            if((.not. bool1) .and. (.not. bool2)) then ! none of the ends of the dihedral coincides with a line (0 or 180 degree bond angle)

               red_int_matrix(1,first_zero) = 3           ! type:  dihedral
               red_int_matrix(2,first_zero) = a_idx(idxi) ! first  atom
               red_int_matrix(3,first_zero) =       idxj  ! second, vertex atom
               red_int_matrix(4,first_zero) = a_idx(idxk) ! third  atom
               red_int_matrix(5,first_zero) = a_idx(idxl) ! fourth atom
               first_zero = first_zero + 1
            endif
 
          enddo

          if(allocated(a_idx)) deallocate(a_idx)
        endif

      endif
      ! end for auxiliary dihedrals 

      return
      end subroutine get_dihedrals
!************************************************************************

!************************************************************************
      integer function nint_dihedrals(bond_matrix,coords)
!************************************************************************
! number of internal coordinate dihedrals is returned
!************************************************************************
      implicit none
      integer, dimension(:,:),intent(in)    :: bond_matrix
!     integer, dimension(:), intent(in)                   :: atnum ! atomic/proton number
      double precision,dimension(3,size(bond_matrix,1)),intent(in):: coords

      !local
      integer :: i,j,k,l,natoms,mloc,mvalue,nb,n_linear
      integer, dimension(:), allocatable :: a_idx
      integer :: idxi,idxj,idxk,idxl
      logical :: bool1, bool2


      natoms = size(bond_matrix,1)
      nint_dihedrals = 0
      n_linear       = 0

      do i=1,natoms
        do j=1,natoms 
          if(bond_matrix(i,j) .eq. 1) then
            do k=1,natoms
              if((bond_matrix(j,k) .eq. 1) .and. (k .ne. i) ) then
                do l=i,natoms
                  if((bond_matrix(k,l) .eq. 1) .and. &
                    ( (l .ne. j) .and. (l .ne. i) ) ) then
                      nint_dihedrals = nint_dihedrals + 1
                      bool1 = is_parallel(coords(1:3,i),&
                                    coords(1:3,j),coords(1:3,k))
                      bool2 = is_parallel(coords(1:3,j),&
                                    coords(1:3,k),coords(1:3,l))
                      if(bool1 .or. bool2) then ! at least one end of the dihedral coincides with a line (0 or 180 degree bond angle)
                        n_linear = n_linear + 1  
                        nint_dihedrals = nint_dihedrals - 1
                      endif
                  endif
                enddo !l
              endif
            enddo !k
          endif
        enddo !j
      enddo !i
102   format(A20,I4,A4,I4,A4,I4,F12.6)      

      mloc   = 1 ! location of the atom with maximum coordination number
      mvalue = 0 ! number of atoms that are coordinated

!     if(natoms .ge. 4 .and. (nint_dihedrals+n_linear .eq. 0) ) then 
      if(natoms .ge. 4 .and. (nint_dihedrals .eq. 0) ) then 
        do j=1,natoms
          nb = 0
          do i=1,natoms
            if(bond_matrix(i,j) .eq. 1) then
              nb = nb+1
            endif
          enddo
          if(nb .gt. mvalue) then
            mvalue = nb
            mloc   = j
          endif
        enddo

        if(mvalue .gt. 0) then
          allocate(a_idx(mvalue))
          j = 0
          do i=1,natoms
            if(bond_matrix(i,mloc) .eq. 1) then
              j = j+1
              a_idx(j) = i
            endif
          enddo


          idxj = mloc    ! i-j-k-l the central atom is always the second in the definition of the dihedral
          do i=1,mvalue  ! the cyclic permutation goes here among the atoms bonded to the central atom
            idxi = i
            idxk = mod(i+1,mvalue)
            idxl = mod(i+2,mvalue)
            if(idxk .eq. 0) idxk = mvalue
            if(idxl .eq. 0) idxl = mvalue
            bool1 = is_parallel(coords(1:3,a_idx(idxi)),&
                                coords(1:3,idxj),&
                                coords(1:3,a_idx(idxk)))
            bool2 = is_parallel(coords(1:3,idxj),&
                                coords(1:3,a_idx(idxk)),&
                                coords(1:3,a_idx(idxl)))
            if((.not. bool1) .and. (.not. bool2)) then ! none of the ends of the dihedral coincides with a line (0 or 180 degree bond angle)
               nint_dihedrals=nint_dihedrals+1
            endif
          enddo

          if(allocated(a_idx)) deallocate(a_idx)
        endif

      endif

      return
      end function nint_dihedrals
!************************************************************************

!************************************************************************
      function get_intcoords(natoms,cartcoords,nintcoord,red_int_matrix)
!************************************************************************
! it returns the values of the internal coordinates 
! using the matrix of redundant internal coordinates and 
! the Cartesian coordinates
! no constrains are handled at the moment (csonti)
!************************************************************************
      implicit none
      integer, intent(in) :: natoms
      double precision, dimension(3,natoms), intent(in):: cartcoords 
      integer, intent(in) :: nintcoord
      integer, dimension(:,:),intent(in) :: red_int_matrix
      double precision, dimension(nintcoord)  :: get_intcoords
      !local
      integer :: i,k,l,m,n,nrow,ncol,rim

      nrow = size(red_int_matrix,1)
      ncol = size(red_int_matrix,2)

      do i=1,ncol
        rim = red_int_matrix(1,i) ! type of the internal coordinate
        if (rim .eq. 0) then      ! last element
          exit
        elseif(rim .eq. 1) then   ! bond
          k = red_int_matrix(2,i)
          l = red_int_matrix(3,i)
!         write(*,*) 'distance',k,l,&
!         distance2(cartcoords(1:3,k),cartcoords(1:3,l))
          get_intcoords(i) = &
               distance2(cartcoords(1:3,k),cartcoords(1:3,l))
        elseif(rim .eq. 2) then   ! angle
          k = red_int_matrix(2,i)
          l = red_int_matrix(3,i)
          m = red_int_matrix(4,i)
!         write(*,*) 'angle',k,l,m,&
!              angle3(cartcoords(1:3,k),&
!                     cartcoords(1:3,l),&
!                     cartcoords(1:3,m))
          get_intcoords(i) = &
               angle3(cartcoords(1:3,k),&
                      cartcoords(1:3,l),&
                      cartcoords(1:3,m))
        elseif(rim .eq. 3) then   ! dihedral
          k = red_int_matrix(2,i)
          l = red_int_matrix(3,i)
          m = red_int_matrix(4,i)
          n = red_int_matrix(5,i)
!         write(*,*) 'dihedral',k,l,m,n,&
!              dihedral4(cartcoords(1:3,k),&
!                        cartcoords(1:3,l),&
!                        cartcoords(1:3,m),&
!                        cartcoords(1:3,n))
          get_intcoords(i) = &
               dihedral4(cartcoords(1:3,k),&
                         cartcoords(1:3,l),&
                         cartcoords(1:3,m),&
                         cartcoords(1:3,n))
        elseif(rim .eq. 4) then   ! auxiliary bond
          k = red_int_matrix(2,i)
          l = red_int_matrix(3,i)
          get_intcoords(i) = &
               distance2(cartcoords(1:3,k),cartcoords(1:3,l))
        endif
      enddo

      return
      end function get_intcoords
!************************************************************************

!************************************************************************
      subroutine read_intcoords(nintcoord,intcoords)
!************************************************************************
! it reads values of the internal coordinates from a scrfile
!************************************************************************
      implicit none
      integer, intent(in) :: nintcoord
      double precision, dimension(nintcoord), intent(out) :: intcoords
      !local
      integer :: i

      open(icoordfile,file='COORD.int',&
                    status='old',form='unformatted')
      do i=1,nintcoord
        read(icoordfile) intcoords(i)
      enddo
      close(icoordfile)

      end subroutine read_intcoords
!************************************************************************

!************************************************************************
      subroutine wrt_intcoords(nintcoord,intcoords)
!************************************************************************
! it writes values of the internal coordinates into a scrfile
!************************************************************************
      implicit none
      integer, intent(in) :: nintcoord
      double precision, dimension(nintcoord), intent(in) :: intcoords
      !local
      integer :: i

      open(icoordfile,file='COORD.int',&
                    status='unknown',form='unformatted')
      do i=1,nintcoord
        write(icoordfile) intcoords(i)
      enddo
      close(icoordfile)

      end subroutine wrt_intcoords
!************************************************************************

!************************************************************************
      subroutine wrt_Pmat(nintcoord,Pmat)
!************************************************************************
! it writes the projection matrix into a scrfile
!************************************************************************
      implicit none
      integer, intent(in) :: nintcoord
      double precision,dimension(nintcoord,nintcoord),intent(in) :: Pmat
      !local
      integer :: i,j

      open(pmatfile,file='PMAT.int',&
                    status='unknown',form='unformatted')
      do j=1,nintcoord
        do i=1,nintcoord
          write(pmatfile) Pmat(i,j)
        enddo
      enddo
      close(pmatfile)

      end subroutine wrt_Pmat
!************************************************************************

!************************************************************************
      subroutine read_Pmat(nintcoord,Pmat)
!************************************************************************
! it reads the projection matrix from a scrfile
!************************************************************************
      implicit none
      integer, intent(in) :: nintcoord
      double precision,dimension(nintcoord,nintcoord),intent(out):: Pmat
      !local
      integer :: i,j

      open(pmatfile,file='PMAT.int',&
                    status='old',form='unformatted')
      do j=1,nintcoord
        do i=1,nintcoord
          read(pmatfile) Pmat(i,j)
        enddo
      enddo
      close(pmatfile)

      end subroutine read_Pmat
!************************************************************************

!************************************************************************
      subroutine step_transform(x0,q0,B,sq,xk,&
                                natoms,nintcoord,red_int_matrix)
!************************************************************************
! transformation of the projected internal coordinate step sq 
! to Cartesian coordinates; JCP, 117, 9160 (2002), section 5
!
! x0  - initial Cartesian coordinate vector
! q0  - initial internal coordinate vector
! B   - Wilson's B-matrix
! Bp  - pseudo-inverse /B+/ of Wilson's B-matrix
! sq  - internal coordinate step
!
!************************************************************************
      implicit none
      double precision, dimension(:), intent(in) :: x0
      double precision, dimension(:), intent(in) :: q0
      double precision, dimension(size(q0),size(x0)), intent(in) :: B
      double precision, dimension(size(q0)), intent(in) :: sq
      double precision, dimension(size(x0)), intent(out) :: xk
      integer, intent(in)                :: natoms,nintcoord
      integer, dimension(:,:),intent(in) :: red_int_matrix
      !local
      double precision, dimension(size(x0),size(q0)) :: Bp
      double precision, dimension(size(x0)) :: xk1,x1
      double precision, dimension(size(sq)) :: qk,dqk,dqk1,dq
      double precision, parameter :: eps = 1.d-6 ! JCP, 117, 9160 (2002), section 5
      integer :: k,i
      logical :: conv

      conv= .False.
      k   = 0
      call matrix_pseudoinverse(B,Bp)
!     write(*,*)'Bp from step_transform' 
!     call print_matrixR(Bp,10)
!     write(*,*)'B from step_transform'
!     call print_matrixR(B,10)
      x1 = x0 + matrix_dot_vec(Bp,sq)
      xk = x1 ! initial estimate

      do while ( (.not.conv) .and. k .le. 25)
        k  = k+1
!       write(*,*) 'xk'
!       call print_vectorR(xk,10)
!       write(*,*)'red_int_matrix'
!       call print_matrixR(red_int_matrix,10)

        qk = get_intcoords(natoms,xk,nintcoord,red_int_matrix)

        do i=1,nintcoord
         if(red_int_matrix(1,i) .eq. 2 &
                .or. &
            red_int_matrix(1,i) .eq. 3) then        ! it handles the periodicity of angles and dihedrals
            qk(i) = mod(qk(i),2.0d0*PI)
         endif
        enddo

!       write(*,*) 'qk'
!       call print_vectorR(qk,10)
!       write(*,*) 'q0'
!       call print_vectorR(q0,10)
!       write(*,*) 'x1'
!       call print_vectorR(x1,10)

        dq = qk-q0

        do i=1,nintcoord
         if(red_int_matrix(1,i) .eq. 3 &
            .and. &
            dabs(dq(i)) .gt. PI) then        ! fix the dihedrals around -180 or 180 -> the difference can be large about 2*PI
              if(dq(i) .gt. 0.d0) then
                dq(i) = dq(i) + 2.0d0*PI
              else
                dq(i) = dq(i) - 2.0d0*PI
              endif
         endif
         if(red_int_matrix(1,i) .eq. 2 &
                .or. &
            red_int_matrix(1,i) .eq. 3) then        ! it handles the periodicity of angles and dihedrals
            qk(i) = mod(qk(i),2.0d0*PI)
         endif
        enddo

        dqk  = sq - (dq)
!       dqk  = sq - (qk-q0)

        if(k .eq. 1) then 
          dqk1  = dqk
        else
          if(vecnorm(dqk,2) .gt. vecnorm(dqk1,2)) then
            xk = x1
            exit
          endif
        endif

!       write(*,*) 'dqk1'
!       call print_vectorR(dqk1,10)

        xk1  = xk + matrix_dot_vec(Bp,dqk)
!       write(*,'(A,I4,F12.8)') &
!           'Iteration number (backtransform) : ',k,vecnorm(dqk,2)

        conv = check_conv(xk1,xk,eps)
        xk   = xk1

      enddo

      end subroutine step_transform
!************************************************************************

!************************************************************************
      logical function check_conv(xk,x,eps)
!************************************************************************
! it returns .True. if xk is "close" to x;
! the rmsd of xk from x is less than eps
!************************************************************************
      implicit none
      double precision, dimension(:), intent(in) :: xk,x
      double precision, intent(in) :: eps
      !local
      double precision :: s
      integer :: i,ndim

      check_conv = .False.
      ndim=size(xk)

      if(ndim .ne. size(x)) then
        write(*,*) 'Error in check_conv!!!'
        return
      endif

      s = 0.d0
      do i=1,ndim
        s = s + dabs( xk(i) - x(i) )**2
      enddo

      s = dsqrt(s/ndim)

      if(s .le. eps) check_conv = .True.

      return

      end function check_conv
!************************************************************************


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! STRING HANDLING
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      integer function integer_(str)
!************************************************************************
! it returns an integer from string (str)
!************************************************************************
      implicit none
      character(len=*), intent(in) :: str
      !local 
      character(len=len(str)) :: stmp
      integer :: code ! ASCII
      integer :: i
      
      stmp = trim(adjustl(str))

      do i=1,len_trim(stmp)
        code = ichar(stmp(i:i))
        if(code .lt. 48 .or. code .gt. 57) then
          write(*,*) 'Non numeric character in ', str,code
          write(*,*) 'invalid conversion to integer'
          call mrccend(1)
        endif
      enddo

      read(str,*) integer_
      end function integer_
!************************************************************************

!************************************************************************
      function int2char(i)
!************************************************************************
! it converts an integer to a string
!************************************************************************
      implicit none
      integer,intent(in) :: i
      character(len=512)   :: int2char
      !local 
      character(len=512) :: stmp
      character(len=512) :: fmts
      
      fmts = '(I0)' 
      write(stmp,fmts) i
      int2char=trim(adjustl(stmp))

      end function int2char
!************************************************************************

!************************************************************************
      double precision function double_(str)
!************************************************************************
! it returns a double from string (str)
!************************************************************************
      implicit none
      character(len=*), intent(in) :: str

      read(str,*) double_
      end function double_
!************************************************************************

!************************************************************************
      subroutine str_srch(str,substr,pos)
!************************************************************************
! it search for the position (pos) of the substring (substr) in the 
! string (str) 
! if the string contains the substring it returns the position of its 
! first character, otherwise returns 0
!************************************************************************
      implicit none
      character(len=*),intent(in) :: str
      character(len=*),intent(in) :: substr
      integer,intent(out) :: pos
      !local
      integer :: lstr,lsub,i
      logical :: bool

      lstr = len(str)
      lsub = len(substr)

      if(lstr.lt.lsub) then
        write(*,*) 'The substring is longer than the string!!!'
        return
      endif

      bool=.True.
      pos=0
      i=1
      do while (i .le. lstr-lsub+1 .and. bool)
        if(str(i:i+lsub-1) .eq. substr(1:lsub)) then
          pos=i
          bool = .False.
        endif
        i=i+1
      enddo

!     strb=adjustl(trim(strb))
!     stra=adjustl(trim(stra))
!     write(*,*) 'Characters before ',ch,' ',strb
!     write(*,*) 'Characters after  ',ch,' ',stra
      return
      end subroutine str_srch
!************************************************************************

!************************************************************************
      subroutine str_cut(str,ch,strb,stra)
!************************************************************************
! it cuts the string at first occurence of character ch 
! then returns strb (part before character ch) and 
! stra (part after character ch)
!************************************************************************
      implicit none
      character(len=*),intent(in) :: str
      character(len=1),intent(in) :: ch
      character(len=*),intent(out) :: strb,stra
      !local
      integer :: i,length
      character(len=len(str)) :: stmp
      
      length = len(stmp)
      strb=''
      stra=''
      i=1
      do while (i.le.length .and. str(i:i).ne.ch)
        strb(i:i)=str(i:i)
        i=i+1
      enddo
      stra(:)=str(i+1:)
!     strb=adjustl(trim(strb))
!     stra=adjustl(trim(stra))
!     write(*,*) 'Characters before ',ch,' ',strb
!     write(*,*) 'Characters after  ',ch,' ',stra
      return
      end subroutine str_cut
!************************************************************************

!************************************************************************
      subroutine str_del(str,pos,len)
!************************************************************************
! it deletes 'len' characters from position 'pos' from the string (str) 
!************************************************************************
      implicit none
      character(len=*), intent(inout) :: str
      integer, intent(in) :: pos,len

      str(pos:) = str(pos+len:)
      end subroutine str_del
!************************************************************************

!************************************************************************
      subroutine str_dtest(str,double,bool,sm)
!************************************************************************
! it converts the string (str) to double (double) 
! using the function double_(str)
! and also checks for special marks (bool) and returns them in sm
!************************************************************************
      implicit none
      character(len=*), intent(in) :: str
      double precision, intent(out) :: double
      logical, intent(inout) :: bool
      character(len=2), intent(inout) :: sm

      character(len=len(trim(adjustl(str)))) :: stmp

      double = 0.d0
      sm   = '  ' ! special mark i.e., ++ or --; default return value is '  '
      stmp = trim(adjustl(str))

      if(index(stmp,'++').gt.0) then
        bool = .True.
        sm   = '++'
        call str_del(stmp,index(stmp,'++'),2)
        double = double_(stmp)
      else
        if(index(stmp,'--').gt.0) then
          bool = .False.
          sm   = '--'
          call str_del(stmp,index(stmp,'--'),2)
          double = double_(stmp)
        else
          double = double_(stmp)
        endif
      endif
      end subroutine str_dtest
!************************************************************************

!************************************************************************
      subroutine str_itest(str,int_,bool,sm)
!************************************************************************
! it converts the string (str) to integer (int_) 
! using the function integer_(str)
! and also checks for special marks (bool) and returns them in sm
!************************************************************************
      implicit none
      character(len=*), intent(in) :: str
      integer, intent(out) :: int_
      logical, intent(inout) :: bool
      character(len=2), intent(inout) :: sm

      character(len=len(trim(adjustl(str)))) :: stmp

      int_ = 0
      sm   = '  ' ! special mark i.e., ++ or --; default return value is '  '
      stmp = trim(adjustl(str))

      if(index(stmp,'++').gt.0) then
        bool = .True.
        sm   = '++'
        call str_del(stmp,index(stmp,'++'),2)
        int_ = integer_(stmp)
      else
        if(index(stmp,'--').gt.0) then
          bool = .False.
          sm   = '--'
          call str_del(stmp,index(stmp,'--'),2)
          int_ = integer_(stmp)
        else
          int_ = integer_(stmp)
        endif
      endif
      end subroutine str_itest
!************************************************************************


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! PERIODIC TABLE RELATED
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      integer function get_atomic_number(atsymbol)
!************************************************************************
! Get atomic number from atomic symbol or from atomic number :)
! /if atomic number is given as a string(atsymbol) instead of an integer/
! this seemed practical during file processing
!************************************************************************
      character(len=*), intent(in) :: atsymbol
      !local
      integer, parameter  :: natmax = 118
      character(len=3)    :: chl3
      character(len=2)    :: atsym(0:natmax),chl2a,chl2b
      integer :: length,atnum,code
      logical :: bool
      data atsym /'Bf',&
                  'H ','He','Li','Be','B ','C ','N ','O ','F ','Ne',&
                  'Na','Mg','Al','Si','P ','S ','Cl','Ar','K ','Ca',&
                  'Sc','Ti','V ','Cr','Mn','Fe','Co','Ni','Cu','Zn',&
                  'Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y ','Zr',&
                  'Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In','Sn',&
                  'Sb','Te','I ','Xe','Cs','Ba','La','Ce','Pr','Nd',&
                  'Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb',&
                  'Lu','Hf','Ta','W ','Re','Os','Ir','Pt','Au','Hg',&
                  'Tl','Pb','Bi','Po','At','Rn','Fr','Ra','Ac','Th',&
                  'Pa','U ','Np','Pu','Am','Cm','Bk','Cf','Es','Fm',&
                  'Md','No','Lr','Rf','Db','Sg','Bh','Hs','Mt','Ds',&
                  'Rg','Cn','Nh','Fl','Mc','Lv','Ts','Og'/
!
      length = len_trim(atsymbol)
      
      if(length .gt. 3 .or. length .lt. 1) then
        write(*,*) 'Unknown atomic symbol0 '//trim(atsymbol)//'!'
        call mrccend(1)
      endif

      chl3   = trim(adjustl(atsymbol)) ! remove blanks
      code   = ichar(chl3(1:1))
      if(code .gt. 47 .and. code .lt. 58) then                       ! atomic number in atsymbol
        atnum = integer_(chl3)
        if(atnum .gt. natmax .or. atnum .lt. 0) then
            write(*,*) 'Unknown atomic symbol1 '//trim(atsymbol)//'!'
            call mrccend(1)
        else
          get_atomic_number = atnum
        endif
      else                                                           ! atomic symbol
        if(length .ne. 3) then
          chl2a = trim(adjustl(atsymbol)) ! remove blanks

          call lowercase(chl2a,chl2a,2)
          bool = .False.
          do i=1,natmax
            call lowercase(atsym(i),chl2b,2)
            if(chl2a .eq. chl2b) then
               get_atomic_number = i
               bool = .True.
            endif
          enddo

          if(.not. bool) then
            write(*,*) 'Unknown atomic symbol2 '//trim(atsymbol)//'!'
            call mrccend(1)
          endif
        else 
          write(*,*) 'Unknown atomic symbol3 '//trim(atsymbol)//'!'
          call mrccend(1)
        endif
      endif
      
      return
      end function get_atomic_number
!************************************************************************

!************************************************************************
      double precision function get_cov_radius(atnum)
!************************************************************************
! it returns the covalent radius of an atom
! Values are extracted from Chemistry A European Journal, 15, 186 (2009)
!************************************************************************
      implicit none
      integer, intent(in) :: atnum
      !local
      integer, parameter :: maxnum = 118
      double precision, dimension(maxnum) :: radii
      integer :: i
      data (radii(i), i = 1, maxnum)&
      /&
        32,                                46,                      & ! H  -> He
       133, 102,  85,  75,  71,  63,  64,  67,                      & ! Li -> Ne
       155, 139, 126, 116, 111, 103,  99,  96,                      & ! Na -> Ar
       196, 171,                                                    & ! K  -> Ca
       148, 136, 134, 122, 119, 116, 111, 110, 112, 118,            & ! Sc -> Zn Transition Metals 
                 124, 121, 121, 116, 114, 117,                      & ! Ga -> Kr
       210, 185,                                                    & ! Rb -> Sr
       163, 154, 147, 138, 128, 125, 125, 120, 128, 136,            & ! Y  -> Cd Transition Metals
                 142, 140, 140, 136, 133, 131,                      & ! In -> Xe
       232, 196,                                                    & ! Cs -> Ba
       180,163,176,174,173,172,168,169,168,167,166,165,164,170,162, & ! La -> Lu Rare Earth Elements
            152, 146, 137, 131, 129, 122, 123, 124, 133,            & ! Hf -> Hg Transition Metals 
                 144, 144, 151, 145, 147, 142,                      & ! Tl -> Rn!
       223, 201,                                                    & ! Fr -> Ra
       186,175,169,170,171,172,166,166,168,168,165,167,173,176,161, & ! Ac -> Lr Rare Earth Elements
            157, 149, 143, 141, 134, 129, 128, 121, 122,            & ! Rf -> Cn Transition Metals 
                 136 ,143 ,162 ,175 ,165 ,157                       & ! Nh -> Og [Pyykko & Atsumi;2009;10.1002/chem.200800987]
      /
!
      if (atnum .lt. 1 .or. atnum .gt. maxnum) then
        write(*,*) 'Not valid atomic number: ', atnum
        write(*,*) 'Allowed atomic numbers are between 1 and ', maxnum
      else
        get_cov_radius = radii(atnum)/100.d0            ! pm to Angstrom
        get_cov_radius = get_cov_radius*ang2bohr      ! Angstrom to Bohr
      endif

      end function get_cov_radius
!************************************************************************

!************************************************************************
      double precision function get_atomic_mass(atnum)
!************************************************************************
! it returns the atomic mass (values were copied from integ.f; 2015-09-15)
!************************************************************************
      implicit none
      integer, intent(in) :: atnum
      !local
      integer, parameter :: maxnum = 118+1
      double precision, dimension(0:maxnum-1) :: atmas
      integer :: i 
      data (atmas(i), i=0, maxnum-1) &
      / &
                   0.d0,&
                   1.007825035d+00,  4.00260324d+00,   7.0160030d+00 , &
                     9.0121822d+00,  11.0093054d+00,  12.0000000d+00 , &
                  14.003074002d+00, 15.99491463d+00, 18.99840322d+00 , &
                    19.9924356d+00,  22.9897677d+00,  23.9850423d+00 , &
                    26.9815386d+00,  27.9769271d+00,  30.9737620d+00 , &
                   31.97207070d+00,34.968852721d+00,  39.9623837d+00 , &
                    38.9637074d+00,  39.9625906d+00,  44.9559100d+00 , &
                    47.9479473d+00,  50.9439617d+00,  51.9405098d+00 , &
                    54.9380471d+00,  55.9349393d+00,  58.9331976d+00 , &
                    57.9353462d+00,  62.9295989d+00,  63.9291448d+00 , &
                     68.925580d+00,  73.9211774d+00,  74.9215942d+00 , &
                    79.9165196d+00,  78.9183361d+00,   83.911507d+00 , &
                     84.911794d+00,  87.9056188d+00,   88.905849d+00 , &
                    89.9047026d+00,  92.9063772d+00,  97.9054073d+00 , &
                     97.907215d+00, 101.9043485d+00,  102.905500d+00 , &
                    105.903478d+00,  106.905092d+00,  113.903357d+00 , &
                    114.903882d+00, 119.9021991d+00,  120.9038212d+00, &
                    129.906229d+00,  126.904473d+00,  131.904144d+00 , &
                    132.905429d+00,  137.905232d+00,  138.906347d+00 , &
                    139.905433d+00,  140.907647d+00,  141.907719d+00 , &
                    144.912743d+00,  151.919728d+00,  152.921225d+00 , &
                    157.924019d+00,  158.925342d+00,  163.929171d+00 , &
                    164.930319d+00,  165.930290d+00,  168.934212d+00 , &
                    173.938859d+00,  174.940770d+00, 179.9465457d+00 , &
                    180.947462d+00,  183.950928d+00,  186.955744d+00 , &
                    191.961467d+00,  192.962917d+00,  194.964766d+00 , &
                    196.966543d+00,  201.970617d+00,  204.974401d+00 , &
                    207.976627d+00,  208.980374d+00,  208.982404d+00 , &
                    209.987126d+00,  222.017571d+00,  223.019733d+00 , &
                    226.025403d+00,  227.027750d+00,  232.038051d+00 , &
                    231.035880d+00,  238.050785d+00,  237.048168d+00 , &
                    244.064199d+00,  243.061373d+00,  247.070347d+00 , &
                    247.070300d+00,  251.079580d+00,  252.082944d+00 , &
                    257.095099d+00,  258.098570d+00,  259.100931d+00 , &
                    260.105320d+00,  261.108690d+00,  262.113760d+00 , &
                    263.118220d+00,  262.122930d+00,  269.134100d+00 , &
                    267.138000d+00,  268.143500d+00,  272.000000d+00 , &
                    277.000000d+00,  283.000000d+00,  289.000000d+00 , &
                    294.000000d+00,  300.000000d+00,  306.000000d+00 , &
                    310.000000d+00 /

      if (atnum .lt. 0 .or. atnum .gt. maxnum-1) then
        write(*,*) 'Not valid atomic number: ', atnum
        write(*,*) 'Allowed atomic numbers are between 0 and ', maxnum-1
      else
        get_atomic_mass = atmas(atnum)
      endif

      end function get_atomic_mass
!************************************************************************


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Graph
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      subroutine breadth_first(A,v)
!************************************************************************
! The breadth-first search algorithm
! A is the adjacensy matrix, v is the initial vertex
!************************************************************************
      implicit none
      integer, dimension(:,:), intent(in) :: A
      integer, intent(in) :: v
      !local
      integer, dimension (:),   allocatable :: stk
      integer, dimension (:),   allocatable :: Q,color,distance,parent
      integer :: nnodes,s,r,w,d,n,i
      
      nnodes = size(A,1)

      allocate(stk(nnodes))
      allocate(Q(nnodes))

      do i=1,nnodes
        color(i)    =  0 ! white
        distance(i) = -1 ! infinity
        parent(i)   =  0 ! nil
      enddo

      color(v)    = 1 ! gray
      distance(v) = 0
      parent(v)   = 0

      do s = 1,nnodes
        stk = 0
        stk(1) = s
        r = 1
        w = 2

      end do
!     write(*,*)'SPLMAT:'
!     call print_matrixI(splmat,10)

      contains

      subroutine enqueue(queue,element)
      implicit none
      integer,dimension(:),intent(inout) :: queue
      integer,intent(in) :: element



      end subroutine enqueue

      end subroutine breadth_first
!************************************************************************

!************************************************************************
      subroutine network_comp(A)
!************************************************************************
! The network components of an undirected, unweighted graph are 
! determined using the breadth-first search algorithm
!************************************************************************
      implicit none

      ! Input: Adjacency matrix a
      integer, dimension(:,:), intent(in) :: A

      ! Output: Shortest path length matrix
      integer, dimension (:,:), allocatable :: splmat
      integer, dimension (:),   allocatable :: stk
      
      ! Variables: nnodes (size of network); s (source node); r (read ptr);
      ! w (write pointer) ; d (distance from s to node n), n (the node
      ! pointed to by the value of the stack at r); j (loop variable);
      integer :: nnodes,s,r,w,d,n,j
      
      nnodes = size(A,1)

      allocate(splmat(nnodes,nnodes))
      splmat = 0
      
      allocate (stk(nnodes))
      
      do s = 1,nnodes
        stk = 0
        stk(1) = s
        r = 1
        w = 2

        do while (r/=w)
          n = stk(r)
          r = r+1
          d = splmat(s,n)

          do j=1,nnodes
            if (A(n,j) .eq. 1 .and.(splmat(s,j)==0)) then
              splmat(s,j) = d+1
              stk(w) = j
              w = w+1
            end if
          end do
        end do
      end do
      write(*,*)'SPLMAT:'
      call print_matrixI(splmat,10)

      end subroutine network_comp
!************************************************************************


!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! I/O 
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!************************************************************************
      double precision function get_last_energy()
!************************************************************************
! it extracts the last energy from the file iface
!************************************************************************
      implicit none
      !local
      character(len=512) :: line,stmp
      character(len=8)   :: cscr8
      character(len=15)  :: cscr15
      integer            :: check,i


      check = 0
      open(ifcfile,file='iface',status='old')

      do while (check.eq.0)
        read(ifcfile,200,iostat=check) line        ! lines before keyword geom
        stmp=adjustl(trim(line))
      enddo

      read(stmp,7596) cscr8,cscr15,i,i,i,get_last_energy

      close(ifcfile)

200   format(a512)
7596  format(a8,1x,a15,1x,3(i2,1x),1pe23.15)

      return
      end function get_last_energy
!************************************************************************

!************************************************************************
      subroutine print_vectorR(vec,cols)
!************************************************************************
! nice :) printout for real values
!************************************************************************
      implicit none
      double precision,dimension(:),intent(in) :: vec
      integer,intent(in) :: cols ! # columns on the screen (<=99)
      !local
      integer :: i,j,k,l,m,n,n_columns
      character(len=10) :: formats
      character(len=2) :: snum

      snum      = int2char(cols)
      n_columns = size(vec)
      m = mod(n_columns,cols)
      n = n_columns/cols

      formats = '(A6,'//snum//'I12)'

      if(n .gt. 0) then
        do j=1,n
          write(*,formats,advance='no') 'col-',(cols*(j-1)+i, i=1,cols)
          write(*,*) ''
          write(*,998,advance='no') '| '
          do k=1,cols
            write(*,999,advance='no') vec(cols*(j-1)+k)
          enddo
          write(*,*)
          write(*,*)
        enddo
        ! remainder
        write(*,formats,advance='no') 'col-',(cols*n+i, i=1,m)
        write(*,*) ''
        write(*,998,advance='no') '| '
        do k=1,m
          write(*,999,advance='no') vec(cols*n+k)
        enddo
        write(*,*)
      else
        write(*,formats,advance='no') 'col-',(cols*n+i, i=1,m)
        write(*,*) ''
        write(*,998,advance='no') '| '
        do k=1,m
          write(*,999,advance='no') vec(k)
        enddo
        write(*,*)
      endif

      write(*,*) ''

998   format(A6)
999   format(f14.8)
      end subroutine print_vectorR
!************************************************************************

!************************************************************************
      subroutine print_matrixR(A,cols)
!************************************************************************
! nice :) printout for real values
!************************************************************************
      implicit none
      double precision,dimension(:,:),intent(in) :: A
      integer,intent(in) :: cols ! # columns on the screen (<=99)
      !local
      integer :: i,j,k,l,m,n,n_rows,n_columns
      character(len=10) :: formats
      character(len=2) :: snum

      snum      = int2char(cols)
      n_rows    = size(A,1)
      n_columns = size(A,2)
      m = mod(n_columns,cols)
      n = n_columns/cols

      formats = '(A6,'//snum//'I12)'

      if(n .gt. 0) then
        do j=1,n
          write(*,formats,advance='no') 'col-',(cols*(j-1)+i, i=1,cols)
          write(*,*) ''
          write(*,997) 'row|'
          do i=1,n_rows
             write(*,998,advance='no') i,'| '
             do k=1,cols
               write(*,999,advance='no') A(i,cols*(j-1)+k)
             enddo
             write(*,*)
          enddo
          write(*,*)
        enddo
        ! remainder
        write(*,formats,advance='no') 'col-',(cols*n+i, i=1,m)
        write(*,*) ''
        write(*,997) 'row|'
        do i=1,n_rows
          write(*,998,advance='no') i,'| '
          do k=1,m
            write(*,999,advance='no') A(i,cols*n+k)
          enddo
          write(*,*)
        enddo
      else
        write(*,formats,advance='no') 'col-',(cols*n+i, i=1,m)
        write(*,*) ''
        write(*,997) 'row|'
        do i=1,n_rows
          write(*,998,advance='no') i,'| '
          do k=1,m
            write(*,999,advance='no') A(i,k)
          enddo
          write(*,*)
        enddo
      endif

      write(*,*) ''

997   format(A6)
998   format(I4,A2)
999   format(f14.8)
!99   format(es13.5e2)
      end subroutine print_matrixR
!************************************************************************

!************************************************************************
      subroutine print_matrixI(A,cols)
!************************************************************************
! nice :) printout for integer values
!************************************************************************
      implicit none
      integer,dimension(:,:),intent(in) :: A
      integer,intent(in) :: cols ! # columns on the screen (<=99)
      !local
      integer :: i,j,k,l,m,n,n_rows,n_columns
      character(len=9) :: formats
      character(len=2) :: snum

      snum      = int2char(cols)
      n_rows    = size(A,1)
      n_columns = size(A,2)
      m = mod(n_columns,cols)
      n = n_columns/cols

      formats = '(A6,'//snum//'I8)'

      if(n .gt. 0) then
        do j=1,n
          write(*,formats,advance='no') 'col-',(cols*(j-1)+i, i=1,cols)
          write(*,*) ''
          write(*,997) 'row|'
          do i=1,n_rows
             write(*,998,advance='no') i,'| '
             do k=1,cols
               write(*,999,advance='no') A(i,cols*(j-1)+k)
             enddo
             write(*,*)
          enddo
          write(*,*)
        enddo
        ! remainder
        write(*,formats,advance='no') 'col-',(cols*n+i, i=1,m)
        write(*,*) ''
        write(*,997) 'row|'
        do i=1,n_rows
          write(*,998,advance='no') i,'| '
          do k=1,m
            write(*,999,advance='no') A(i,cols*n+k)
          enddo
          write(*,*)
        enddo
      else
        write(*,formats,advance='no') 'col-',(cols*n+i, i=1,m)
        write(*,*) ''
        write(*,997) 'row|'
        do i=1,n_rows
          write(*,998,advance='no') i,'| '
          do k=1,m
            write(*,999,advance='no') A(i,k)
          enddo
          write(*,*)
        enddo
      endif

      write(*,*) ''

997   format(A6)
998   format(I4,A2)
999   format(I8)
      end subroutine print_matrixI
!************************************************************************


!************************************************************************
      end module optim
!************************************************************************



